@@ -20,7 +20,6 @@
 #include "nest/iface.h"
 #include "nest/cli.h"
 #include "filter/filter.h"
-#include "filter/f-inst.h"
 
 pool *proto_pool;
 list  proto_list;
@@ -28,8 +27,7 @@ list  proto_list;
 static list protocol_list;
 struct protocol *class_to_protocol[PROTOCOL__MAX];
 
-#define CD(c, msg, args...) ({ if (c->debug & D_STATES) log(L_TRACE "%s.%s: " msg, c->proto->name, c->name ?: "?", ## args); })
-#define PD(p, msg, args...) ({ if (p->debug & D_STATES) log(L_TRACE "%s: " msg, p->name, ## args); })
+#define PD(pr, msg, args...) do { if (pr->debug & D_STATES) { log(L_TRACE "%s: " msg, pr->name , ## args); } } while(0)
 
 static timer *proto_shutdown_timer;
 static timer *gr_wait_timer;
@@ -44,11 +42,9 @@ static u32 graceful_restart_locks;
 
 static char *p_states[] = { "DOWN", "START", "UP", "STOP" };
 static char *c_states[] = { "DOWN", "START", "UP", "FLUSHING" };
-static char *e_states[] = { "DOWN", "FEEDING", "READY" };
 
 extern struct protocol proto_unix_iface;
 
-static void channel_request_reload(struct channel *c);
 static void proto_shutdown_loop(timer *);
 static void proto_rethink_goal(struct proto *p);
 static char *proto_state_name(struct proto *p);
@@ -62,18 +58,6 @@ static inline int proto_is_done(struct proto *p)
 static inline int channel_is_active(struct channel *c)
 { return (c->channel_state == CS_START) || (c->channel_state == CS_UP); }
 
-static inline int channel_reloadable(struct channel *c)
-{ return c->proto->reload_routes && c->reloadable; }
-
-static inline void
-channel_log_state_change(struct channel *c)
-{
-  if (c->export_state)
-    CD(c, "State changed to %s/%s", c_states[c->channel_state], e_states[c->export_state]);
-  else
-    CD(c, "State changed to %s", c_states[c->channel_state]);
-}
-
 static void
 proto_log_state_change(struct proto *p)
 {
@@ -175,33 +159,30 @@ proto_add_channel(struct proto *p, struct channel_config *cf)
   c->net_type = cf->net_type;
   c->ra_mode = cf->ra_mode;
   c->preference = cf->preference;
-  c->debug = cf->debug;
   c->merge_limit = cf->merge_limit;
   c->in_keep_filtered = cf->in_keep_filtered;
-  c->rpki_reload = cf->rpki_reload;
 
   c->channel_state = CS_DOWN;
   c->export_state = ES_DOWN;
   c->last_state_change = current_time();
+  c->last_tx_filter_change = current_time();
   c->reloadable = 1;
 
-  init_list(&c->roa_subscriptions);
-
   CALL(c->channel->init, c, cf);
 
   add_tail(&p->channels, &c->n);
 
-  CD(c, "Connected to table %s", c->table->name);
+  PD(p, "Channel %s connected to table %s", c->name, c->table->name);
 
   return c;
 }
 
 void
-proto_remove_channel(struct proto *p UNUSED, struct channel *c)
+proto_remove_channel(struct proto *p, struct channel *c)
 {
   ASSERT(c->channel_state == CS_DOWN);
 
-  CD(c, "Removed", c->name);
+  PD(p, "Channel %s removed", c->name);
 
   rem_node(&c->n);
   mb_free(c);
@@ -252,7 +233,7 @@ channel_schedule_feed(struct channel *c, int initial)
   c->export_state = ES_FEEDING;
   c->refeeding = !initial;
 
-  ev_schedule_work(c->feed_event);
+  ev_schedule(c->feed_event);
 }
 
 static void
@@ -263,187 +244,26 @@ channel_feed_loop(void *ptr)
   if (c->export_state != ES_FEEDING)
     return;
 
-  /* Start feeding */
   if (!c->feed_active)
-  {
     if (c->proto->feed_begin)
       c->proto->feed_begin(c, !c->refeeding);
 
-    c->refeed_pending = 0;
-  }
-
   // DBG("Feeding protocol %s continued\n", p->name);
   if (!rt_feed_channel(c))
   {
-    ev_schedule_work(c->feed_event);
-    return;
-  }
-
-  /* Reset export limit if the feed ended with acceptable number of exported routes */
-  struct channel_limit *l = &c->out_limit;
-  if (c->refeeding &&
-      (l->state == PLS_BLOCKED) &&
-      (c->refeed_count <= l->limit) &&
-      (c->stats.exp_routes <= l->limit))
-  {
-    log(L_INFO "Protocol %s resets route export limit (%u)", c->proto->name, l->limit);
-    channel_reset_limit(&c->out_limit);
-
-    /* Continue in feed - it will process routing table again from beginning */
-    c->refeed_count = 0;
-    ev_schedule_work(c->feed_event);
+    ev_schedule(c->feed_event);
     return;
   }
 
   // DBG("Feeding protocol %s finished\n", p->name);
   c->export_state = ES_READY;
-  channel_log_state_change(c);
+  // proto_log_state_change(p);
 
   if (c->proto->feed_end)
     c->proto->feed_end(c);
-
-  /* Restart feeding */
-  if (c->refeed_pending)
-    channel_request_feeding(c);
 }
 
 
-static void
-channel_roa_in_changed(struct rt_subscription *s)
-{
-  struct channel *c = s->data;
-  int active = c->reload_event && ev_active(c->reload_event);
-
-  CD(c, "Reload triggered by RPKI change%s", active ? " - already active" : "");
-
-  if (!active)
-    channel_request_reload(c);
-  else
-    c->reload_pending = 1;
-}
-
-static void
-channel_roa_out_changed(struct rt_subscription *s)
-{
-  struct channel *c = s->data;
-  int active = (c->export_state == ES_FEEDING);
-
-  CD(c, "Feeding triggered by RPKI change%s", active ? " - already active" : "");
-
-  if (!active)
-    channel_request_feeding(c);
-  else
-    c->refeed_pending = 1;
-}
-
-/* Temporary code, subscriptions should be changed to resources */
-struct roa_subscription {
-  struct rt_subscription s;
-  node roa_node;
-};
-
-static int
-channel_roa_is_subscribed(struct channel *c, rtable *tab, int dir)
-{
-  void (*hook)(struct rt_subscription *) =
-    dir ? channel_roa_in_changed : channel_roa_out_changed;
-
-  struct roa_subscription *s;
-  node *n;
-
-  WALK_LIST2(s, n, c->roa_subscriptions, roa_node)
-    if ((s->s.tab == tab) && (s->s.hook == hook))
-      return 1;
-
-  return 0;
-}
-
-
-static void
-channel_roa_subscribe(struct channel *c, rtable *tab, int dir)
-{
-  if (channel_roa_is_subscribed(c, tab, dir))
-    return;
-
-  struct roa_subscription *s = mb_allocz(c->proto->pool, sizeof(struct roa_subscription));
-
-  s->s.hook = dir ? channel_roa_in_changed : channel_roa_out_changed;
-  s->s.data = c;
-  rt_subscribe(tab, &s->s);
-
-  add_tail(&c->roa_subscriptions, &s->roa_node);
-}
-
-static void
-channel_roa_unsubscribe(struct roa_subscription *s)
-{
-  rt_unsubscribe(&s->s);
-  rem_node(&s->roa_node);
-  mb_free(s);
-}
-
-static void
-channel_roa_subscribe_filter(struct channel *c, int dir)
-{
-  const struct filter *f = dir ? c->in_filter : c->out_filter;
-  struct rtable *tab;
-  int valid = 1, found = 0;
-
-  if ((f == FILTER_ACCEPT) || (f == FILTER_REJECT))
-    return;
-
-  /* No automatic reload for non-reloadable channels */
-  if (dir && !channel_reloadable(c))
-    valid = 0;
-
-#ifdef CONFIG_BGP
-  /* No automatic reload for BGP channels without in_table / out_table */
-  if (c->channel == &channel_bgp)
-    valid = dir ? !!c->in_table : !!c->out_table;
-#endif
-
-  struct filter_iterator fit;
-  FILTER_ITERATE_INIT(&fit, f, c->proto->pool);
-
-  FILTER_ITERATE(&fit, fi)
-  {
-    switch (fi->fi_code)
-    {
-    case FI_ROA_CHECK_IMPLICIT:
-      tab = fi->i_FI_ROA_CHECK_IMPLICIT.rtc->table;
-      if (valid) channel_roa_subscribe(c, tab, dir);
-      found = 1;
-      break;
-
-    case FI_ROA_CHECK_EXPLICIT:
-      tab = fi->i_FI_ROA_CHECK_EXPLICIT.rtc->table;
-      if (valid) channel_roa_subscribe(c, tab, dir);
-      found = 1;
-      break;
-
-    default:
-      break;
-    }
-  }
-  FILTER_ITERATE_END;
-
-  FILTER_ITERATE_CLEANUP(&fit);
-
-  if (!valid && found)
-    log(L_WARN "%s.%s: Automatic RPKI reload not active for %s",
-	c->proto->name, c->name ?: "?", dir ? "import" : "export");
-}
-
-static void
-channel_roa_unsubscribe_all(struct channel *c)
-{
-  struct roa_subscription *s;
-  node *n, *x;
-
-  WALK_LIST2_DELSAFE(s, n, x, c->roa_subscriptions, roa_node)
-    channel_roa_unsubscribe(s);
-}
-
 static void
 channel_start_export(struct channel *c)
 {
@@ -462,7 +282,6 @@ channel_stop_export(struct channel *c)
 
   c->export_state = ES_DOWN;
   c->stats.exp_routes = 0;
-  bmap_reset(&c->export_map, 1024);
 }
 
 
@@ -473,7 +292,7 @@ channel_schedule_reload(struct channel *c)
   ASSERT(c->channel_state == CS_UP);
 
   rt_reload_channel_abort(c);
-  ev_schedule_work(c->reload_event);
+  ev_schedule(c->reload_event);
 }
 
 static void
@@ -481,19 +300,11 @@ channel_reload_loop(void *ptr)
 {
   struct channel *c = ptr;
 
-  /* Start reload */
-  if (!c->reload_active)
-    c->reload_pending = 0;
-
   if (!rt_reload_channel(c))
   {
-    ev_schedule_work(c->reload_event);
+    ev_schedule(c->reload_event);
     return;
   }
-
-  /* Restart reload */
-  if (c->reload_pending)
-    channel_request_reload(c);
 }
 
 static void
@@ -506,40 +317,20 @@ channel_reset_import(struct channel *c)
   rt_prune_sync(c->in_table, 1);
 }
 
-static void
-channel_reset_export(struct channel *c)
-{
-  /* Just free the routes */
-  rt_prune_sync(c->out_table, 1);
-}
-
 /* Called by protocol to activate in_table */
 void
 channel_setup_in_table(struct channel *c)
 {
   struct rtable_config *cf = mb_allocz(c->proto->pool, sizeof(struct rtable_config));
-
   cf->name = "import";
   cf->addr_type = c->net_type;
-  cf->internal = 1;
 
-  c->in_table = rt_setup(c->proto->pool, cf);
+  c->in_table = mb_allocz(c->proto->pool, sizeof(struct rtable));
+  rt_setup(c->proto->pool, c->in_table, cf);
 
   c->reload_event = ev_new_init(c->proto->pool, channel_reload_loop, c);
 }
 
-/* Called by protocol to activate out_table */
-void
-channel_setup_out_table(struct channel *c)
-{
-  struct rtable_config *cf = mb_allocz(c->proto->pool, sizeof(struct rtable_config));
-  cf->name = "export";
-  cf->addr_type = c->net_type;
-  cf->internal = 1;
-
-  c->out_table = rt_setup(c->proto->pool, cf);
-}
-
 
 static void
 channel_do_start(struct channel *c)
@@ -550,9 +341,6 @@ channel_do_start(struct channel *c)
 
   c->feed_event = ev_new_init(c->proto->pool, channel_feed_loop, c);
 
-  bmap_init(&c->export_map, c->proto->pool, 1024);
-  memset(&c->stats, 0, sizeof(struct proto_stats));
-
   channel_reset_limit(&c->rx_limit);
   channel_reset_limit(&c->in_limit);
   channel_reset_limit(&c->out_limit);
@@ -560,17 +348,6 @@ channel_do_start(struct channel *c)
   CALL(c->channel->start, c);
 }
 
-static void
-channel_do_up(struct channel *c)
-{
-  /* Register RPKI/ROA subscriptions */
-  if (c->rpki_reload)
-  {
-    channel_roa_subscribe_filter(c, 1);
-    channel_roa_subscribe_filter(c, 0);
-  }
-}
-
 static void
 channel_do_flush(struct channel *c)
 {
@@ -581,14 +358,6 @@ channel_do_flush(struct channel *c)
     channel_graceful_restart_unlock(c);
 
   CALL(c->channel->shutdown, c);
-
-  /* This have to be done in here, as channel pool is freed before channel_do_down() */
-  bmap_free(&c->export_map);
-  c->in_table = NULL;
-  c->reload_event = NULL;
-  c->out_table = NULL;
-
-  channel_roa_unsubscribe_all(c);
 }
 
 static void
@@ -603,14 +372,10 @@ channel_do_down(struct channel *c)
   if ((c->stats.imp_routes + c->stats.filt_routes) != 0)
     log(L_ERR "%s: Channel %s is down but still has some routes", c->proto->name, c->name);
 
-  // bmap_free(&c->export_map);
   memset(&c->stats, 0, sizeof(struct proto_stats));
 
   c->in_table = NULL;
   c->reload_event = NULL;
-  c->out_table = NULL;
-
-  /* The in_table and out_table are going to be freed by freeing their resource pools. */
 
   CALL(c->channel->cleanup, c);
 
@@ -646,9 +411,6 @@ channel_set_state(struct channel *c, uint state)
     if (c->in_table && (cs == CS_UP))
       channel_reset_import(c);
 
-    if (c->out_table && (cs == CS_UP))
-      channel_reset_export(c);
-
     break;
 
   case CS_UP:
@@ -660,7 +422,6 @@ channel_set_state(struct channel *c, uint state)
     if (!c->gr_wait && c->proto->rt_notify)
       channel_start_export(c);
 
-    channel_do_up(c);
     break;
 
   case CS_FLUSHING:
@@ -672,9 +433,6 @@ channel_set_state(struct channel *c, uint state)
     if (c->in_table && (cs == CS_UP))
       channel_reset_import(c);
 
-    if (c->out_table && (cs == CS_UP))
-      channel_reset_export(c);
-
     channel_do_flush(c);
     break;
 
@@ -687,8 +445,7 @@ channel_set_state(struct channel *c, uint state)
   default:
     ASSERT(0);
   }
-
-  channel_log_state_change(c);
+  // XXXX proto_log_state_change(c);
 }
 
 /**
@@ -706,8 +463,6 @@ channel_request_feeding(struct channel *c)
 {
   ASSERT(c->channel_state == CS_UP);
 
-  CD(c, "Feeding requested");
-
   /* Do nothing if we are still waiting for feeding */
   if (c->export_state == ES_DOWN)
     return;
@@ -722,11 +477,19 @@ channel_request_feeding(struct channel *c)
     rt_feed_channel_abort(c);
   }
 
-  /* Track number of exported routes during refeed */
-  c->refeed_count = 0;
+  channel_reset_limit(&c->out_limit);
+
+  /* Hack: reset exp_routes during refeed, and do not decrease it later */
+  c->stats.exp_routes = 0;
 
   channel_schedule_feed(c, 0);	/* Sets ES_FEEDING */
-  channel_log_state_change(c);
+  // proto_log_state_change(c);
+}
+
+static inline int
+channel_reloadable(struct channel *c)
+{
+  return c->proto->reload_routes && c->reloadable;
 }
 
 static void
@@ -735,8 +498,6 @@ channel_request_reload(struct channel *c)
   ASSERT(c->channel_state == CS_UP);
   ASSERT(channel_reloadable(c));
 
-  CD(c, "Reload requested");
-
   c->proto->reload_routes(c);
 
   /*
@@ -782,8 +543,6 @@ channel_config_new(const struct channel_class *cc, const char *name, uint net_ty
   cf->net_type = net_type;
   cf->ra_mode = RA_OPTIMAL;
   cf->preference = proto->protocol->preference;
-  cf->debug = new_config->channel_default_debug;
-  cf->rpki_reload = 1;
 
   add_tail(&proto->channels, &cf->n);
 
@@ -816,7 +575,6 @@ channel_copy_config(struct channel_config *src, struct proto_config *proto)
   struct channel_config *dst = cfg_alloc(src->channel->config_size);
 
   memcpy(dst, src, src->channel->config_size);
-  memset(&dst->n, 0, sizeof(node));
   add_tail(&proto->channels, &dst->n);
   CALL(src->channel->copy_config, dst, src);
 
@@ -836,7 +594,6 @@ channel_reconfigure(struct channel *c, struct channel_config *cf)
   /* Note that filter_same() requires arguments in (new, old) order */
   int import_changed = !filter_same(cf->in_filter, c->in_filter);
   int export_changed = !filter_same(cf->out_filter, c->out_filter);
-  int rpki_reload_changed = (cf->rpki_reload != c->rpki_reload);
 
   if (c->preference != cf->preference)
     import_changed = 1;
@@ -854,31 +611,20 @@ channel_reconfigure(struct channel *c, struct channel_config *cf)
   // c->ra_mode = cf->ra_mode;
   c->merge_limit = cf->merge_limit;
   c->preference = cf->preference;
-  c->debug = cf->debug;
   c->in_keep_filtered = cf->in_keep_filtered;
-  c->rpki_reload = cf->rpki_reload;
 
   channel_verify_limits(c);
 
+  if (export_changed)
+    c->last_tx_filter_change = current_time();
+
   /* Execute channel-specific reconfigure hook */
-  if (c->channel->reconfigure && !c->channel->reconfigure(c, cf, &import_changed, &export_changed))
+  if (c->channel->reconfigure && !c->channel->reconfigure(c, cf))
     return 0;
 
   /* If the channel is not open, it has no routes and we cannot reload it anyways */
   if (c->channel_state != CS_UP)
-    goto done;
-
-  /* Update RPKI/ROA subscriptions */
-  if (import_changed || export_changed || rpki_reload_changed)
-  {
-    channel_roa_unsubscribe_all(c);
-
-    if (c->rpki_reload)
-    {
-      channel_roa_subscribe_filter(c, 1);
-      channel_roa_subscribe_filter(c, 0);
-    }
-  }
+    return 1;
 
   if (reconfigure_type == RECONFIG_SOFT)
   {
@@ -888,7 +634,7 @@ channel_reconfigure(struct channel *c, struct channel_config *cf)
     if (export_changed)
       log(L_INFO "Channel %s.%s changed export", c->proto->name, c->name);
 
-    goto done;
+    return 1;
   }
 
   /* Route reload may be not supported */
@@ -904,8 +650,6 @@ channel_reconfigure(struct channel *c, struct channel_config *cf)
   if (export_changed)
     channel_request_feeding(c);
 
-done:
-  CD(c, "Reconfigured");
   return 1;
 }
 
@@ -1021,7 +765,6 @@ proto_init(struct proto_config *c, node *n)
   p->proto_state = PS_DOWN;
   p->last_state_change = current_time();
   p->vrf = c->vrf;
-  p->vrf_set = c->vrf_set;
   insert_node(&p->n, n);
 
   p->event = ev_new_init(proto_pool, proto_event, p);
@@ -1098,7 +841,7 @@ proto_copy_config(struct proto_config *dest, struct proto_config *src)
   struct channel_config *cc;
   node old_node;
   int old_class;
-  const char *old_name;
+  char *old_name;
 
   if (dest->protocol != src->protocol)
     cf_error("Can't copy configuration from a different protocol type");
@@ -1131,28 +874,6 @@ proto_copy_config(struct proto_config *dest, struct proto_config *src)
   dest->protocol->copy_config(dest, src);
 }
 
-void
-proto_clone_config(struct symbol *sym, struct proto_config *parent)
-{
-  struct proto_config *cf = proto_config_new(parent->protocol, SYM_PROTO);
-  proto_copy_config(cf, parent);
-  cf->name = sym->name;
-  cf->proto = NULL;
-  cf->parent = parent;
-
-  sym->class = cf->class;
-  sym->proto = cf;
-}
-
-static void
-proto_undef_clone(struct symbol *sym, struct proto_config *cf)
-{
-  rem_node(&cf->n);
-
-  sym->class = SYM_VOID;
-  sym->proto = NULL;
-}
-
 /**
  * protos_preconfig - pre-configuration processing
  * @c: new configuration
@@ -1189,8 +910,7 @@ proto_reconfigure(struct proto *p, struct proto_config *oc, struct proto_config
   if ((nc->protocol != oc->protocol) ||
       (nc->net_type != oc->net_type) ||
       (nc->disabled != p->disabled) ||
-      (nc->vrf != oc->vrf) ||
-      (nc->vrf_set != oc->vrf_set))
+      (nc->vrf != oc->vrf))
     return 0;
 
   p->name = nc->name;
@@ -1253,41 +973,17 @@ protos_commit(struct config *new, struct config *old, int force_reconfig, int ty
     {
       p = oc->proto;
       sym = cf_find_symbol(new, oc->name);
-
-      /* Handle dynamic protocols */
-      if (!sym && oc->parent && !new->shutdown)
-      {
-	struct symbol *parsym = cf_find_symbol(new, oc->parent->name);
-	if (parsym && parsym->class == SYM_PROTO)
-	{
-	  /* This is hack, we would like to share config, but we need to copy it now */
-	  new_config = new;
-	  cfg_mem = new->mem;
-	  conf_this_scope = new->root_scope;
-	  sym = cf_get_symbol(oc->name);
-	  proto_clone_config(sym, parsym->proto);
-	  new_config = NULL;
-	  cfg_mem = NULL;
-	}
-      }
-
       if (sym && sym->class == SYM_PROTO && !new->shutdown)
       {
 	/* Found match, let's check if we can smoothly switch to new configuration */
 	/* No need to check description */
-	nc = sym->proto;
+	nc = sym->def;
 	nc->proto = p;
 
 	/* We will try to reconfigure protocol p */
 	if (! force_reconfig && proto_reconfigure(p, oc, nc, type))
 	  continue;
 
-	if (nc->parent)
-	{
-	  proto_undef_clone(sym, nc);
-	  goto remove;
-	}
-
 	/* Unsuccessful, we will restart it */
 	if (!p->disabled && !nc->disabled)
 	  log(L_INFO "Restarting protocol %s", p->name);
@@ -1301,16 +997,10 @@ protos_commit(struct config *new, struct config *old, int force_reconfig, int ty
       }
       else if (!new->shutdown)
       {
-      remove:
 	log(L_INFO "Removing protocol %s", p->name);
 	p->down_code = PDC_CF_REMOVE;
 	p->cf_new = NULL;
       }
-      else if (new->gr_down)
-      {
-	p->down_code = PDC_CMD_GR_DOWN;
-	p->cf_new = NULL;
-      }
       else /* global shutdown */
       {
 	p->down_code = PDC_CMD_SHUTDOWN;
@@ -1415,15 +1105,6 @@ proto_rethink_goal(struct proto *p)
   }
 }
 
-struct proto *
-proto_spawn(struct proto_config *cf, uint disabled)
-{
-  struct proto *p = proto_init(cf, TAIL(proto_list));
-  p->disabled = disabled;
-  proto_rethink_goal(p);
-  return p;
-}
-
 
 /**
  * DOC: Graceful restart recovery
@@ -2066,16 +1747,6 @@ channel_show_info(struct channel *c)
     channel_show_stats(c);
 }
 
-void
-channel_cmd_debug(struct channel *c, uint mask)
-{
-  if (cli_access_restricted())
-    return;
-
-  c->debug = mask;
-  cli_msg(0, "");
-}
-
 void
 proto_cmd_show(struct proto *p, uintptr_t verbose, int cnt)
 {
@@ -2106,8 +1777,8 @@ proto_cmd_show(struct proto *p, uintptr_t verbose, int cnt)
       cli_msg(-1006, "  Message:        %s", p->message);
     if (p->cf->router_id)
       cli_msg(-1006, "  Router ID:      %R", p->cf->router_id);
-    if (p->vrf_set)
-      cli_msg(-1006, "  VRF:            %s", p->vrf ? p->vrf->name : "default");
+    if (p->vrf)
+      cli_msg(-1006, "  VRF:            %s", p->vrf->name);
 
     if (p->proto->show_proto_info)
       p->proto->show_proto_info(p);
@@ -2192,7 +1863,7 @@ proto_cmd_reload(struct proto *p, uintptr_t dir, int cnt UNUSED)
   /* All channels must support reload */
   if (dir != CMD_RELOAD_OUT)
     WALK_LIST(c, p->channels)
-      if ((c->channel_state == CS_UP) && !channel_reloadable(c))
+      if (!channel_reloadable(c))
       {
 	cli_msg(-8006, "%s: reload failed", p->name);
 	return;
@@ -2203,29 +1874,20 @@ proto_cmd_reload(struct proto *p, uintptr_t dir, int cnt UNUSED)
   /* re-importing routes */
   if (dir != CMD_RELOAD_OUT)
     WALK_LIST(c, p->channels)
-      if (c->channel_state == CS_UP)
-	channel_request_reload(c);
+      channel_request_reload(c);
 
   /* re-exporting routes */
   if (dir != CMD_RELOAD_IN)
     WALK_LIST(c, p->channels)
-      if (c->channel_state == CS_UP)
-	channel_request_feeding(c);
+      channel_request_feeding(c);
 
   cli_msg(-15, "%s: reloading", p->name);
 }
 
-extern void pipe_update_debug(struct proto *P);
-
 void
 proto_cmd_debug(struct proto *p, uintptr_t mask, int cnt UNUSED)
 {
   p->debug = mask;
-
-#ifdef CONFIG_PIPE
-  if (p->proto == &proto_pipe)
-    pipe_update_debug(p);
-#endif
 }
 
 void
@@ -2235,7 +1897,7 @@ proto_cmd_mrtdump(struct proto *p, uintptr_t mask, int cnt UNUSED)
 }
 
 static void
-proto_apply_cmd_symbol(const struct symbol *s, void (* cmd)(struct proto *, uintptr_t, int), uintptr_t arg)
+proto_apply_cmd_symbol(struct symbol *s, void (* cmd)(struct proto *, uintptr_t, int), uintptr_t arg)
 {
   if (s->class != SYM_PROTO)
   {
@@ -2243,12 +1905,12 @@ proto_apply_cmd_symbol(const struct symbol *s, void (* cmd)(struct proto *, uint
     return;
   }
 
-  cmd(s->proto->proto, arg, 0);
+  cmd(((struct proto_config *)s->def)->proto, arg, 0);
   cli_msg(0, "");
 }
 
 static void
-proto_apply_cmd_patt(const char *patt, void (* cmd)(struct proto *, uintptr_t, int), uintptr_t arg)
+proto_apply_cmd_patt(char *patt, void (* cmd)(struct proto *, uintptr_t, int), uintptr_t arg)
 {
   struct proto *p;
   int cnt = 0;
@@ -2286,7 +1948,7 @@ proto_get_named(struct symbol *sym, struct protocol *pr)
     if (sym->class != SYM_PROTO)
       cf_error("%s: Not a protocol", sym->name);
 
-    p = sym->proto->proto;
+    p = ((struct proto_config *) sym->def)->proto;
     if (!p || p->proto != pr)
       cf_error("%s: Not a %s protocol", sym->name, pr->name);
   }
@@ -2306,47 +1968,3 @@ proto_get_named(struct symbol *sym, struct protocol *pr)
 
   return p;
 }
-
-struct proto *
-proto_iterate_named(struct symbol *sym, struct protocol *proto, struct proto *old)
-{
-  if (sym)
-  {
-    /* Just the first pass */
-    if (old)
-    {
-      cli_msg(0, "");
-      return NULL;
-    }
-
-    if (sym->class != SYM_PROTO)
-      cf_error("%s: Not a protocol", sym->name);
-
-    struct proto *p = sym->proto->proto;
-    if (!p || (p->proto != proto))
-      cf_error("%s: Not a %s protocol", sym->name, proto->name);
-
-    return p;
-  }
-  else
-  {
-    for (struct proto *p = !old ? HEAD(proto_list) : NODE_NEXT(old);
-	 NODE_VALID(p);
-	 p = NODE_NEXT(p))
-    {
-      if ((p->proto == proto) && (p->proto_state != PS_DOWN))
-      {
-	cli_separator(this_cli);
-	return p;
-      }
-    }
-
-    /* Not found anything during first pass */
-    if (!old)
-      cf_error("There is no %s protocol running", proto->name);
-
-    /* No more items */
-    cli_msg(0, "");
-    return NULL;
-  }
-}
