@@ -38,8 +38,6 @@
 #include <stdlib.h>
 #include "babel.h"
 
-#define LOG_PKT_AUTH(msg, args...) \
-  log_rl(&p->log_pkt_tbf, L_AUTH "%s: " msg, p->p.name, args)
 
 /*
  * Is one number greater or equal than another mod 2^16? This is based on the
@@ -58,12 +56,19 @@ static void babel_update_cost(struct babel_neighbor *n);
 static inline void babel_kick_timer(struct babel_proto *p);
 static inline void babel_iface_kick_timer(struct babel_iface *ifa);
 
+static inline void babel_lock_neighbor(struct babel_neighbor *nbr)
+{ if (nbr) nbr->uc++; }
+
+static inline void babel_unlock_neighbor(struct babel_neighbor *nbr)
+{ if (nbr && !--nbr->uc) mb_free(nbr); }
+
+
 /*
  *	Functions to maintain data structures
  */
 
 static void
-babel_init_entry(struct fib *f UNUSED, void *E)
+babel_init_entry(void *E)
 {
   struct babel_entry *e = E;
 
@@ -108,7 +113,7 @@ babel_get_source(struct babel_proto *p, struct babel_entry *e, u64 router_id)
   if (s)
     return s;
 
-  s = sl_allocz(p->source_slab);
+  s = sl_alloc(p->source_slab);
   s->router_id = router_id;
   s->expires = current_time() + BABEL_GARBAGE_INTERVAL;
   s->seqno = 0;
@@ -119,7 +124,7 @@ babel_get_source(struct babel_proto *p, struct babel_entry *e, u64 router_id)
 }
 
 static void
-babel_expire_sources(struct babel_proto *p UNUSED, struct babel_entry *e)
+babel_expire_sources(struct babel_proto *p, struct babel_entry *e)
 {
   struct babel_source *n, *nx;
   btime now_ = current_time();
@@ -129,7 +134,7 @@ babel_expire_sources(struct babel_proto *p UNUSED, struct babel_entry *e)
     if (n->expires && n->expires <= now_)
     {
       rem_node(NODE n);
-      sl_free(n);
+      sl_free(p->source_slab, n);
     }
   }
 }
@@ -154,7 +159,8 @@ babel_get_route(struct babel_proto *p, struct babel_entry *e, struct babel_neigh
   if (r)
     return r;
 
-  r = sl_allocz(p->route_slab);
+  r = sl_alloc(p->route_slab);
+  memset(r, 0, sizeof(*r));
 
   r->e = e;
   r->neigh = nbr;
@@ -174,7 +180,7 @@ babel_retract_route(struct babel_proto *p, struct babel_route *r)
 }
 
 static void
-babel_flush_route(struct babel_proto *p UNUSED, struct babel_route *r)
+babel_flush_route(struct babel_proto *p, struct babel_route *r)
 {
   DBG("Babel: Flush route %N router_id %lR neigh %I\n",
       r->e->n.addr, r->router_id, r->neigh->addr);
@@ -185,7 +191,7 @@ babel_flush_route(struct babel_proto *p UNUSED, struct babel_route *r)
   if (r->e->selected == r)
     r->e->selected = NULL;
 
-  sl_free(r);
+  sl_free(p->route_slab, r);
 }
 
 static void
@@ -311,16 +317,13 @@ babel_add_seqno_request(struct babel_proto *p, struct babel_entry *e,
 	return;
 
       /* Found older */
+      babel_unlock_neighbor(sr->nbr);
       rem_node(NODE sr);
-
-      if (sr->nbr)
-        rem_node(&sr->nbr_node);
-
       goto found;
     }
 
   /* No entries found */
-  sr = sl_allocz(p->seqno_slab);
+  sr = sl_alloc(p->seqno_slab);
 
 found:
   sr->router_id = router_id;
@@ -328,23 +331,18 @@ found:
   sr->hop_count = hop_count;
   sr->count = 0;
   sr->expires = current_time() + BABEL_SEQNO_REQUEST_EXPIRY;
-
-  if (sr->nbr = nbr)
-    add_tail(&nbr->requests, &sr->nbr_node);
-
+  babel_lock_neighbor(sr->nbr = nbr);
   add_tail(&e->requests, NODE sr);
 
   babel_send_seqno_request(p, e, sr);
 }
 
 static void
-babel_remove_seqno_request(struct babel_proto *p UNUSED, struct babel_seqno_request *sr)
+babel_remove_seqno_request(struct babel_proto *p, struct babel_seqno_request *sr)
 {
-  if (sr->nbr)
-    rem_node(&sr->nbr_node);
-
+  babel_unlock_neighbor(sr->nbr);
   rem_node(NODE sr);
-  sl_free(sr);
+  sl_free(p->seqno_slab, sr);
 }
 
 static int
@@ -429,9 +427,8 @@ babel_get_neighbor(struct babel_iface *ifa, ip_addr addr)
   nbr->rxcost = BABEL_INFINITY;
   nbr->txcost = BABEL_INFINITY;
   nbr->cost = BABEL_INFINITY;
-  nbr->init_expiry = current_time() + BABEL_INITIAL_NEIGHBOR_TIMEOUT;
   init_list(&nbr->routes);
-  init_list(&nbr->requests);
+  babel_lock_neighbor(nbr);
   add_tail(&ifa->neigh_list, NODE nbr);
 
   return nbr;
@@ -452,13 +449,9 @@ babel_flush_neighbor(struct babel_proto *p, struct babel_neighbor *nbr)
     babel_flush_route(p, r);
   }
 
-  struct babel_seqno_request *sr;
-  WALK_LIST_FIRST2(sr, nbr_node, nbr->requests)
-    babel_remove_seqno_request(p, sr);
-
   nbr->ifa = NULL;
   rem_node(NODE nbr);
-  mb_free(nbr);
+  babel_unlock_neighbor(nbr);
 }
 
 static void
@@ -509,15 +502,13 @@ babel_expire_neighbors(struct babel_proto *p)
       if (nbr->ihu_expiry && nbr->ihu_expiry <= now_)
         babel_expire_ihu(p, nbr);
 
-      if (nbr->init_expiry && nbr->init_expiry <= now_)
-      { babel_flush_neighbor(p, nbr); continue; }
-
       if (nbr->hello_expiry && nbr->hello_expiry <= now_)
-      { babel_expire_hello(p, nbr, now_); continue; }
+        babel_expire_hello(p, nbr, now_);
     }
   }
 }
 
+
 /*
  *	Best route selection
  */
@@ -640,48 +631,21 @@ babel_announce_rte(struct babel_proto *p, struct babel_entry *e)
   if (r)
   {
     rta a0 = {
+      .src = p->p.main_source,
       .source = RTS_BABEL,
       .scope = SCOPE_UNIVERSE,
       .dest = RTD_UNICAST,
-      .pref = c->preference,
       .from = r->neigh->addr,
       .nh.gw = r->next_hop,
       .nh.iface = r->neigh->ifa->iface,
-      .eattrs = alloca(sizeof(ea_list) + 3*sizeof(eattr)),
     };
 
-    *a0.eattrs = (ea_list) { .count = 3 };
-    a0.eattrs->attrs[0] = (eattr) {
-      .id = EA_BABEL_METRIC,
-      .type = EAF_TYPE_INT,
-      .u.data = r->metric,
-    };
-
-    struct adata *ad = alloca(sizeof(struct adata) + sizeof(u64));
-    ad->length = sizeof(u64);
-    memcpy(ad->data, &(r->router_id), sizeof(u64));
-    a0.eattrs->attrs[1] = (eattr) {
-      .id = EA_BABEL_ROUTER_ID,
-      .type = EAF_TYPE_OPAQUE,
-      .u.ptr = ad,
-    };
-
-    a0.eattrs->attrs[2] = (eattr) {
-      .id = EA_BABEL_SEQNO,
-      .type = EAF_TYPE_INT,
-      .u.data = r->seqno,
-    };
-
-    /*
-     * If we cannot find a reachable neighbour, set the entry to be onlink. This
-     * makes it possible to, e.g., assign /32 addresses on a mesh interface and
-     * have routing work.
-     */
-    if (!neigh_find(&p->p, r->next_hop, r->neigh->ifa->iface, 0))
-      a0.nh.flags = RNF_ONLINK;
-
     rta *a = rta_lookup(&a0);
-    rte *rte = rte_get_temp(a, p->p.main_source);
+    rte *rte = rte_get_temp(a);
+    rte->u.babel.seqno = r->seqno;
+    rte->u.babel.metric = r->metric;
+    rte->u.babel.router_id = r->router_id;
+    rte->pflags = 0;
 
     e->unreachable = 0;
     rte_update2(c, e->n.addr, rte, p->p.main_source);
@@ -690,15 +654,17 @@ babel_announce_rte(struct babel_proto *p, struct babel_entry *e)
   {
     /* Unreachable */
     rta a0 = {
+      .src = p->p.main_source,
       .source = RTS_BABEL,
       .scope = SCOPE_UNIVERSE,
       .dest = RTD_UNREACHABLE,
-      .pref = 1,
     };
 
     rta *a = rta_lookup(&a0);
-    rte *rte = rte_get_temp(a, p->p.main_source);
+    rte *rte = rte_get_temp(a);
+    memset(&rte->u.babel, 0, sizeof(rte->u.babel));
     rte->pflags = 0;
+    rte->pref = 1;
 
     e->unreachable = 1;
     rte_update2(c, e->n.addr, rte, p->p.main_source);
@@ -858,14 +824,14 @@ babel_send_ihus(struct babel_iface *ifa)
 }
 
 static void
-babel_send_hello(struct babel_iface *ifa, uint interval)
+babel_send_hello(struct babel_iface *ifa)
 {
   struct babel_proto *p = ifa->proto;
   union babel_msg msg = {};
 
   msg.type = BABEL_TLV_HELLO;
   msg.hello.seqno = ifa->hello_seqno++;
-  msg.hello.interval = interval ?: ifa->cf->hello_interval;
+  msg.hello.interval = ifa->cf->hello_interval;
 
   TRACE(D_PACKETS, "Sending hello on %s with seqno %d interval %t",
 	ifa->ifname, msg.hello.seqno, (btime) msg.hello.interval);
@@ -1130,9 +1096,6 @@ babel_update_hello_history(struct babel_neighbor *n, u16 seqno, uint interval)
   /* Update expiration */
   n->hello_expiry = current_time() + BABEL_HELLO_EXPIRY_FACTOR(interval);
   n->last_hello_int = interval;
-
-  /* Disable initial timeout */
-  n->init_expiry = 0;
 }
 
 
@@ -1412,130 +1375,6 @@ babel_handle_seqno_request(union babel_msg *m, struct babel_iface *ifa)
   }
 }
 
-/*
- *      Authentication functions
- */
-
-/**
- * babel_auth_reset_index - Reset authentication index on interface
- * @ifa: Interface to reset
- *
- * This function resets the authentication index and packet counter for an
- * interface, and should be called on interface configuration, or when the
- * packet counter overflows.
- */
-void
-babel_auth_reset_index(struct babel_iface *ifa)
-{
-  random_bytes(ifa->auth_index, BABEL_AUTH_INDEX_LEN);
-  ifa->auth_pc = 1;
-}
-
-static void
-babel_auth_send_challenge_request(struct babel_iface *ifa, struct babel_neighbor *n)
-{
-  struct babel_proto *p = ifa->proto;
-  union babel_msg msg = {};
-
-  TRACE(D_PACKETS, "Sending challenge request to %I on %s",
-	n->addr, ifa->ifname);
-
-  random_bytes(n->auth_nonce, BABEL_AUTH_NONCE_LEN);
-  n->auth_nonce_expiry = current_time() + BABEL_AUTH_CHALLENGE_TIMEOUT;
-  n->auth_next_challenge = current_time() + BABEL_AUTH_CHALLENGE_INTERVAL;
-
-  msg.type = BABEL_TLV_CHALLENGE_REQUEST;
-  msg.challenge.nonce_len = BABEL_AUTH_NONCE_LEN;
-  msg.challenge.nonce = n->auth_nonce;
-
-  babel_send_unicast(&msg, ifa, n->addr);
-}
-
-static void
-babel_auth_send_challenge_reply(struct babel_iface *ifa, struct babel_neighbor *n, struct babel_msg_auth *rcv)
-{
-  struct babel_proto *p = ifa->proto;
-  union babel_msg msg = {};
-
-  TRACE(D_PACKETS, "Sending challenge reply to %I on %s",
-	n->addr, ifa->ifname);
-
-  n->auth_next_challenge_reply = current_time() + BABEL_AUTH_CHALLENGE_INTERVAL;
-
-  msg.type = BABEL_TLV_CHALLENGE_REPLY;
-  msg.challenge.nonce_len = rcv->challenge_len;
-  msg.challenge.nonce = rcv->challenge;
-
-  babel_send_unicast(&msg, ifa, n->addr);
-}
-
-int
-babel_auth_check_pc(struct babel_iface *ifa, struct babel_msg_auth *msg)
-{
-  struct babel_proto *p = ifa->proto;
-  struct babel_neighbor *n;
-
-  /*
-   * We create the neighbour entry at this point because it makes it easier to
-   * rate limit challenge replies; this is explicitly allowed by the spec (see
-   * Section 4.3).
-   */
-  n = babel_get_neighbor(ifa, msg->sender);
-
-  /* (3b) Handle challenge request */
-  if (msg->challenge_seen && (n->auth_next_challenge_reply <= current_time()))
-    babel_auth_send_challenge_reply(ifa, n, msg);
-
-  /* (4a) If PC TLV is missing, drop the packet */
-  if (!msg->pc_seen)
-  {
-    LOG_PKT_AUTH("Authentication failed for %I on %s - missing or invalid PC",
-                 msg->sender, ifa->ifname);
-    return 0;
-  }
-
-  /* (4b) On successful challenge, update PC and index to current values */
-  if (msg->challenge_reply_seen &&
-      (n->auth_nonce_expiry > current_time()) &&
-      !memcmp(msg->challenge_reply, n->auth_nonce, BABEL_AUTH_NONCE_LEN))
-  {
-    n->auth_index_len = msg->index_len;
-    memcpy(n->auth_index, msg->index, msg->index_len);
-
-    n->auth_pc = msg->pc;
-    n->auth_passed = 1;
-
-    return 1;
-  }
-
-  /* (5) If index differs, send challenge and drop the packet */
-  if ((n->auth_index_len != msg->index_len) ||
-      memcmp(n->auth_index, msg->index, msg->index_len))
-  {
-    TRACE(D_PACKETS, "Index mismatch for packet from %I via %s",
-	  msg->sender, ifa->ifname);
-
-    if (n->auth_next_challenge <= current_time())
-      babel_auth_send_challenge_request(ifa, n);
-
-    return 0;
-  }
-
-  /* (6) Index matches; only accept if PC is greater than last */
-  if (n->auth_pc >= msg->pc)
-  {
-    LOG_PKT_AUTH("Authentication failed for %I on %s - "
-		 "lower packet counter (rcv %u, old %u)",
-                 msg->sender, ifa->ifname, msg->pc, n->auth_pc);
-    return 0;
-  }
-
-  n->auth_pc = msg->pc;
-  n->auth_passed = 1;
-
-  return 1;
-}
-
 
 /*
  *	Babel interfaces
@@ -1573,7 +1412,7 @@ babel_iface_timer(timer *t)
 
   if (now_ >= ifa->next_hello)
   {
-    babel_send_hello(ifa, 0);
+    babel_send_hello(ifa);
     ifa->next_hello += hello_period * (1 + (now_ - ifa->next_hello) / hello_period);
   }
 
@@ -1620,7 +1459,7 @@ babel_iface_start(struct babel_iface *ifa)
   tm_start(ifa->timer, 100 MS);
   ifa->up = 1;
 
-  babel_send_hello(ifa, 0);
+  babel_send_hello(ifa);
   babel_send_wildcard_retraction(ifa);
   babel_send_wildcard_request(ifa);
   babel_send_update(ifa, 0);	/* Full update */
@@ -1704,8 +1543,6 @@ babel_iface_update_buffers(struct babel_iface *ifa)
   sk_set_tbsize(ifa->sk, tbsize);
 
   ifa->tx_length = tbsize - BABEL_OVERHEAD;
-
-  babel_auth_set_tx_overhead(ifa);
 }
 
 static struct babel_iface*
@@ -1765,9 +1602,6 @@ babel_add_iface(struct babel_proto *p, struct iface *new, struct babel_iface_con
   init_list(&ifa->neigh_list);
   ifa->hello_seqno = 1;
 
-  if (ic->auth_type != BABEL_AUTH_NONE)
-    babel_auth_reset_index(ifa);
-
   ifa->timer = tm_new_init(ifa->pool, babel_iface_timer, ifa, 0, 0);
 
   init_list(&ifa->msg_queue);
@@ -1798,20 +1632,6 @@ babel_remove_iface(struct babel_proto *p, struct babel_iface *ifa)
   rfree(ifa->pool); /* contains ifa itself, locks, socket, etc */
 }
 
-static int
-iface_is_valid(struct babel_proto *p, struct iface *iface)
-{
-  if (!(iface->flags & IF_MULTICAST))
-  {
-    log(L_ERR "%s: Interface %s does not support multicast",
-	p->p.name, iface->name);
-
-    return 0;
-  }
-
-  return 1;
-}
-
 static void
 babel_if_notify(struct proto *P, unsigned flags, struct iface *iface)
 {
@@ -1831,13 +1651,16 @@ babel_if_notify(struct proto *P, unsigned flags, struct iface *iface)
     if (!(iface->flags & IF_UP))
       return;
 
+    /* We only speak multicast */
+    if (!(iface->flags & IF_MULTICAST))
+      return;
+
     /* Ignore ifaces without link-local address */
     if (!iface->llv6)
       return;
 
     struct babel_iface_config *ic = (void *) iface_patt_find(&cf->iface_list, iface, NULL);
-
-    if (ic && iface_is_valid(p, iface))
+    if (ic)
       babel_add_iface(p, iface, ic);
 
     return;
@@ -1875,11 +1698,6 @@ babel_reconfigure_iface(struct babel_proto *p, struct babel_iface *ifa, struct b
   ifa->next_hop_ip4 = ipa_nonzero(new->next_hop_ip4) ? new->next_hop_ip4 : addr4;
   ifa->next_hop_ip6 = ipa_nonzero(new->next_hop_ip6) ? new->next_hop_ip6 : ifa->addr;
 
-  babel_iface_update_buffers(ifa);
-
-  if ((new->auth_type != BABEL_AUTH_NONE) && (new->auth_type != old->auth_type))
-    babel_auth_reset_index(ifa);
-
   if (ipa_zero(ifa->next_hop_ip4) && p->ip4_channel)
     log(L_WARN "%s: Missing IPv4 next hop address for %s", p->p.name, ifa->ifname);
 
@@ -1889,6 +1707,9 @@ babel_reconfigure_iface(struct babel_proto *p, struct babel_iface *ifa, struct b
   if (ifa->next_regular > (current_time() + new->update_interval))
     ifa->next_regular = current_time() + (random() % new->update_interval);
 
+  if ((new->tx_length != old->tx_length) || (new->rx_buffer != old->rx_buffer))
+    babel_iface_update_buffers(ifa);
+
   if (new->check_link != old->check_link)
     babel_iface_update_state(ifa);
 
@@ -1908,6 +1729,10 @@ babel_reconfigure_ifaces(struct babel_proto *p, struct babel_config *cf)
     if (!(iface->flags & IF_UP))
       continue;
 
+    /* Ignore non-multicast ifaces */
+    if (!(iface->flags & IF_MULTICAST))
+      continue;
+
     /* Ignore ifaces without link-local address */
     if (!iface->llv6)
       continue;
@@ -1915,9 +1740,6 @@ babel_reconfigure_ifaces(struct babel_proto *p, struct babel_config *cf)
     struct babel_iface *ifa = babel_find_iface(p, iface);
     struct babel_iface_config *ic = (void *) iface_patt_find(&cf->iface_list, iface, NULL);
 
-    if (ic && !iface_is_valid(p, iface))
-      ic = NULL;
-
     if (ifa && ic)
     {
       if (babel_reconfigure_iface(p, ifa, ic))
@@ -2026,23 +1848,14 @@ babel_dump(struct proto *P)
 static void
 babel_get_route_info(rte *rte, byte *buf)
 {
-  u64 rid = 0;
-  eattr *e = ea_find(rte->attrs->eattrs, EA_BABEL_ROUTER_ID);
-  if (e)
-    memcpy(&rid, e->u.ptr->data, sizeof(u64));
-
-  buf += bsprintf(buf, " (%d/%d) [%lR]", rte->attrs->pref,
-      ea_get_int(rte->attrs->eattrs, EA_BABEL_METRIC, BABEL_INFINITY), rid);
+  buf += bsprintf(buf, " (%d/%d) [%lR]", rte->pref, rte->u.babel.metric, rte->u.babel.router_id);
 }
 
 static int
-babel_get_attr(const eattr *a, byte *buf, int buflen UNUSED)
+babel_get_attr(eattr *a, byte *buf, int buflen UNUSED)
 {
   switch (a->id)
   {
-  case EA_BABEL_SEQNO:
-    return GA_FULL;
-
   case EA_BABEL_METRIC:
     bsprintf(buf, "metric: %d", a->u.data);
     return GA_FULL;
@@ -2061,7 +1874,7 @@ babel_get_attr(const eattr *a, byte *buf, int buflen UNUSED)
 }
 
 void
-babel_show_interfaces(struct proto *P, const char *iff)
+babel_show_interfaces(struct proto *P, char *iff)
 {
   struct babel_proto *p = (void *) P;
   struct babel_iface *ifa = NULL;
@@ -2070,12 +1883,13 @@ babel_show_interfaces(struct proto *P, const char *iff)
   if (p->p.proto_state != PS_UP)
   {
     cli_msg(-1023, "%s: is not up", p->p.name);
+    cli_msg(0, "");
     return;
   }
 
   cli_msg(-1023, "%s:", p->p.name);
-  cli_msg(-1023, "%-10s %-6s %-5s %7s %6s %7s %-15s %s",
-	  "Interface", "State", "Auth", "RX cost", "Nbrs", "Timer",
+  cli_msg(-1023, "%-10s %-6s %7s %6s %7s %-15s %s",
+	  "Interface", "State", "RX cost", "Nbrs", "Timer",
 	  "Next hop (v4)", "Next hop (v6)");
 
   WALK_LIST(ifa, p->interfaces)
@@ -2088,17 +1902,17 @@ babel_show_interfaces(struct proto *P, const char *iff)
 	nbrs++;
 
     btime timer = MIN(ifa->next_regular, ifa->next_hello) - current_time();
-    cli_msg(-1023, "%-10s %-6s %-5s %7u %6u %7t %-15I %I",
+    cli_msg(-1023, "%-10s %-6s %7u %6u %7t %-15I %I",
 	    ifa->iface->name, (ifa->up ? "Up" : "Down"),
-            (ifa->cf->auth_type == BABEL_AUTH_MAC ?
-             (ifa->cf->auth_permissive ? "Perm" : "Yes") : "No"),
 	    ifa->cf->rxcost, nbrs, MAX(timer, 0),
 	    ifa->next_hop_ip4, ifa->next_hop_ip6);
   }
+
+  cli_msg(0, "");
 }
 
 void
-babel_show_neighbors(struct proto *P, const char *iff)
+babel_show_neighbors(struct proto *P, char *iff)
 {
   struct babel_proto *p = (void *) P;
   struct babel_iface *ifa = NULL;
@@ -2108,12 +1922,13 @@ babel_show_neighbors(struct proto *P, const char *iff)
   if (p->p.proto_state != PS_UP)
   {
     cli_msg(-1024, "%s: is not up", p->p.name);
+    cli_msg(0, "");
     return;
   }
 
   cli_msg(-1024, "%s:", p->p.name);
-  cli_msg(-1024, "%-25s %-10s %6s %6s %6s %7s %4s",
-	  "IP address", "Interface", "Metric", "Routes", "Hellos", "Expires", "Auth");
+  cli_msg(-1024, "%-25s %-10s %6s %6s %6s %7s",
+	  "IP address", "Interface", "Metric", "Routes", "Hellos", "Expires");
 
   WALK_LIST(ifa, p->interfaces)
   {
@@ -2127,12 +1942,13 @@ babel_show_neighbors(struct proto *P, const char *iff)
         rts++;
 
       uint hellos = u32_popcount(n->hello_map);
-      btime timer = (n->hello_expiry ?: n->init_expiry) - current_time();
-      cli_msg(-1024, "%-25I %-10s %6u %6u %6u %7t %-4s",
-	      n->addr, ifa->iface->name, n->cost, rts, hellos, MAX(timer, 0),
-              n->auth_passed ? "Yes" : "No");
+      btime timer = n->hello_expiry - current_time();
+      cli_msg(-1024, "%-25I %-10s %6u %6u %6u %7t",
+	      n->addr, ifa->iface->name, n->cost, rts, hellos, MAX(timer, 0));
     }
   }
+
+  cli_msg(0, "");
 }
 
 static void
@@ -2174,6 +1990,7 @@ babel_show_entries(struct proto *P)
   if (p->p.proto_state != PS_UP)
   {
     cli_msg(-1025, "%s: is not up", p->p.name);
+    cli_msg(0, "");
     return;
   }
 
@@ -2183,6 +2000,8 @@ babel_show_entries(struct proto *P)
 
   babel_show_entries_(p, &p->ip4_rtable);
   babel_show_entries_(p, &p->ip6_rtable);
+
+  cli_msg(0, "");
 }
 
 static void
@@ -2214,6 +2033,7 @@ babel_show_routes(struct proto *P)
   if (p->p.proto_state != PS_UP)
   {
     cli_msg(-1025, "%s: is not up", p->p.name);
+    cli_msg(0, "");
     return;
   }
 
@@ -2223,6 +2043,8 @@ babel_show_routes(struct proto *P)
 
   babel_show_routes_(p, &p->ip4_rtable);
   babel_show_routes_(p, &p->ip6_rtable);
+
+  cli_msg(0, "");
 }
 
 
@@ -2255,17 +2077,56 @@ babel_kick_timer(struct babel_proto *p)
 }
 
 
-static int
-babel_preexport(struct channel *C, struct rte *new)
+static struct ea_list *
+babel_prepare_attrs(struct linpool *pool, ea_list *next, uint metric, u64 router_id)
 {
-  struct rta *a = new->attrs;
+  struct ea_list *l = lp_alloc(pool, sizeof(struct ea_list) + 2*sizeof(eattr));
+  struct adata *rid = lp_alloc(pool, sizeof(struct adata) + sizeof(u64));
+  rid->length = sizeof(u64);
+  memcpy(&rid->data, &router_id, sizeof(u64));
+
+  l->next = next;
+  l->flags = EALF_SORTED;
+  l->count = 2;
+
+  l->attrs[0].id = EA_BABEL_METRIC;
+  l->attrs[0].flags = 0;
+  l->attrs[0].type = EAF_TYPE_INT | EAF_TEMP;
+  l->attrs[0].u.data = metric;
+
+  l->attrs[1].id = EA_BABEL_ROUTER_ID;
+  l->attrs[1].flags = 0;
+  l->attrs[1].type = EAF_TYPE_OPAQUE | EAF_TEMP;
+  l->attrs[1].u.ptr = rid;
+
+  return l;
+}
+
+
+static int
+babel_preexport(struct proto *P, struct rte **new, struct linpool *pool UNUSED)
+{
+  struct rta *a = (*new)->attrs;
+
   /* Reject our own unreachable routes */
-  if ((a->dest == RTD_UNREACHABLE) && (new->src->proto == C->proto))
+  if ((a->dest == RTD_UNREACHABLE) && (a->src->proto == P))
     return -1;
 
   return 0;
 }
 
+static struct ea_list *
+babel_make_tmp_attrs(struct rte *rt, struct linpool *pool)
+{
+  return babel_prepare_attrs(pool, NULL, rt->u.babel.metric, rt->u.babel.router_id);
+}
+
+static void
+babel_store_tmp_attrs(struct rte *rt)
+{
+  rt->u.babel.metric = ea_get_int(rt->attrs->eattrs, EA_BABEL_METRIC, 0);
+}
+
 /*
  * babel_rt_notify - core tells us about new route (possibly our own),
  * so store it into our data structures.
@@ -2280,22 +2141,10 @@ babel_rt_notify(struct proto *P, struct channel *c UNUSED, struct network *net,
   if (new)
   {
     /* Update */
-    uint rt_seqno;
+    uint internal = (new->attrs->src->proto == P);
+    uint rt_seqno = internal ? new->u.babel.seqno : p->update_seqno;
     uint rt_metric = ea_get_int(new->attrs->eattrs, EA_BABEL_METRIC, 0);
-    u64 rt_router_id = 0;
-
-    if (new->src->proto == P)
-    {
-      rt_seqno = ea_find(new->attrs->eattrs, EA_BABEL_SEQNO)->u.data;
-      eattr *e = ea_find(new->attrs->eattrs, EA_BABEL_ROUTER_ID);
-      if (e)
-	memcpy(&rt_router_id, e->u.ptr->data, sizeof(u64));
-    }
-    else
-    {
-      rt_seqno = p->update_seqno;
-      rt_router_id = p->router_id;
-    }
+    u64 rt_router_id = internal ? new->u.babel.router_id : p->router_id;
 
     if (rt_metric > BABEL_INFINITY)
     {
@@ -2338,16 +2187,15 @@ babel_rt_notify(struct proto *P, struct channel *c UNUSED, struct network *net,
 static int
 babel_rte_better(struct rte *new, struct rte *old)
 {
-  uint new_metric = ea_get_int(new->attrs->eattrs, EA_BABEL_METRIC, BABEL_INFINITY);
-  uint old_metric = ea_get_int(old->attrs->eattrs, EA_BABEL_METRIC, BABEL_INFINITY);
-
-  return new_metric < old_metric;
+  return new->u.babel.metric < old->u.babel.metric;
 }
 
-static u32
-babel_rte_igp_metric(struct rte *rt)
+static int
+babel_rte_same(struct rte *new, struct rte *old)
 {
-  return ea_get_int(rt->attrs->eattrs, EA_BABEL_METRIC, BABEL_INFINITY);
+  return ((new->u.babel.seqno == old->u.babel.seqno) &&
+	  (new->u.babel.metric == old->u.babel.metric) &&
+	  (new->u.babel.router_id == old->u.babel.router_id));
 }
 
 
@@ -2381,8 +2229,10 @@ babel_init(struct proto_config *CF)
   P->if_notify = babel_if_notify;
   P->rt_notify = babel_rt_notify;
   P->preexport = babel_preexport;
+  P->make_tmp_attrs = babel_make_tmp_attrs;
+  P->store_tmp_attrs = babel_store_tmp_attrs;
   P->rte_better = babel_rte_better;
-  P->rte_igp_metric = babel_rte_igp_metric;
+  P->rte_same = babel_rte_same;
 
   return P;
 }
@@ -2431,11 +2281,6 @@ babel_iface_shutdown(struct babel_iface *ifa)
 {
   if (ifa->sk)
   {
-    /*
-     * Retract all our routes and lower the hello interval so peers' neighbour
-     * state expires quickly
-     */
-    babel_send_hello(ifa, BABEL_MIN_INTERVAL);
     babel_send_wildcard_retraction(ifa);
     babel_send_queue(ifa);
   }
@@ -2498,9 +2343,3 @@ struct protocol proto_babel = {
   .get_route_info =	babel_get_route_info,
   .get_attr =		babel_get_attr
 };
-
-void
-babel_build(void)
-{
-  proto_build(&proto_babel);
-}
