@@ -92,7 +92,6 @@
  * RFC 6286 - AS-Wide Unique BGP Identifier
  * RFC 6608 - Subcodes for BGP Finite State Machine Error
  * RFC 6793 - BGP Support for 4-Octet AS Numbers
- * RFC 7311 - Accumulated IGP Metric Attribute for BGP
  * RFC 7313 - Enhanced Route Refresh Capability for BGP
  * RFC 7606 - Revised Error Handling for BGP UPDATE Messages
  * RFC 7911 - Advertisement of Multiple Paths in BGP
@@ -100,12 +99,8 @@
  * RFC 8092 - BGP Large Communities Attribute
  * RFC 8203 - BGP Administrative Shutdown Communication
  * RFC 8212 - Default EBGP Route Propagation Behavior without Policies
- * RFC 8654 - Extended Message Support for BGP
- * RFC 9117 - Revised Validation Procedure for BGP Flow Specifications
- * RFC 9234 - Route Leak Prevention and Detection Using Roles
- * draft-ietf-idr-ext-opt-param-07
+ * draft-ietf-idr-bgp-extended-messages-27
  * draft-uttaro-idr-bgp-persistence-04
- * draft-walton-bgp-hostname-capability-02
  */
 
 #undef LOCAL_DEBUG
@@ -127,15 +122,14 @@
 #include "bgp.h"
 
 
-static list STATIC_LIST_INIT(bgp_sockets);		/* Global list of listening sockets */
+struct linpool *bgp_linpool;		/* Global temporary pool */
+struct linpool *bgp_linpool2;		/* Global temporary pool for bgp_rt_notify() */
+static list bgp_sockets;		/* Global list of listening sockets */
 
 
 static void bgp_connect(struct bgp_proto *p);
 static void bgp_active(struct bgp_proto *p);
-static void bgp_setup_conn(struct bgp_proto *p, struct bgp_conn *conn);
-static void bgp_setup_sk(struct bgp_conn *conn, sock *s);
-static void bgp_send_open(struct bgp_conn *conn);
-static void bgp_update_bfd(struct bgp_proto *p, const struct bfd_options *bfd);
+static void bgp_update_bfd(struct bgp_proto *p, int use_bfd);
 
 static int bgp_incoming_connection(sock *sk, uint dummy UNUSED);
 static void bgp_listen_sock_err(sock *sk UNUSED, int err);
@@ -155,19 +149,18 @@ bgp_open(struct bgp_proto *p)
   struct bgp_socket *bs = NULL;
   struct iface *ifa = p->cf->strict_bind ? p->cf->iface : NULL;
   ip_addr addr = p->cf->strict_bind ? p->cf->local_ip :
-    (p->ipv4 ? IPA_NONE4 : IPA_NONE6);
+    (ipa_is_ip4(p->cf->remote_ip) ? IPA_NONE4 : IPA_NONE6);
   uint port = p->cf->local_port;
-  uint flags = p->cf->free_bind ? SKF_FREEBIND : 0;
-  uint flag_mask = SKF_FREEBIND;
+
+  /* FIXME: Add some global init? */
+  if (!bgp_linpool)
+    init_list(&bgp_sockets);
 
   /* We assume that cf->iface is defined iff cf->local_ip is link-local */
 
   WALK_LIST(bs, bgp_sockets)
-    if (ipa_equal(bs->sk->saddr, addr) &&
-	(bs->sk->sport == port) &&
-	(bs->sk->iface == ifa) &&
-	(bs->sk->vrf == p->p.vrf) &&
-	((bs->sk->flags & flag_mask) == flags))
+    if (ipa_equal(bs->sk->saddr, addr) && (bs->sk->sport == port) &&
+	(bs->sk->iface == ifa) && (bs->sk->vrf == p->p.vrf))
     {
       bs->uc++;
       p->sock = bs;
@@ -181,7 +174,7 @@ bgp_open(struct bgp_proto *p)
   sk->sport = port;
   sk->iface = ifa;
   sk->vrf = p->p.vrf;
-  sk->flags = flags;
+  sk->flags = 0;
   sk->tos = IP_PREC_INTERNET_CONTROL;
   sk->rbsize = BGP_RX_BUFFER_SIZE;
   sk->tbsize = BGP_TX_BUFFER_SIZE;
@@ -199,6 +192,12 @@ bgp_open(struct bgp_proto *p)
 
   add_tail(&bgp_sockets, &bs->n);
 
+  if (!bgp_linpool)
+  {
+    bgp_linpool  = lp_new_default(proto_pool);
+    bgp_linpool2 = lp_new_default(proto_pool);
+  }
+
   return 0;
 
 err:
@@ -227,6 +226,15 @@ bgp_close(struct bgp_proto *p)
   rfree(bs->sk);
   rem_node(&bs->n);
   mb_free(bs);
+
+  if (!EMPTY_LIST(bgp_sockets))
+    return;
+
+  rfree(bgp_linpool);
+  bgp_linpool = NULL;
+
+  rfree(bgp_linpool2);
+  bgp_linpool2 = NULL;
 }
 
 static inline int
@@ -234,17 +242,8 @@ bgp_setup_auth(struct bgp_proto *p, int enable)
 {
   if (p->cf->password)
   {
-    ip_addr prefix = p->cf->remote_ip;
-    int pxlen = -1;
-
-    if (p->cf->remote_range)
-    {
-      prefix = net_prefix(p->cf->remote_range);
-      pxlen = net_pxlen(p->cf->remote_range);
-    }
-
     int rv = sk_set_md5_auth(p->sock->sk,
-			     p->cf->local_ip, prefix, pxlen, p->cf->iface,
+			     p->cf->local_ip, p->cf->remote_ip, p->cf->iface,
 			     enable ? p->cf->password : NULL, p->cf->setkey);
 
     if (rv < 0)
@@ -260,7 +259,7 @@ static inline struct bgp_channel *
 bgp_find_channel(struct bgp_proto *p, u32 afi)
 {
   struct bgp_channel *c;
-  BGP_WALK_CHANNELS(p, c)
+  WALK_LIST(c, p->p.channels)
     if (c->afi == afi)
       return c;
 
@@ -273,17 +272,8 @@ bgp_startup(struct bgp_proto *p)
   BGP_TRACE(D_EVENTS, "Started");
   p->start_state = BSS_CONNECT;
 
-  if (!p->passive)
+  if (!p->cf->passive)
     bgp_active(p);
-
-  if (p->postponed_sk)
-  {
-    /* Apply postponed incoming connection */
-    bgp_setup_conn(p, &p->incoming_conn);
-    bgp_setup_sk(&p->incoming_conn, p->postponed_sk);
-    bgp_send_open(&p->incoming_conn);
-    p->postponed_sk = NULL;
-  }
 }
 
 static void
@@ -323,8 +313,6 @@ err2:
 err1:
   p->p.disabled = 1;
   bgp_store_error(p, NULL, BE_MISC, err_val);
-
-  p->neigh = NULL;
   proto_notify_state(&p->p, PS_DOWN);
 
   return;
@@ -399,7 +387,7 @@ bgp_close_conn(struct bgp_conn *conn)
 void
 bgp_update_startup_delay(struct bgp_proto *p)
 {
-  const struct bgp_config *cf = p->cf;
+  struct bgp_config *cf = p->cf;
 
   DBG("BGP: Updating startup delay\n");
 
@@ -422,7 +410,7 @@ bgp_update_startup_delay(struct bgp_proto *p)
 }
 
 static void
-bgp_graceful_close_conn(struct bgp_conn *conn, int subcode, byte *data, uint len)
+bgp_graceful_close_conn(struct bgp_conn *conn, uint subcode, byte *data, uint len)
 {
   switch (conn->state)
   {
@@ -438,13 +426,7 @@ bgp_graceful_close_conn(struct bgp_conn *conn, int subcode, byte *data, uint len
   case BS_OPENSENT:
   case BS_OPENCONFIRM:
   case BS_ESTABLISHED:
-    if (subcode < 0)
-    {
-      bgp_conn_enter_close_state(conn);
-      bgp_schedule_packet(conn, NULL, PKT_SCHEDULE_CLOSE);
-    }
-    else
-      bgp_error(conn, 6, subcode, data, len);
+    bgp_error(conn, 6, subcode, data, len);
     return;
 
   default:
@@ -461,8 +443,6 @@ bgp_down(struct bgp_proto *p)
     bgp_close(p);
   }
 
-  p->neigh = NULL;
-
   BGP_TRACE(D_EVENTS, "Down");
   proto_notify_state(&p->p, PS_DOWN);
 }
@@ -476,7 +456,7 @@ bgp_decision(void *vp)
   if ((p->p.proto_state == PS_START) &&
       (p->outgoing_conn.state == BS_IDLE) &&
       (p->incoming_conn.state != BS_OPENCONFIRM) &&
-      !p->passive)
+      !p->cf->passive)
     bgp_active(p);
 
   if ((p->p.proto_state == PS_STOP) &&
@@ -485,31 +465,8 @@ bgp_decision(void *vp)
     bgp_down(p);
 }
 
-static struct bgp_proto *
-bgp_spawn(struct bgp_proto *pp, ip_addr remote_ip)
-{
-  struct symbol *sym;
-  char fmt[SYM_MAX_LEN];
-
-  bsprintf(fmt, "%s%%0%dd", pp->cf->dynamic_name, pp->cf->dynamic_name_digits);
-
-  /* This is hack, we would like to share config, but we need to copy it now */
-  new_config = config;
-  cfg_mem = config->mem;
-  conf_this_scope = config->root_scope;
-  sym = cf_default_name(fmt, &(pp->dynamic_name_counter));
-  proto_clone_config(sym, pp->p.cf);
-  new_config = NULL;
-  cfg_mem = NULL;
-
-  /* Just pass remote_ip to bgp_init() */
-  ((struct bgp_config *) sym->proto)->remote_ip = remote_ip;
-
-  return (void *) proto_spawn(sym->proto, 0);
-}
-
 void
-bgp_stop(struct bgp_proto *p, int subcode, byte *data, uint len)
+bgp_stop(struct bgp_proto *p, uint subcode, byte *data, uint len)
 {
   proto_notify_state(&p->p, PS_STOP);
   bgp_graceful_close_conn(&p->outgoing_conn, subcode, data, len);
@@ -534,7 +491,6 @@ bgp_conn_enter_openconfirm_state(struct bgp_conn *conn)
 }
 
 static const struct bgp_af_caps dummy_af_caps = { };
-static const struct bgp_af_caps basic_af_caps = { .ready = 1 };
 
 void
 bgp_conn_enter_established_state(struct bgp_conn *conn)
@@ -545,16 +501,10 @@ bgp_conn_enter_established_state(struct bgp_conn *conn)
   struct bgp_channel *c;
 
   BGP_TRACE(D_EVENTS, "BGP session established");
-  p->last_established = current_time();
-  p->stats.fsm_established_transitions++;
 
   /* For multi-hop BGP sessions */
-  if (ipa_zero(p->local_ip))
-    p->local_ip = conn->sk->saddr;
-
-  /* For promiscuous sessions */
-  if (!p->remote_as)
-    p->remote_as = conn->received_as;
+  if (ipa_zero(p->source_addr))
+    p->source_addr = conn->sk->saddr;
 
   /* In case of LLv6 is not valid during BGP start */
   if (ipa_zero(p->link_addr) && p->neigh && p->neigh->iface && p->neigh->iface->llv6)
@@ -586,18 +536,11 @@ bgp_conn_enter_established_state(struct bgp_conn *conn)
   /* Summary state of ADD_PATH RX for active channels */
   uint summary_add_path_rx = 0;
 
-  BGP_WALK_CHANNELS(p, c)
+  WALK_LIST(c, p->p.channels)
   {
     const struct bgp_af_caps *loc = bgp_find_af_caps(local, c->afi);
     const struct bgp_af_caps *rem = bgp_find_af_caps(peer,  c->afi);
 
-    /* Use default if capabilities were not announced */
-    if (!local->length && (c->afi == BGP_AF_IPV4))
-      loc = &basic_af_caps;
-
-    if (!peer->length && (c->afi == BGP_AF_IPV4))
-      rem = &basic_af_caps;
-
     /* Ignore AFIs that were not announced in multiprotocol capability */
     if (!loc || !loc->ready)
       loc = &dummy_af_caps;
@@ -668,7 +611,7 @@ bgp_conn_enter_established_state(struct bgp_conn *conn)
   p->channel_count = num;
   p->summary_add_path_rx = summary_add_path_rx;
 
-  BGP_WALK_CHANNELS(p, c)
+  WALK_LIST(c, p->p.channels)
   {
     if (c->c.disabled)
       continue;
@@ -687,7 +630,6 @@ static void
 bgp_conn_leave_established_state(struct bgp_proto *p)
 {
   BGP_TRACE(D_EVENTS, "BGP session closed");
-  p->last_established = current_time();
   p->conn = NULL;
 
   if (p->p.proto_state == PS_UP)
@@ -747,7 +689,7 @@ bgp_handle_graceful_restart(struct bgp_proto *p)
   p->gr_active_num = 0;
 
   struct bgp_channel *c;
-  BGP_WALK_CHANNELS(p, c)
+  WALK_LIST(c, p->p.channels)
   {
     /* FIXME: perhaps check for channel state instead of disabled flag? */
     if (c->c.disabled)
@@ -842,7 +784,7 @@ bgp_graceful_restart_timeout(timer *t)
   if (p->llgr_ready)
   {
     struct bgp_channel *c;
-    BGP_WALK_CHANNELS(p, c)
+    WALK_LIST(c, p->p.channels)
     {
       /* Channel is not in GR and is already flushed */
       if (!c->gr_active)
@@ -938,7 +880,6 @@ bgp_send_open(struct bgp_conn *conn)
   conn->sk->rx_hook = bgp_rx;
   conn->sk->tx_hook = bgp_tx;
   tm_stop(conn->connect_timer);
-  bgp_prepare_capabilities(conn);
   bgp_schedule_packet(conn, NULL, PKT_OPEN);
   bgp_conn_set_state(conn, BS_OPENSENT);
   bgp_start_timer(conn->hold_timer, conn->bgp->cf->initial_hold_time);
@@ -1098,8 +1039,8 @@ bgp_connect(struct bgp_proto *p)	/* Enter Connect state and start establishing c
   DBG("BGP: Connecting\n");
   sock *s = sk_new(p->p.pool);
   s->type = SK_TCP_ACTIVE;
-  s->saddr = p->local_ip;
-  s->daddr = p->remote_ip;
+  s->saddr = p->source_addr;
+  s->daddr = p->cf->remote_ip;
   s->dport = p->cf->remote_port;
   s->iface = p->neigh ? p->neigh->iface : NULL;
   s->vrf = p->p.vrf;
@@ -1134,9 +1075,6 @@ err:
   return;
 }
 
-static inline int bgp_is_dynamic(struct bgp_proto *p)
-{ return ipa_zero(p->remote_ip); }
-
 /**
  * bgp_find_proto - find existing proto for incoming connection
  * @sk: TCP socket
@@ -1145,7 +1083,6 @@ static inline int bgp_is_dynamic(struct bgp_proto *p)
 static struct bgp_proto *
 bgp_find_proto(sock *sk)
 {
-  struct bgp_proto *best = NULL;
   struct bgp_proto *p;
 
   /* sk->iface is valid only if src or dst address is link-local */
@@ -1153,20 +1090,13 @@ bgp_find_proto(sock *sk)
 
   WALK_LIST(p, proto_list)
     if ((p->p.proto == &proto_bgp) &&
-	(ipa_equal(p->remote_ip, sk->daddr) || bgp_is_dynamic(p)) &&
-	(!p->cf->remote_range || ipa_in_netX(sk->daddr, p->cf->remote_range)) &&
-	(p->p.vrf == sk->vrf) &&
-	(p->cf->local_port == sk->sport) &&
+	(p->sock == sk->data) &&
+	ipa_equal(p->cf->remote_ip, sk->daddr) &&
 	(!link || (p->cf->iface == sk->iface)) &&
 	(ipa_zero(p->cf->local_ip) || ipa_equal(p->cf->local_ip, sk->saddr)))
-    {
-      best = p;
+      return p;
 
-      if (!bgp_is_dynamic(p))
-	break;
-    }
-
-  return best;
+  return NULL;
 }
 
 /**
@@ -1245,16 +1175,6 @@ bgp_incoming_connection(sock *sk, uint dummy UNUSED)
     sk_reallocate(sk);
   }
 
-  /* For dynamic BGP, spawn new instance and postpone the socket */
-  if (bgp_is_dynamic(p))
-  {
-    p = bgp_spawn(p, sk->daddr);
-    p->postponed_sk = sk;
-    rmove(sk, p->p.pool);
-    return 0;
-  }
-
-  rmove(sk, p->p.pool);
   bgp_setup_conn(p, &p->incoming_conn);
   bgp_setup_sk(&p->incoming_conn, sk);
   bgp_send_open(&p->incoming_conn);
@@ -1281,11 +1201,11 @@ bgp_start_neighbor(struct bgp_proto *p)
 {
   /* Called only for single-hop BGP sessions */
 
-  if (ipa_zero(p->local_ip))
-    p->local_ip = p->neigh->ifa->ip;
+  if (ipa_zero(p->source_addr))
+    p->source_addr = p->neigh->ifa->ip;
 
-  if (ipa_is_link_local(p->local_ip))
-    p->link_addr = p->local_ip;
+  if (ipa_is_link_local(p->source_addr))
+    p->link_addr = p->source_addr;
   else if (p->neigh->iface->llv6)
     p->link_addr = p->neigh->iface->llv6->ip;
 
@@ -1348,7 +1268,7 @@ bgp_bfd_notify(struct bfd_request *req)
     BGP_TRACE(D_EVENTS, "BFD session down");
     bgp_store_error(p, NULL, BE_MISC, BEM_BFD_DOWN);
 
-    if (req->opts.mode == BGP_BFD_GRACEFUL)
+    if (p->cf->bfd == BGP_BFD_GRACEFUL)
     {
       /* Trigger graceful restart */
       if (p->conn && (p->conn->state == BS_ESTABLISHED) && p->gr_ready)
@@ -1371,17 +1291,14 @@ bgp_bfd_notify(struct bfd_request *req)
 }
 
 static void
-bgp_update_bfd(struct bgp_proto *p, const struct bfd_options *bfd)
+bgp_update_bfd(struct bgp_proto *p, int use_bfd)
 {
-  if (bfd && p->bfd_req)
-    bfd_update_request(p->bfd_req, bfd);
-
-  if (bfd && !p->bfd_req && !bgp_is_dynamic(p))
-    p->bfd_req = bfd_request_session(p->p.pool, p->remote_ip, p->local_ip,
+  if (use_bfd && !p->bfd_req)
+    p->bfd_req = bfd_request_session(p->p.pool, p->cf->remote_ip, p->source_addr,
 				     p->cf->multihop ? NULL : p->neigh->iface,
-				     p->p.vrf, bgp_bfd_notify, p, bfd);
+				     bgp_bfd_notify, p);
 
-  if (!bfd && p->bfd_req)
+  if (!use_bfd && p->bfd_req)
   {
     rfree(p->bfd_req);
     p->bfd_req = NULL;
@@ -1394,10 +1311,6 @@ bgp_reload_routes(struct channel *C)
   struct bgp_proto *p = (void *) C->proto;
   struct bgp_channel *c = (void *) C;
 
-  /* Ignore non-BGP channels */
-  if (C->channel != &channel_bgp)
-    return;
-
   ASSERT(p->conn && (p->route_refresh || c->c.in_table));
 
   if (c->c.in_table)
@@ -1412,10 +1325,6 @@ bgp_feed_begin(struct channel *C, int initial)
   struct bgp_proto *p = (void *) C->proto;
   struct bgp_channel *c = (void *) C;
 
-  /* Ignore non-BGP channels */
-  if (C->channel != &channel_bgp)
-    return;
-
   /* This should not happen */
   if (!p->conn)
     return;
@@ -1441,10 +1350,6 @@ bgp_feed_end(struct channel *C)
   struct bgp_proto *p = (void *) C->proto;
   struct bgp_channel *c = (void *) C;
 
-  /* Ignore non-BGP channels */
-  if (C->channel != &channel_bgp)
-    return;
-
   /* This should not happen */
   if (!p->conn)
     return;
@@ -1470,7 +1375,7 @@ static void
 bgp_start_locked(struct object_lock *lock)
 {
   struct bgp_proto *p = lock->data;
-  const struct bgp_config *cf = p->cf;
+  struct bgp_config *cf = p->cf;
 
   if (p->p.proto_state != PS_START)
   {
@@ -1480,17 +1385,17 @@ bgp_start_locked(struct object_lock *lock)
 
   DBG("BGP: Got lock\n");
 
-  if (cf->multihop || bgp_is_dynamic(p))
+  if (cf->multihop)
   {
     /* Multi-hop sessions do not use neighbor entries */
     bgp_initiate(p);
     return;
   }
 
-  neighbor *n = neigh_find(&p->p, p->remote_ip, cf->iface, NEF_STICKY);
+  neighbor *n = neigh_find(&p->p, cf->remote_ip, cf->iface, NEF_STICKY);
   if (!n)
   {
-    log(L_ERR "%s: Invalid remote address %I%J", p->p.name, p->remote_ip, cf->iface);
+    log(L_ERR "%s: Invalid remote address %I%J", p->p.name, cf->remote_ip, cf->iface);
     /* As we do not start yet, we can just disable protocol */
     p->p.disabled = 1;
     bgp_store_error(p, NULL, BE_MISC, BEM_INVALID_NEXT_HOP);
@@ -1501,7 +1406,7 @@ bgp_start_locked(struct object_lock *lock)
   p->neigh = n;
 
   if (n->scope <= 0)
-    BGP_TRACE(D_EVENTS, "Waiting for %I%J to become my neighbor", p->remote_ip, cf->iface);
+    BGP_TRACE(D_EVENTS, "Waiting for %I%J to become my neighbor", cf->remote_ip, cf->iface);
   else if (p->cf->check_link && !(n->iface->flags & IF_LINK_UP))
     BGP_TRACE(D_EVENTS, "Waiting for link on %s", n->iface->name);
   else
@@ -1512,38 +1417,17 @@ static int
 bgp_start(struct proto *P)
 {
   struct bgp_proto *p = (struct bgp_proto *) P;
-  const struct bgp_config *cf = p->cf;
-
-  p->local_ip = cf->local_ip;
-  p->local_as = cf->local_as;
-  p->remote_as = cf->remote_as;
-  p->public_as = cf->local_as;
-
-  /* For dynamic BGP childs, remote_ip is already set */
-  if (ipa_nonzero(cf->remote_ip))
-    p->remote_ip = cf->remote_ip;
-
-  /* Confederation ID is used for truly external peers */
-  if (p->cf->confederation && !p->is_interior)
-    p->public_as = cf->confederation;
-
-  p->passive = cf->passive || bgp_is_dynamic(p);
+  struct object_lock *lock;
 
+  DBG("BGP: Startup.\n");
   p->start_state = BSS_PREPARE;
   p->outgoing_conn.state = BS_IDLE;
   p->incoming_conn.state = BS_IDLE;
   p->neigh = NULL;
   p->bfd_req = NULL;
-  p->postponed_sk = NULL;
   p->gr_ready = 0;
   p->gr_active_num = 0;
 
-  /* Reset some stats */
-  p->stats.rx_messages = p->stats.tx_messages = 0;
-  p->stats.rx_updates = p->stats.tx_updates = 0;
-  p->stats.rx_bytes = p->stats.tx_bytes = 0;
-  p->last_rx_update = 0;
-
   p->event = ev_new_init(p->p.pool, bgp_decision, p);
   p->startup_timer = tm_new_init(p->p.pool, bgp_startup_timeout, p, 0, 0);
   p->gr_timer = tm_new_init(p->p.pool, bgp_graceful_restart_timeout, p, 0, 0);
@@ -1553,13 +1437,14 @@ bgp_start(struct proto *P)
     p->rr_cluster_id = p->cf->rr_cluster_id ? p->cf->rr_cluster_id : p->local_id;
 
   p->remote_id = 0;
+  p->source_addr = p->cf->local_ip;
   p->link_addr = IPA_NONE;
 
   /* Lock all channels when in GR recovery mode */
   if (p->p.gr_recovery && p->cf->gr_mode)
   {
     struct bgp_channel *c;
-    BGP_WALK_CHANNELS(p, c)
+    WALK_LIST(c, p->p.channels)
       channel_graceful_restart_lock(&c->c);
   }
 
@@ -1567,23 +1452,15 @@ bgp_start(struct proto *P)
    * Before attempting to create the connection, we need to lock the port,
    * so that we are the only instance attempting to talk with that neighbor.
    */
-  struct object_lock *lock;
+
   lock = p->lock = olock_new(P->pool);
-  lock->addr = p->remote_ip;
+  lock->addr = p->cf->remote_ip;
   lock->port = p->cf->remote_port;
   lock->iface = p->cf->iface;
   lock->vrf = p->cf->iface ? NULL : p->p.vrf;
   lock->type = OBJLOCK_TCP;
   lock->hook = bgp_start_locked;
   lock->data = p;
-
-  /* For dynamic BGP, we use inst 1 to avoid collisions with regular BGP */
-  if (bgp_is_dynamic(p))
-  {
-    lock->addr = net_prefix(p->cf->remote_range);
-    lock->inst = 1;
-  }
-
   olock_acquire(lock);
 
   return PS_START;
@@ -1595,7 +1472,7 @@ static int
 bgp_shutdown(struct proto *P)
 {
   struct bgp_proto *p = (struct bgp_proto *) P;
-  int subcode = 0;
+  uint subcode = 0;
 
   char *message = NULL;
   byte *data = NULL;
@@ -1616,7 +1493,6 @@ bgp_shutdown(struct proto *P)
 
   case PDC_CMD_DISABLE:
   case PDC_CMD_SHUTDOWN:
-  shutdown:
     subcode = 2; // Errcode 6, 2 - administrative shutdown
     message = P->message;
     break;
@@ -1626,14 +1502,6 @@ bgp_shutdown(struct proto *P)
     message = P->message;
     break;
 
-  case PDC_CMD_GR_DOWN:
-    if ((p->cf->gr_mode != BGP_GR_ABLE) &&
-	(p->cf->llgr_mode != BGP_LLGR_ABLE))
-      goto shutdown;
-
-    subcode = -1; // Do not send NOTIFICATION, just close the connection
-    break;
-
   case PDC_RX_LIMIT_HIT:
   case PDC_IN_LIMIT_HIT:
     subcode = 1; // Errcode 6, 1 - max number of prefixes reached
@@ -1660,7 +1528,7 @@ bgp_shutdown(struct proto *P)
   if (message)
   {
     uint msg_len = strlen(message);
-    msg_len = MIN(msg_len, 255);
+    msg_len = MIN(msg_len, 128);
 
     /* Buffer will be freed automatically by protocol shutdown */
     data = mb_alloc(p->p.pool, msg_len + 1);
@@ -1692,28 +1560,23 @@ bgp_init(struct proto_config *CF)
   P->rte_mergable = bgp_rte_mergable;
   P->rte_recalculate = cf->deterministic_med ? bgp_rte_recalculate : NULL;
   P->rte_modify = bgp_rte_modify_stale;
-  P->rte_igp_metric = bgp_rte_igp_metric;
 
   p->cf = cf;
+  p->local_as = cf->local_as;
+  p->remote_as = cf->remote_as;
+  p->public_as = cf->local_as;
   p->is_internal = (cf->local_as == cf->remote_as);
   p->is_interior = p->is_internal || cf->confederation_member;
   p->rs_client = cf->rs_client;
   p->rr_client = cf->rr_client;
 
-  p->ipv4 = ipa_nonzero(cf->remote_ip) ?
-    ipa_is_ip4(cf->remote_ip) :
-    (cf->remote_range && (cf->remote_range->type == NET_IP4));
-
-  p->remote_ip = cf->remote_ip;
-  p->remote_as = cf->remote_as;
-
-  /* Hack: We use cf->remote_ip just to pass remote_ip from bgp_spawn() */
-  if (cf->c.parent)
-    cf->remote_ip = IPA_NONE;
+  /* Confederation ID is used for truly external peers */
+  if (cf->confederation && !p->is_interior)
+    p->public_as = cf->confederation;
 
   /* Add all channels */
   struct bgp_channel_config *cc;
-  BGP_CF_WALK_CHANNELS(cf, cc)
+  WALK_LIST(cc, CF->channels)
     proto_add_channel(P, &cc->c);
 
   return P;
@@ -1734,9 +1597,6 @@ bgp_channel_init(struct channel *C, struct channel_config *CF)
 
   if (cf->igp_table_ip6)
     c->igp_table_ip6 = cf->igp_table_ip6->table;
-
-  if (cf->base_table)
-    c->base_table = cf->base_table->table;
 }
 
 static int
@@ -1744,7 +1604,7 @@ bgp_channel_start(struct channel *C)
 {
   struct bgp_proto *p = (void *) C->proto;
   struct bgp_channel *c = (void *) C;
-  ip_addr src = p->local_ip;
+  ip_addr src = p->source_addr;
 
   if (c->igp_table_ip4)
     rt_lock_table(c->igp_table_ip4);
@@ -1752,12 +1612,6 @@ bgp_channel_start(struct channel *C)
   if (c->igp_table_ip6)
     rt_lock_table(c->igp_table_ip6);
 
-  if (c->base_table)
-  {
-    rt_lock_table(c->base_table);
-    rt_flowspec_link(c->base_table, c->c.table);
-  }
-
   c->pool = p->p.pool; // XXXX
   bgp_init_bucket_table(c);
   bgp_init_prefix_table(c);
@@ -1765,9 +1619,6 @@ bgp_channel_start(struct channel *C)
   if (c->cf->import_table)
     channel_setup_in_table(C);
 
-  if (c->cf->export_table)
-    channel_setup_out_table(C);
-
   c->stale_timer = tm_new_init(c->pool, bgp_long_lived_stale_timeout, c, 0, 0);
 
   c->next_hop_addr = c->cf->next_hop_addr;
@@ -1842,12 +1693,6 @@ bgp_channel_cleanup(struct channel *C)
   if (c->igp_table_ip6)
     rt_unlock_table(c->igp_table_ip6);
 
-  if (c->base_table)
-  {
-    rt_flowspec_unlink(c->base_table, c->c.table);
-    rt_unlock_table(c->base_table);
-  }
-
   c->index = 0;
 
   /* Cleanup rest of bgp_channel starting at pool field */
@@ -1859,7 +1704,7 @@ bgp_find_channel_config(struct bgp_config *cf, u32 afi)
 {
   struct bgp_channel_config *cc;
 
-  BGP_CF_WALK_CHANNELS(cf, cc)
+  WALK_LIST(cc, cf->c.channels)
     if (cc->afi == afi)
       return cc;
 
@@ -1895,43 +1740,19 @@ bgp_default_igp_table(struct bgp_config *cf, struct bgp_channel_config *cc, u32
   cf_error("Undefined IGP table");
 }
 
-static struct rtable_config *
-bgp_default_base_table(struct bgp_config *cf, struct bgp_channel_config *cc)
-{
-  /* Expected table type */
-  u32 type = (cc->afi == BGP_AF_FLOW4) ? NET_IP4 : NET_IP6;
-
-  /* First, try appropriate IP channel */
-  u32 afi2 = BGP_AF(BGP_AFI(cc->afi), BGP_SAFI_UNICAST);
-  struct bgp_channel_config *cc2 = bgp_find_channel_config(cf, afi2);
-  if (cc2 && (cc2->c.table->addr_type == type))
-    return cc2->c.table;
-
-  /* Last, try default table of given type */
-  struct rtable_config *tab = cf->c.global->def_tables[type];
-  if (tab)
-    return tab;
-
-  cf_error("Undefined base table");
-}
 
 void
 bgp_postconfig(struct proto_config *CF)
 {
   struct bgp_config *cf = (void *) CF;
+  int internal = (cf->local_as == cf->remote_as);
+  int interior = internal || cf->confederation_member;
 
   /* Do not check templates at all */
   if (cf->c.class == SYM_TEMPLATE)
     return;
 
 
-  /* Handle undefined remote_as, zero should mean unspecified external */
-  if (!cf->remote_as && (cf->peer_type == BGP_PT_INTERNAL))
-    cf->remote_as = cf->local_as;
-
-  int internal = (cf->local_as == cf->remote_as);
-  int interior = internal || cf->confederation_member;
-
   /* EBGP direct by default, IBGP multihop by default */
   if (cf->multihop < 0)
     cf->multihop = internal ? 64 : 0;
@@ -1948,20 +1769,11 @@ bgp_postconfig(struct proto_config *CF)
   if (!cf->local_as)
     cf_error("Local AS number must be set");
 
-  if (ipa_zero(cf->remote_ip) && !cf->remote_range)
+  if (ipa_zero(cf->remote_ip))
     cf_error("Neighbor must be configured");
 
-  if (ipa_zero(cf->local_ip) && cf->strict_bind)
-    cf_error("Local address must be configured for strict bind");
-
-  if (!cf->remote_as && !cf->peer_type)
-    cf_error("Remote AS number (or peer type) must be set");
-
-  if ((cf->peer_type == BGP_PT_INTERNAL) && !internal)
-    cf_error("IBGP cannot have different ASNs");
-
-  if ((cf->peer_type == BGP_PT_EXTERNAL) &&  internal)
-    cf_error("EBGP cannot have the same ASNs");
+  if (!cf->remote_as)
+    cf_error("Remote AS number must be set");
 
   if (!cf->iface && (ipa_is_link_local(cf->local_ip) ||
 		     ipa_is_link_local(cf->remote_ip)))
@@ -1976,15 +1788,6 @@ bgp_postconfig(struct proto_config *CF)
   if (internal && cf->rs_client)
     cf_error("Only external neighbor can be RS client");
 
-  if (internal && (cf->local_role != BGP_ROLE_UNDEFINED))
-    cf_error("Local role cannot be set on IBGP sessions");
-
-  if (interior && (cf->local_role != BGP_ROLE_UNDEFINED))
-    log(L_WARN "BGP roles are not recommended to be used within AS confederations");
-
-  if (cf->require_roles && (cf->local_role == BGP_ROLE_UNDEFINED))
-    cf_error("Local role must be set if roles are required");
-
   if (!cf->confederation && cf->confederation_member)
     cf_error("Confederation ID must be set for member sessions");
 
@@ -2004,27 +1807,9 @@ bgp_postconfig(struct proto_config *CF)
   if (!cf->gr_mode && cf->llgr_mode)
     cf_error("Long-lived graceful restart requires basic graceful restart");
 
-  if (internal && cf->enforce_first_as)
-    cf_error("Enforce first AS check is requires EBGP sessions");
-
-  if (cf->keepalive_time > cf->hold_time)
-    cf_error("Keepalive time must be at most hold time");
-
-  if (cf->keepalive_time > (cf->hold_time / 2))
-    log(L_WARN "Keepalive time should be at most 1/2 of hold time");
-
-  if (cf->min_hold_time > cf->hold_time)
-    cf_error("Min hold time (%u) exceeds hold time (%u)",
-	     cf->min_hold_time, cf->hold_time);
-
-  uint keepalive_time = cf->keepalive_time ?: cf->hold_time / 3;
-  if (cf->min_keepalive_time > keepalive_time)
-    cf_error("Min keepalive time (%u) exceeds keepalive time (%u)",
-	     cf->min_keepalive_time, keepalive_time);
-
 
   struct bgp_channel_config *cc;
-  BGP_CF_WALK_CHANNELS(cf, cc)
+  WALK_LIST(cc, CF->channels)
   {
     /* Handle undefined import filter */
     if (cc->c.in_filter == FILTER_UNDEF)
@@ -2048,14 +1833,14 @@ bgp_postconfig(struct proto_config *CF)
     if (cc->next_hop_keep == 0xff)
       cc->next_hop_keep = cf->rr_client ? NH_IBGP : (cf->rs_client ? NH_ALL : NH_NO);
 
+    /* Different default based on rs_client */
+    if (!cc->missing_lladdr)
+      cc->missing_lladdr = cf->rs_client ? MLL_IGNORE : MLL_SELF;
+
     /* Different default for gw_mode */
     if (!cc->gw_mode)
       cc->gw_mode = cf->multihop ? GW_RECURSIVE : GW_DIRECT;
 
-    /* Different default for next_hop_prefer */
-    if (!cc->next_hop_prefer)
-      cc->next_hop_prefer = (cc->gw_mode == GW_DIRECT) ? NHP_GLOBAL : NHP_LOCAL;
-
     /* Defaults based on proto config */
     if (cc->gr_able == 0xff)
       cc->gr_able = (cf->gr_mode == BGP_GR_ABLE);
@@ -2066,10 +1851,6 @@ bgp_postconfig(struct proto_config *CF)
     if (cc->llgr_time == ~0U)
       cc->llgr_time = cf->llgr_time;
 
-    /* AIGP enabled by default on interior sessions */
-    if (cc->aigp == 0xff)
-      cc->aigp = interior;
-
     /* Default values of IGP tables */
     if ((cc->gw_mode == GW_RECURSIVE) && !cc->desc->no_igp)
     {
@@ -2086,14 +1867,6 @@ bgp_postconfig(struct proto_config *CF)
 	cf_error("Mismatched IGP table type");
     }
 
-    /* Default value of base table */
-    if ((BGP_SAFI(cc->afi) == BGP_SAFI_FLOW) && cc->validate && !cc->base_table)
-      cc->base_table = bgp_default_base_table(cf, cc);
-
-    if (cc->base_table && !cc->base_table->trie_used)
-      cf_error("Flowspec validation requires base table (%s) with trie",
-	       cc->base_table->name);
-
     if (cf->multihop && (cc->gw_mode == GW_DIRECT))
       cf_error("Multihop BGP cannot use direct gateway mode");
 
@@ -2112,8 +1885,8 @@ static int
 bgp_reconfigure(struct proto *P, struct proto_config *CF)
 {
   struct bgp_proto *p = (void *) P;
-  const struct bgp_config *new = (void *) CF;
-  const struct bgp_config *old = p->cf;
+  struct bgp_config *new = (void *) CF;
+  struct bgp_config *old = p->cf;
 
   if (proto_get_router_id(CF) != p->local_id)
     return 0;
@@ -2122,11 +1895,8 @@ bgp_reconfigure(struct proto *P, struct proto_config *CF)
 		     ((byte *) new) + sizeof(struct proto_config),
 		     // password item is last and must be checked separately
 		     OFFSETOF(struct bgp_config, password) - sizeof(struct proto_config))
-    && !bstrcmp(old->password, new->password)
-    && ((!old->remote_range && !new->remote_range)
-	|| (old->remote_range && new->remote_range && net_equal(old->remote_range, new->remote_range)))
-    && !bstrcmp(old->dynamic_name, new->dynamic_name)
-    && (old->dynamic_name_digits == new->dynamic_name_digits);
+    && ((!old->password && !new->password)
+	|| (old->password && new->password && !strcmp(old->password, new->password)));
 
   /* FIXME: Move channel reconfiguration to generic protocol code ? */
   struct channel *C, *C2;
@@ -2135,16 +1905,20 @@ bgp_reconfigure(struct proto *P, struct proto_config *CF)
   WALK_LIST(C, p->p.channels)
     C->stale = 1;
 
-  BGP_CF_WALK_CHANNELS(new, cc)
+  WALK_LIST(cc, new->c.channels)
   {
     C = (struct channel *) bgp_find_channel(p, cc->afi);
     same = proto_configure_channel(P, &C, &cc->c) && same;
+
+    if (C)
+      C->stale = 0;
   }
 
   WALK_LIST_DELSAFE(C, C2, p->p.channels)
     if (C->stale)
       same = proto_configure_channel(P, &C, NULL) && same;
 
+
   if (same && (p->start_state > BSS_PREPARE))
     bgp_update_bfd(p, new->bfd);
 
@@ -2152,75 +1926,37 @@ bgp_reconfigure(struct proto *P, struct proto_config *CF)
   if (same)
     p->cf = new;
 
-  /* Reset name counter */
-  p->dynamic_name_counter = 0;
-
   return same;
 }
 
-#define TABLE(cf, NAME) ((cf)->NAME ? (cf)->NAME->table : NULL )
+#define IGP_TABLE(cf, sym) ((cf)->igp_table_##sym ? (cf)->igp_table_##sym ->table : NULL )
 
 static int
-bgp_channel_reconfigure(struct channel *C, struct channel_config *CC, int *import_changed, int *export_changed)
+bgp_channel_reconfigure(struct channel *C, struct channel_config *CC)
 {
-  struct bgp_proto *p = (void *) C->proto;
   struct bgp_channel *c = (void *) C;
   struct bgp_channel_config *new = (void *) CC;
   struct bgp_channel_config *old = c->cf;
 
-  if ((new->secondary != old->secondary) ||
-      (new->validate != old->validate) ||
-      (new->gr_able != old->gr_able) ||
-      (new->llgr_able != old->llgr_able) ||
-      (new->llgr_time != old->llgr_time) ||
-      (new->ext_next_hop != old->ext_next_hop) ||
-      (new->add_path != old->add_path) ||
-      (new->import_table != old->import_table) ||
-      (new->export_table != old->export_table) ||
-      (TABLE(new, igp_table_ip4) != TABLE(old, igp_table_ip4)) ||
-      (TABLE(new, igp_table_ip6) != TABLE(old, igp_table_ip6)) ||
-      (TABLE(new, base_table) != TABLE(old, base_table)))
+  if (memcmp(((byte *) old) + sizeof(struct channel_config),
+	     ((byte *) new) + sizeof(struct channel_config),
+	     /* Remaining items must be checked separately */
+	     OFFSETOF(struct bgp_channel_config, rest) - sizeof(struct channel_config)))
     return 0;
 
-  if (new->mandatory && !old->mandatory && (C->channel_state != CS_UP))
+  /* Check change in IGP tables */
+  if ((IGP_TABLE(old, ip4) != IGP_TABLE(new, ip4)) ||
+      (IGP_TABLE(old, ip6) != IGP_TABLE(new, ip6)))
     return 0;
 
-  if ((new->gw_mode != old->gw_mode) ||
-      (new->next_hop_prefer != old->next_hop_prefer) ||
-      (new->aigp != old->aigp) ||
-      (new->cost != old->cost))
-  {
-    /* import_changed itself does not force ROUTE_REFRESH when import_table is active */
-    if (c->c.in_table && (c->c.channel_state == CS_UP))
-      bgp_schedule_packet(p->conn, c, PKT_ROUTE_REFRESH);
-
-    *import_changed = 1;
-  }
-
-  if (!ipa_equal(new->next_hop_addr, old->next_hop_addr) ||
-      (new->next_hop_self != old->next_hop_self) ||
-      (new->next_hop_keep != old->next_hop_keep) ||
-      (new->aigp != old->aigp) ||
-      (new->aigp_originate != old->aigp_originate))
-    *export_changed = 1;
-
   c->cf = new;
   return 1;
 }
 
 static void
-bgp_copy_config(struct proto_config *dest, struct proto_config *src)
+bgp_copy_config(struct proto_config *dest UNUSED, struct proto_config *src UNUSED)
 {
-  struct bgp_config *d = (void *) dest;
-  struct bgp_config *s = (void *) src;
-
-  /* Copy BFD options */
-  if (s->bfd)
-  {
-    struct bfd_options *opts = cfg_alloc(sizeof(struct bfd_options));
-    memcpy(opts, s->bfd, sizeof(struct bfd_options));
-    d->bfd = opts;
-  }
+  /* Just a shallow copy */
 }
 
 
@@ -2320,7 +2056,7 @@ bgp_state_dsc(struct bgp_proto *p)
     return "Down";
 
   int state = MAX(p->incoming_conn.state, p->outgoing_conn.state);
-  if ((state == BS_IDLE) && (p->start_state >= BSS_CONNECT) && p->passive)
+  if ((state == BS_IDLE) && (p->start_state >= BSS_CONNECT) && p->cf->passive)
     return "Passive";
 
   return bgp_state_names[state];
@@ -2363,15 +2099,6 @@ bgp_show_afis(int code, char *s, u32 *afis, uint count)
   cli_msg(code, b.start);
 }
 
-static const char *
-bgp_format_role_name(u8 role)
-{
-  static const char *bgp_role_names[] = { "provider", "rs_server", "rs_client", "customer", "peer" };
-  if (role == BGP_ROLE_UNDEFINED) return "undefined";
-  if (role < ARRAY_SIZE(bgp_role_names)) return bgp_role_names[role];
-  return "?";
-}
-
 static void
 bgp_show_capabilities(struct bgp_proto *p UNUSED, struct bgp_caps *caps)
 {
@@ -2497,12 +2224,6 @@ bgp_show_capabilities(struct bgp_proto *p UNUSED, struct bgp_caps *caps)
     bgp_show_afis(-1006, "        AF supported:", afl1, afn1);
     bgp_show_afis(-1006, "        AF preserved:", afl2, afn2);
   }
-
-  if (caps->hostname)
-    cli_msg(-1006, "      Hostname: %s", caps->hostname);
-
-  if (caps->role != BGP_ROLE_UNDEFINED)
-    cli_msg(-1006, "      Role: %s", bgp_format_role_name(caps->role));
 }
 
 static void
@@ -2511,17 +2232,8 @@ bgp_show_proto_info(struct proto *P)
   struct bgp_proto *p = (struct bgp_proto *) P;
 
   cli_msg(-1006, "  BGP state:          %s", bgp_state_dsc(p));
-
-  if (bgp_is_dynamic(p) && p->cf->remote_range)
-    cli_msg(-1006, "    Neighbor range:   %N", p->cf->remote_range);
-  else
-    cli_msg(-1006, "    Neighbor address: %I%J", p->remote_ip, p->cf->iface);
-
-  if ((p->conn == &p->outgoing_conn) && (p->cf->remote_port != BGP_PORT))
-    cli_msg(-1006, "    Neighbor port:    %u", p->cf->remote_port);
-
+  cli_msg(-1006, "    Neighbor address: %I%J", p->cf->remote_ip, p->cf->iface);
   cli_msg(-1006, "    Neighbor AS:      %u", p->remote_as);
-  cli_msg(-1006, "    Local AS:         %u", p->cf->local_as);
 
   if (p->gr_active_num)
     cli_msg(-1006, "    Neighbor graceful restart active");
@@ -2557,25 +2269,13 @@ bgp_show_proto_info(struct proto *P)
 	    p->rr_client ? " route-reflector" : "",
 	    p->rs_client ? " route-server" : "",
 	    p->as4_session ? " AS4" : "");
-    cli_msg(-1006, "    Source address:   %I", p->local_ip);
+    cli_msg(-1006, "    Source address:   %I", p->source_addr);
     cli_msg(-1006, "    Hold timer:       %t/%u",
 	    tm_remains(p->conn->hold_timer), p->conn->hold_time);
     cli_msg(-1006, "    Keepalive timer:  %t/%u",
 	    tm_remains(p->conn->keepalive_timer), p->conn->keepalive_time);
   }
 
-#if 0
-  struct bgp_stats *s = &p->stats;
-  cli_msg(-1006, "    FSM established transitions: %u",
-	  s->fsm_established_transitions);
-  cli_msg(-1006, "    Rcvd messages:    %u total / %u updates / %lu bytes",
-	  s->rx_messages, s->rx_updates, s->rx_bytes);
-  cli_msg(-1006, "    Sent messages:    %u total / %u updates / %lu bytes",
-	  s->tx_messages, s->tx_updates, s->tx_bytes);
-  cli_msg(-1006, "    Last rcvd update elapsed time: %t s",
-	  p->last_rx_update ? (current_time() - p->last_rx_update) : 0);
-#endif
-
   if ((p->last_error_class != BE_NONE) &&
       (p->last_error_class != BE_MAN_DOWN))
   {
@@ -2590,9 +2290,6 @@ bgp_show_proto_info(struct proto *P)
     {
       channel_show_info(&c->c);
 
-      if (c->c.channel != &channel_bgp)
-	continue;
-
       if (p->gr_active_num)
 	cli_msg(-1006, "    Neighbor GR:    %s", bgp_gr_states[c->gr_active]);
 
@@ -2612,9 +2309,6 @@ bgp_show_proto_info(struct proto *P)
 
       if (c->igp_table_ip6)
 	cli_msg(-1006, "    IGP IPv6 table: %s", c->igp_table_ip6->name);
-
-      if (c->base_table)
-	cli_msg(-1006, "    Base table:     %s", c->base_table->name);
     }
   }
 }
@@ -2648,8 +2342,3 @@ struct protocol proto_bgp = {
   .get_route_info = 	bgp_get_route_info,
   .show_proto_info = 	bgp_show_proto_info
 };
-
-void bgp_build(void)
-{
-  proto_build(&proto_bgp);
-}
