@@ -20,6 +20,7 @@
 #include "lib/hash.h"
 #include "lib/socket.h"
 
+struct linpool;
 struct eattr;
 
 
@@ -82,10 +83,8 @@ struct bgp_config {
   struct iface *iface;			/* Interface for link-local addresses */
   u16 local_port;			/* Local listening port */
   u16 remote_port; 			/* Neighbor destination port */
-  int peer_type;			/* Internal or external BGP (BGP_PT_*, optional) */
   int multihop;				/* Number of hops if multihop */
   int strict_bind;			/* Bind listening socket to local address */
-  int free_bind;			/* Bind listening socket with SKF_FREEBIND */
   int ttl_security;			/* Enable TTL security [RFC 5082] */
   int compare_path_lengths;		/* Use path lengths when selecting best route */
   int med_metric;			/* Compare MULTI_EXIT_DISC even between routes from differen ASes */
@@ -98,7 +97,6 @@ struct bgp_config {
   int enable_refresh;			/* Enable local support for route refresh [RFC 2918] */
   int enable_as4;			/* Enable local support for 4B AS numbers [RFC 6793] */
   int enable_extended_messages;		/* Enable local support for extended messages [draft] */
-  int enable_hostname;			/* Enable local support for hostname [draft] */
   u32 rr_cluster_id;			/* Route reflector cluster ID, if different from local ID */
   int rr_client;			/* Whether neighbor is RR client of me */
   int rs_client;			/* Whether neighbor is RS client of me */
@@ -108,35 +106,25 @@ struct bgp_config {
   int interpret_communities;		/* Hardwired handling of well-known communities */
   int allow_local_as;			/* Allow that number of local ASNs in incoming AS_PATHs */
   int allow_local_pref;			/* Allow LOCAL_PREF in EBGP sessions */
-  int allow_as_sets;			/* Allow AS_SETs in incoming AS_PATHs */
-  int enforce_first_as;			/* Enable check for neighbor AS as first AS in AS_PATH */
   int gr_mode;				/* Graceful restart mode (BGP_GR_*) */
   int llgr_mode;			/* Long-lived graceful restart mode (BGP_LLGR_*) */
   int setkey;				/* Set MD5 password to system SA/SP database */
-  u8  local_role;			/* Set peering role with neighbor [RFC 9234] */
-  int require_roles;			/* Require configured roles on both sides */
   /* Times below are in seconds */
   unsigned gr_time;			/* Graceful restart timeout */
   unsigned llgr_time;			/* Long-lived graceful restart stale time */
   unsigned connect_delay_time;		/* Minimum delay between connect attempts */
   unsigned connect_retry_time;		/* Timeout for connect attempts */
-  unsigned hold_time;
-  unsigned min_hold_time;		/* Minimum accepted hold time */
-  unsigned initial_hold_time;
+  unsigned hold_time, initial_hold_time;
   unsigned keepalive_time;
-  unsigned min_keepalive_time;		/* Minimum accepted keepalive time */
   unsigned error_amnesia_time;		/* Errors are forgotten after */
   unsigned error_delay_time_min;	/* Time to wait after an error is detected */
   unsigned error_delay_time_max;
   unsigned disable_after_error;		/* Disable the protocol when error is detected */
   u32 disable_after_cease;		/* Disable it when cease is received, bitfield */
 
-  const char *password;			/* Password used for MD5 authentication */
-  net_addr *remote_range;		/* Allowed neighbor range for dynamic BGP */
-  const char *dynamic_name;		/* Name pattern for dynamic BGP */
-  int dynamic_name_digits;		/* Minimum number of digits for dynamic names */
+  char *password;			/* Password used for MD5 authentication */
   int check_link;			/* Use iface link state for liveness detection */
-  struct bfd_options *bfd;		/* Use BFD for liveness detection */
+  int bfd;				/* Use BFD for liveness detection */
 };
 
 struct bgp_channel_config {
@@ -148,37 +136,21 @@ struct bgp_channel_config {
   ip_addr next_hop_addr;		/* Local address for NEXT_HOP attribute */
   u8 next_hop_self;			/* Always set next hop to local IP address (NH_*) */
   u8 next_hop_keep;			/* Do not modify next hop attribute (NH_*) */
-  u8 next_hop_prefer;			/* Prefer global or link-local next hop (NHP_*) */
-  u8 mandatory;				/* Channel is mandatory in capability negotiation */
+  u8 missing_lladdr;			/* What we will do when we don' know link-local addr, see MLL_* */
   u8 gw_mode;				/* How we compute route gateway from next_hop attr, see GW_* */
   u8 secondary;				/* Accept also non-best routes (i.e. RA_ACCEPTED) */
-  u8 validate;				/* Validate Flowspec per RFC 8955 (6) */
   u8 gr_able;				/* Allow full graceful restart for the channel */
   u8 llgr_able;				/* Allow full long-lived GR for the channel */
   uint llgr_time;			/* Long-lived graceful restart stale time */
   u8 ext_next_hop;			/* Allow both IPv4 and IPv6 next hops */
   u8 add_path;				/* Use ADD-PATH extension [RFC 7911] */
-  u8 aigp;				/* AIGP is allowed on this session */
-  u8 aigp_originate;			/* AIGP is originated automatically */
-  u32 cost;				/* IGP cost for direct next hops */
   u8 import_table;			/* Use c.in_table as Adj-RIB-In */
-  u8 export_table;			/* Use c.out_table as Adj-RIB-Out */
 
+  uint rest[0];				/* Remaining items are reconfigured separately */
   struct rtable_config *igp_table_ip4;	/* Table for recursive IPv4 next hop lookups */
   struct rtable_config *igp_table_ip6;	/* Table for recursive IPv6 next hop lookups */
-  struct rtable_config *base_table;	/* Base table for Flowspec validation */
 };
 
-#define BGP_PT_INTERNAL		1
-#define BGP_PT_EXTERNAL		2
-
-#define BGP_ROLE_UNDEFINED 	255
-#define BGP_ROLE_PROVIDER 	0
-#define BGP_ROLE_RS_SERVER 	1
-#define BGP_ROLE_RS_CLIENT 	2
-#define BGP_ROLE_CUSTOMER 	3
-#define BGP_ROLE_PEER 		4
-
 #define NH_NO			0
 #define NH_ALL			1
 #define NH_IBGP			2
@@ -191,9 +163,6 @@ struct bgp_channel_config {
 #define GW_DIRECT		1
 #define GW_RECURSIVE		2
 
-#define NHP_GLOBAL		1
-#define NHP_LOCAL		2
-
 #define BGP_ADD_PATH_RX		1
 #define BGP_ADD_PATH_TX		2
 #define BGP_ADD_PATH_FULL	3
@@ -218,10 +187,6 @@ struct bgp_channel_config {
 
 #define BGP_BFD_GRACEFUL	2	/* BFD down triggers graceful restart */
 
-/* rte->pflags */
-#define BGP_REF_SUPPRESSED	0x1	/* Used for deterministic MED comparison */
-#define BGP_REF_STALE		0x2	/* Route is LLGR_STATE */
-#define BGP_REF_NOT_STALE	0x4	/* Route is NOT LLGR_STATE */
 
 struct bgp_af_caps {
   u32 afi;
@@ -242,20 +207,14 @@ struct bgp_caps {
   u8 ext_messages;			/* Extended message length,  RFC draft */
   u8 route_refresh;			/* Route refresh capability, RFC 2918 */
   u8 enhanced_refresh;			/* Enhanced route refresh,   RFC 7313 */
-  u8 role;				/* BGP role capability,      RFC 9234 */
 
   u8 gr_aware;				/* Graceful restart capability, RFC 4724 */
   u8 gr_flags;				/* Graceful restart flags */
   u16 gr_time;				/* Graceful restart time in seconds */
 
   u8 llgr_aware;			/* Long-lived GR capability, RFC draft */
-  u8 any_ext_next_hop;			/* Bitwise OR of per-AF ext_next_hop */
-  u8 any_add_path;			/* Bitwise OR of per-AF add_path */
-
-  const char *hostname;			/* Hostname, RFC draft */
 
   u16 af_count;				/* Number of af_data items */
-  u16 length;				/* Length of capabilities in OPEN msg */
 
   struct bgp_af_caps af_data[0];	/* Per-AF capability data */
 };
@@ -270,21 +229,12 @@ struct bgp_socket {
   u32 uc;				/* Use count */
 };
 
-struct bgp_stats {
-  uint rx_messages, tx_messages;
-  uint rx_updates, tx_updates;
-  u64  rx_bytes, tx_bytes;
-
-  uint fsm_established_transitions;
-};
-
 struct bgp_conn {
   struct bgp_proto *bgp;
   struct birdsock *sk;
   u8 state;				/* State of connection state machine */
   u8 as4_session;			/* Session uses 4B AS numbers in AS_PATH (both sides support it) */
   u8 ext_messages;			/* Session uses extended message length */
-  u32 received_as;			/* ASN received in OPEN message */
 
   struct bgp_caps *local_caps;
   struct bgp_caps *remote_caps;
@@ -304,21 +254,18 @@ struct bgp_conn {
 
 struct bgp_proto {
   struct proto p;
-  const struct bgp_config *cf;		/* Shortcut to BGP configuration */
-  ip_addr local_ip, remote_ip;
+  struct bgp_config *cf;		/* Shortcut to BGP configuration */
   u32 local_as, remote_as;
   u32 public_as;			/* Externally visible ASN (local_as or confederation id) */
   u32 local_id;				/* BGP identifier of this router */
   u32 remote_id;			/* BGP identifier of the neighbor */
   u32 rr_cluster_id;			/* Route reflector cluster ID */
-  u8 start_state;			/* Substates that partitions BS_START */
+  int start_state;			/* Substates that partitions BS_START */
   u8 is_internal;			/* Internal BGP session (local_as == remote_as) */
   u8 is_interior;			/* Internal or intra-confederation BGP session */
   u8 as4_session;			/* Session uses 4B AS numbers in AS_PATH (both sides support it) */
   u8 rr_client;				/* Whether neighbor is RR client of me */
   u8 rs_client;				/* Whether neighbor is RS client of me */
-  u8 ipv4;				/* Use IPv4 connection, i.e. remote_ip is IPv4 */
-  u8 passive;				/* Do not initiate outgoing connection */
   u8 route_refresh;			/* Route refresh allowed to send [RFC 2918] */
   u8 enhanced_refresh;			/* Enhanced refresh is negotiated [RFC 7313] */
   u8 gr_ready;				/* Neighbor could do graceful restart */
@@ -335,15 +282,11 @@ struct bgp_proto {
   struct neighbor *neigh;		/* Neighbor entry corresponding to remote ip, NULL if multihop */
   struct bgp_socket *sock;		/* Shared listening socket */
   struct bfd_request *bfd_req;		/* BFD request, if BFD is used */
-  struct birdsock *postponed_sk;	/* Postponed incoming socket for dynamic BGP */
-  struct bgp_stats stats;		/* BGP statistics */
-  btime last_established;		/* Last time of enter/leave of established state */
-  btime last_rx_update;			/* Last time of RX update */
-  ip_addr link_addr;			/* Link-local version of local_ip */
+  ip_addr source_addr;			/* Local address used as an advertised next hop */
+  ip_addr link_addr;			/* Link-local version of source_addr */
   event *event;				/* Event for respawning and shutting process */
   timer *startup_timer;			/* Timer used to delay protocol startup due to previous errors (startup_delay) */
   timer *gr_timer;			/* Timer waiting for reestablishment after graceful restart */
-  int dynamic_name_counter;		/* Counter for dynamic BGP names */
   uint startup_delay;			/* Delay (in seconds) of protocol startup due to previous errors */
   btime last_proto_error;		/* Time of last error that leads to protocol stop */
   u8 last_error_class; 			/* Error class of last error */
@@ -363,7 +306,6 @@ struct bgp_channel {
 
   rtable *igp_table_ip4;		/* Table for recursive IPv4 next hop lookups */
   rtable *igp_table_ip6;		/* Table for recursive IPv6 next hop lookups */
-  rtable *base_table;			/* Base table for Flowspec validation */
 
   /* Rest are zeroed when down */
   pool *pool;
@@ -420,8 +362,7 @@ struct bgp_export_state {
   int mpls;
 
   u32 attrs_seen[1];
-  uint err_reject;
-  uint local_next_hop;
+  uint err_withdraw;
 };
 
 struct bgp_write_state {
@@ -435,7 +376,7 @@ struct bgp_write_state {
   int mpls;
 
   eattr *mp_next_hop;
-  const adata *mpls_labels;
+  adata *mpls_labels;
 };
 
 struct bgp_parse_state {
@@ -446,7 +387,6 @@ struct bgp_parse_state {
   int as4_session;
   int add_path;
   int mpls;
-  int reach_nlri_step;
 
   u32 attrs_seen[256/32];
 
@@ -492,9 +432,6 @@ struct bgp_parse_state {
 #define BGP_RX_BUFFER_EXT_SIZE	65535
 #define BGP_TX_BUFFER_EXT_SIZE	65535
 
-#define BGP_CF_WALK_CHANNELS(P,C) WALK_LIST(C, P->c.channels) if (C->c.channel == &channel_bgp)
-#define BGP_WALK_CHANNELS(P,C) WALK_LIST(C, P->p.channels) if (C->c.channel == &channel_bgp)
-
 static inline int bgp_channel_is_ipv4(struct bgp_channel *c)
 { return BGP_AFI(c->afi) == BGP_AFI_IPV4; }
 
@@ -507,12 +444,6 @@ static inline int bgp_cc_is_ipv4(struct bgp_channel_config *c)
 static inline int bgp_cc_is_ipv6(struct bgp_channel_config *c)
 { return BGP_AFI(c->afi) == BGP_AFI_IPV6; }
 
-static inline int bgp_channel_is_role_applicable(struct bgp_channel *c)
-{ return (c->afi == BGP_AF_IPV4 || c->afi == BGP_AF_IPV6); }
-
-static inline int bgp_cc_is_role_applicable(struct bgp_channel_config *c)
-{ return (c->afi == BGP_AF_IPV4 || c->afi == BGP_AF_IPV6); }
-
 static inline uint bgp_max_packet_length(struct bgp_conn *conn)
 { return conn->ext_messages ? BGP_MAX_EXT_MSG_LENGTH : BGP_MAX_MESSAGE_LENGTH; }
 
@@ -523,6 +454,9 @@ bgp_parse_error(struct bgp_parse_state *s, uint subcode)
   longjmp(s->err_jmpbuf, 1);
 }
 
+extern struct linpool *bgp_linpool;
+extern struct linpool *bgp_linpool2;
+
 
 void bgp_start_timer(timer *t, uint value);
 void bgp_check_config(struct bgp_config *c);
@@ -538,16 +472,11 @@ void bgp_graceful_restart_done(struct bgp_channel *c);
 void bgp_refresh_begin(struct bgp_channel *c);
 void bgp_refresh_end(struct bgp_channel *c);
 void bgp_store_error(struct bgp_proto *p, struct bgp_conn *c, u8 class, u32 code);
-void bgp_stop(struct bgp_proto *p, int subcode, byte *data, uint len);
+void bgp_stop(struct bgp_proto *p, uint subcode, byte *data, uint len);
 
 struct rte_source *bgp_find_source(struct bgp_proto *p, u32 path_id);
 struct rte_source *bgp_get_source(struct bgp_proto *p, u32 path_id);
 
-static inline int
-rte_resolvable(rte *rt)
-{
-  return rt->attrs->dest != RTD_UNREACHABLE;
-}
 
 
 #ifdef LOCAL_DEBUG
@@ -578,24 +507,24 @@ bgp_set_attr_u32(ea_list **to, struct linpool *pool, uint code, uint flags, u32
 { bgp_set_attr(to, pool, code, flags, (uintptr_t) val); }
 
 static inline void
-bgp_set_attr_ptr(ea_list **to, struct linpool *pool, uint code, uint flags, const struct adata *val)
+bgp_set_attr_ptr(ea_list **to, struct linpool *pool, uint code, uint flags, struct adata *val)
 { bgp_set_attr(to, pool, code, flags, (uintptr_t) val); }
 
 static inline void
 bgp_set_attr_data(ea_list **to, struct linpool *pool, uint code, uint flags, void *data, uint len)
 {
   struct adata *a = lp_alloc_adata(pool, len);
-  bmemcpy(a->data, data, len);
+  memcpy(a->data, data, len);
   bgp_set_attr(to, pool, code, flags, (uintptr_t) a);
 }
 
-#define bgp_unset_attr(to, pool, code) ea_unset_attr(to, pool, 0, code)
+static inline void
+bgp_unset_attr(ea_list **to, struct linpool *pool, uint code)
+{ eattr *e = bgp_set_attr(to, pool, code, 0, 0); e->type = EAF_TYPE_UNDEF; }
 
-int bgp_encode_mp_reach_mrt(struct bgp_write_state *s, eattr *a, byte *buf, uint size);
 
 int bgp_encode_attrs(struct bgp_write_state *s, ea_list *attrs, byte *buf, byte *end);
 ea_list * bgp_decode_attrs(struct bgp_parse_state *s, byte *data, uint len);
-void bgp_finish_attrs(struct bgp_parse_state *s, rta *a);
 
 void bgp_init_bucket_table(struct bgp_channel *c);
 void bgp_free_bucket_table(struct bgp_channel *c);
@@ -611,31 +540,15 @@ int bgp_rte_better(struct rte *, struct rte *);
 int bgp_rte_mergable(rte *pri, rte *sec);
 int bgp_rte_recalculate(rtable *table, net *net, rte *new, rte *old, rte *old_best);
 struct rte *bgp_rte_modify_stale(struct rte *r, struct linpool *pool);
-u32 bgp_rte_igp_metric(struct rte *);
 void bgp_rt_notify(struct proto *P, struct channel *C, net *n, rte *new, rte *old);
-int bgp_preexport(struct channel *, struct rte *);
-int bgp_get_attr(const struct eattr *e, byte *buf, int buflen);
+int bgp_preexport(struct proto *, struct rte **, struct linpool *);
+int bgp_get_attr(struct eattr *e, byte *buf, int buflen);
 void bgp_get_route_info(struct rte *, byte *buf);
-int bgp_total_aigp_metric_(rte *e, u64 *metric, const struct adata **ad);
-
-#define BGP_AIGP_METRIC		1
-#define BGP_AIGP_MAX		U64(0xffffffffffffffff)
-
-static inline u64
-bgp_total_aigp_metric(rte *r)
-{
-  u64 metric = BGP_AIGP_MAX;
-  const struct adata *ad;
-
-  bgp_total_aigp_metric_(r, &metric, &ad);
-  return metric;
-}
 
 
 /* packets.c */
 
 void bgp_dump_state_change(struct bgp_conn *conn, uint old, uint new);
-void bgp_prepare_capabilities(struct bgp_conn *conn);
 const struct bgp_af_desc *bgp_get_af_desc(u32 afi);
 const struct bgp_af_caps *bgp_find_af_caps(struct bgp_caps *caps, u32 afi);
 void bgp_schedule_packet(struct bgp_conn *conn, struct bgp_channel *c, int type);
@@ -665,8 +578,6 @@ void bgp_update_next_hop(struct bgp_export_state *s, eattr *a, ea_list **to);
 #define BAF_PARTIAL		0x20
 #define BAF_EXT_LEN		0x10
 
-#define BAF_DECODE_FLAGS	0x0100	/* Private flag - attribute flags are handled by the decode hook */
-
 #define BA_ORIGIN		0x01	/* RFC 4271 */		/* WM */
 #define BA_AS_PATH		0x02				/* WM */
 #define BA_NEXT_HOP		0x03				/* WM */
@@ -682,9 +593,7 @@ void bgp_update_next_hop(struct bgp_export_state *s, eattr *a, ea_list **to);
 #define BA_EXT_COMMUNITY	0x10	/* RFC 4360 */
 #define BA_AS4_PATH             0x11	/* RFC 6793 */
 #define BA_AS4_AGGREGATOR       0x12	/* RFC 6793 */
-#define BA_AIGP			0x1a	/* RFC 7311 */
 #define BA_LARGE_COMMUNITY	0x20	/* RFC 8092 */
-#define BA_ONLY_TO_CUSTOMER	0x23	/* RFC 9234 */
 
 /* Bird's private internal BGP attributes */
 #define BA_MPLS_LABEL_STACK	0xfe	/* MPLS label stack transfer attribute */
