@@ -69,11 +69,10 @@
  * structure is finally freed.
  *
  * Supported standards:
- * RFC 1058 - RIPv1
- * RFC 2453 - RIPv2
- * RFC 2080 - RIPng
- * RFC 2091 - Triggered RIP for demand circuits
- * RFC 4822 - RIP cryptographic authentication
+ * - RFC 1058 - RIPv1
+ * - RFC 2453 - RIPv2
+ * - RFC 2080 - RIPng
+ * - RFC 4822 - RIP cryptographic authentication
  */
 
 #include <stdlib.h>
@@ -108,14 +107,14 @@ rip_add_rte(struct rip_proto *p, struct rip_rte **rp, struct rip_rte *src)
 }
 
 static inline void
-rip_remove_rte(struct rip_proto *p UNUSED, struct rip_rte **rp)
+rip_remove_rte(struct rip_proto *p, struct rip_rte **rp)
 {
   struct rip_rte *rt = *rp;
 
   rip_unlock_neighbor(rt->from);
 
   *rp = rt->next;
-  sl_free(rt);
+  sl_free(p->rte_slab, rt);
 }
 
 static inline int rip_same_rte(struct rip_rte *a, struct rip_rte *b)
@@ -124,11 +123,6 @@ static inline int rip_same_rte(struct rip_rte *a, struct rip_rte *b)
 static inline int rip_valid_rte(struct rip_rte *rt)
 { return rt->from->ifa != NULL; }
 
-struct rip_iface_adata {
-  struct adata ad;
-  struct iface *iface;
-};
-
 /**
  * rip_announce_rte - announce route from RIP routing table to the core
  * @p: RIP instance
@@ -150,7 +144,7 @@ rip_announce_rte(struct rip_proto *p, struct rip_entry *en)
   {
     /* Update */
     rta a0 = {
-      .pref = p->p.main_channel->preference,
+      .src = p->p.main_source,
       .source = RTS_RIP,
       .scope = SCOPE_UNIVERSE,
       .dest = RTD_UNICAST,
@@ -173,7 +167,7 @@ rip_announce_rte(struct rip_proto *p, struct rip_entry *en)
 	struct nexthop *nh = allocz(sizeof(struct nexthop));
 
 	nh->gw = rt->next_hop;
-	nh->iface = rt->from->ifa->iface;
+	nh->iface = rt->from->nbr->iface;
 	nh->weight = rt->from->ifa->cf->ecmp_weight;
 
 	nexthop_insert(&nhs, nh);
@@ -190,42 +184,17 @@ rip_announce_rte(struct rip_proto *p, struct rip_entry *en)
       /* Unipath route */
       a0.from = rt->from->nbr->addr;
       a0.nh.gw = rt->next_hop;
-      a0.nh.iface = rt->from->ifa->iface;
+      a0.nh.iface = rt->from->nbr->iface;
     }
 
-    struct {
-      ea_list l;
-      eattr e[3];
-      struct rip_iface_adata riad;
-    } ea_block = {
-      .l = { .count = 3, },
-      .e = {
-	{
-	  .id = EA_RIP_METRIC,
-	  .type = EAF_TYPE_INT,
-	  .u.data = rt_metric,
-	},
-	{
-	  .id = EA_RIP_TAG,
-	  .type = EAF_TYPE_INT,
-	  .u.data = rt_tag,
-	},
-	{
-	  .id = EA_RIP_FROM,
-	  .type = EAF_TYPE_IFACE,
-	  .u.ptr = &ea_block.riad.ad,
-	}
-      },
-      .riad = {
-	.ad = { .length = sizeof(struct rip_iface_adata) - sizeof(struct adata) },
-	.iface = a0.nh.iface,
-      },
-    };
-
-    a0.eattrs = &ea_block.l;
-
     rta *a = rta_lookup(&a0);
-    rte *e = rte_get_temp(a, p->p.main_source);
+    rte *e = rte_get_temp(a);
+
+    e->u.rip.from = a0.nh.iface;
+    e->u.rip.metric = rt_metric;
+    e->u.rip.tag = rt_tag;
+
+    e->pflags = 0;
 
     rte_update(&p->p, en->n.addr, e);
   }
@@ -338,10 +307,8 @@ rip_rt_notify(struct proto *P, struct channel *ch UNUSED, struct network *net, s
   if (new)
   {
     /* Update */
-    u32 rt_tag = ea_get_int(new->attrs->eattrs, EA_RIP_TAG, 0);
     u32 rt_metric = ea_get_int(new->attrs->eattrs, EA_RIP_METRIC, 1);
-    const eattr *rie = ea_find(new->attrs->eattrs, EA_RIP_FROM);
-    struct iface *rt_from = rie ? ((struct rip_iface_adata *) rie->u.ptr)->iface : NULL;
+    u32 rt_tag = ea_get_int(new->attrs->eattrs, EA_RIP_TAG, 0);
 
     if (rt_metric > p->infinity)
     {
@@ -372,7 +339,7 @@ rip_rt_notify(struct proto *P, struct channel *ch UNUSED, struct network *net, s
     en->valid = RIP_ENTRY_VALID;
     en->metric = rt_metric;
     en->tag = rt_tag;
-    en->from = (new->src->proto == P) ? rt_from : NULL;
+    en->from = (new->attrs->src->proto == P) ? new->u.rip.from : NULL;
     en->iface = new->attrs->nh.iface;
     en->next_hop = new->attrs->nh.gw;
   }
@@ -402,20 +369,6 @@ rip_rt_notify(struct proto *P, struct channel *ch UNUSED, struct network *net, s
   }
 }
 
-void
-rip_flush_table(struct rip_proto *p, struct rip_neighbor *n)
-{
-  btime expires = current_time() + n->ifa->cf->timeout_time;
-
-  FIB_WALK(&p->rtable, struct rip_entry, en)
-  {
-    for (struct rip_rte *e = en->routes; e; e = e->next)
-      if ((e->from == n) && (e->expires == TIME_INFINITY))
-	e->expires = expires;
-  }
-  FIB_WALK_END;
-}
-
 
 /*
  *	RIP neighbors
@@ -450,7 +403,7 @@ rip_remove_neighbor(struct rip_proto *p, struct rip_neighbor *n)
 {
   neighbor *nbr = n->nbr;
 
-  TRACE(D_EVENTS, "Removing neighbor %I on %s", nbr->addr, nbr->ifreq->name);
+  TRACE(D_EVENTS, "Removing neighbor %I on %s", nbr->addr, nbr->iface->name);
 
   rem_node(NODE n);
   n->ifa = NULL;
@@ -531,8 +484,7 @@ rip_update_bfd(struct rip_proto *p, struct rip_neighbor *n)
      */
     ip_addr saddr = rip_is_v2(p) ? n->ifa->sk->saddr : n->nbr->ifa->ip;
     n->bfd_req = bfd_request_session(p->p.pool, n->nbr->addr, saddr,
-				     n->nbr->iface, p->p.vrf,
-				     rip_bfd_notify, n, NULL);
+				     n->nbr->iface, rip_bfd_notify, n);
   }
 
   if (!use_bfd && n->bfd_req)
@@ -554,23 +506,14 @@ rip_iface_start(struct rip_iface *ifa)
 
   TRACE(D_EVENTS, "Starting interface %s", ifa->iface->name);
 
-  if (! ifa->cf->demand_circuit)
-  {
-    ifa->next_regular = current_time() + (random() % ifa->cf->update_time) + 100 MS;
-    tm_set(ifa->timer, ifa->next_regular);
-  }
-  else
-  {
-    ifa->next_regular = TIME_INFINITY;
-  }
-
+  ifa->next_regular = current_time() + (random() % ifa->cf->update_time) + 100 MS;
+  ifa->next_triggered = current_time();	/* Available immediately */
+  ifa->want_triggered = 1;		/* All routes in triggered update */
+  tm_start(ifa->timer, 100 MS);
   ifa->up = 1;
 
-  if (ifa->cf->passive)
-    return;
-
-  rip_send_request(p, ifa);
-  rip_send_table(p, ifa, ifa->addr, 0);
+  if (!ifa->cf->passive)
+    rip_send_request(ifa->rip, ifa);
 }
 
 static void
@@ -583,24 +526,10 @@ rip_iface_stop(struct rip_iface *ifa)
 
   rip_reset_tx_session(p, ifa);
 
-  ifa->next_regular = 0;
-  ifa->next_triggered = 0;
-  ifa->want_triggered = 0;
-
-  if (ifa->tx_pending)
-    ifa->tx_seqnum++;
-
-  ifa->tx_pending = 0;
-  ifa->req_pending = 0;
-
-  if (ifa->cf->demand_circuit && !ifa->cf->passive)
-    rip_send_flush(p, ifa);
-
   WALK_LIST_FIRST(n, ifa->neigh_list)
     rip_remove_neighbor(p, n);
 
   tm_stop(ifa->timer);
-  tm_stop(ifa->rxmt_timer);
   ifa->up = 0;
 }
 
@@ -714,7 +643,6 @@ rip_add_iface(struct rip_proto *p, struct iface *iface, struct rip_iface_config
   add_tail(&p->iface_list, NODE ifa);
 
   ifa->timer = tm_new_init(p->p.pool, rip_iface_timer, ifa, 0, 0);
-  ifa->rxmt_timer = tm_new_init(p->p.pool, rip_rxmt_timeout, ifa, 0, 0);
 
   struct object_lock *lock = olock_new(p->p.pool);
   lock->type = OBJLOCK_UDP;
@@ -753,8 +681,7 @@ rip_reconfigure_iface(struct rip_proto *p, struct rip_iface *ifa, struct rip_ifa
       (new->port != old->port) ||
       (new->tx_tos != old->tx_tos) ||
       (new->tx_priority != old->tx_priority) ||
-      (new->ttl_security != old->ttl_security) ||
-      (new->demand_circuit != old->demand_circuit))
+      (new->ttl_security != old->ttl_security))
     return 0;
 
   TRACE(D_EVENTS, "Reconfiguring interface %s", ifa->iface->name);
@@ -763,21 +690,9 @@ rip_reconfigure_iface(struct rip_proto *p, struct rip_iface *ifa, struct rip_ifa
 
   rip_iface_update_buffers(ifa);
 
-  if ((! ifa->cf->demand_circuit) &&
-      (ifa->next_regular > (current_time() + new->update_time)))
+  if (ifa->next_regular > (current_time() + new->update_time))
     ifa->next_regular = current_time() + (random() % new->update_time) + 100 MS;
 
-  if (ifa->up && new->demand_circuit && (new->passive != old->passive))
-  {
-    if (new->passive)
-      rip_send_flush(p, ifa);
-    else
-    {
-      rip_send_request(p, ifa);
-      rip_send_table(p, ifa, ifa->addr, 0);
-    }
-  }
-
   if (new->check_link != old->check_link)
     rip_iface_update_state(ifa);
 
@@ -970,11 +885,6 @@ rip_timer(timer *t)
 
   /* Handling neighbor expiration */
   WALK_LIST(ifa, p->iface_list)
-  {
-    /* No expiration for demand circuit ifaces */
-    if (ifa->cf->demand_circuit)
-      continue;
-
     WALK_LIST_DELSAFE(n, nn, ifa->neigh_list)
       if (n->last_seen)
       {
@@ -985,7 +895,6 @@ rip_timer(timer *t)
 	else
 	  next = MIN(next, expires);
       }
-  }
 
   tm_start(p->timer, MAX(next - now_, 100 MS));
 }
@@ -993,7 +902,7 @@ rip_timer(timer *t)
 static inline void
 rip_kick_timer(struct rip_proto *p)
 {
-  if ((p->timer->expires > (current_time() + 100 MS)))
+  if (p->timer->expires > (current_time() + 100 MS))
     tm_start(p->timer, 100 MS);
 }
 
@@ -1022,8 +931,11 @@ rip_iface_timer(timer *t)
 
   if (ifa->tx_active)
   {
-    tm_start(ifa->timer, 100 MS);
-    return;
+    if (now_ < (ifa->next_regular + period))
+    { tm_start(ifa->timer, 100 MS); return; }
+
+    /* We are too late, reset is done by rip_send_table() */
+    log(L_WARN "%s: Too slow update on %s, resetting", p->p.name, ifa->iface->name);
   }
 
   if (now_ >= ifa->next_regular)
@@ -1045,17 +957,13 @@ rip_iface_timer(timer *t)
     p->triggered = 0;
   }
 
-  if (ifa->want_triggered && (ifa->next_triggered < ifa->next_regular))
-    tm_set(ifa->timer, ifa->next_triggered);
-  else if (ifa->next_regular != TIME_INFINITY)
-    tm_set(ifa->timer, ifa->next_regular);
+  tm_start(ifa->timer, ifa->want_triggered ? (1 S) : (ifa->next_regular - now_));
 }
 
-
 static inline void
 rip_iface_kick_timer(struct rip_iface *ifa)
 {
-  if ((! tm_active(ifa->timer)) || (ifa->timer->expires > (current_time() + 100 MS)))
+  if (ifa->timer->expires > (current_time() + 100 MS))
     tm_start(ifa->timer, 100 MS);
 }
 
@@ -1079,8 +987,9 @@ rip_trigger_update(struct rip_proto *p)
     TRACE(D_EVENTS, "Scheduling triggered updates for %s", ifa->iface->name);
     ifa->want_triggered = current_time();
     rip_iface_kick_timer(ifa);
-    p->triggered = 1;
   }
+
+  p->triggered = 1;
 }
 
 
@@ -1088,6 +997,28 @@ rip_trigger_update(struct rip_proto *p)
  *	RIP protocol glue
  */
 
+static struct ea_list *
+rip_prepare_attrs(struct linpool *pool, ea_list *next, u8 metric, u16 tag)
+{
+  struct ea_list *l = lp_alloc(pool, sizeof(struct ea_list) + 2 * sizeof(eattr));
+
+  l->next = next;
+  l->flags = EALF_SORTED;
+  l->count = 2;
+
+  l->attrs[0].id = EA_RIP_METRIC;
+  l->attrs[0].flags = 0;
+  l->attrs[0].type = EAF_TYPE_INT | EAF_TEMP;
+  l->attrs[0].u.data = metric;
+
+  l->attrs[1].id = EA_RIP_TAG;
+  l->attrs[1].flags = 0;
+  l->attrs[1].type = EAF_TYPE_INT | EAF_TEMP;
+  l->attrs[1].u.data = tag;
+
+  return l;
+}
+
 static void
 rip_reload_routes(struct channel *C)
 {
@@ -1101,25 +1032,41 @@ rip_reload_routes(struct channel *C)
   rip_kick_timer(p);
 }
 
-static u32
-rip_rte_igp_metric(struct rte *rt)
+static struct ea_list *
+rip_make_tmp_attrs(struct rte *rt, struct linpool *pool)
 {
-  return ea_get_int(rt->attrs->eattrs, EA_RIP_METRIC, IGP_METRIC_UNKNOWN);
+  return rip_prepare_attrs(pool, NULL, rt->u.rip.metric, rt->u.rip.tag);
+}
+
+static void
+rip_store_tmp_attrs(struct rte *rt)
+{
+  rt->u.rip.metric = ea_get_int(rt->attrs->eattrs, EA_RIP_METRIC, 1);
+  rt->u.rip.tag = ea_get_int(rt->attrs->eattrs, EA_RIP_TAG, 0);
 }
 
 static int
 rip_rte_better(struct rte *new, struct rte *old)
 {
-  return rip_rte_igp_metric(new) < rip_rte_igp_metric(old);
+  return new->u.rip.metric < old->u.rip.metric;
 }
 
+static int
+rip_rte_same(struct rte *new, struct rte *old)
+{
+  return ((new->u.rip.metric == old->u.rip.metric) &&
+	  (new->u.rip.tag == old->u.rip.tag) &&
+	  (new->u.rip.from == old->u.rip.from));
+}
+
+
 static void
 rip_postconfig(struct proto_config *CF)
 {
   // struct rip_config *cf = (void *) CF;
 
   /* Define default channel */
-  if (! proto_cf_main_channel(CF))
+  if (EMPTY_LIST(CF->channels))
     channel_config_new(NULL, net_label[CF->net_type], CF->net_type, CF);
 }
 
@@ -1134,8 +1081,10 @@ rip_init(struct proto_config *CF)
   P->rt_notify = rip_rt_notify;
   P->neigh_notify = rip_neigh_notify;
   P->reload_routes = rip_reload_routes;
+  P->make_tmp_attrs = rip_make_tmp_attrs;
+  P->store_tmp_attrs = rip_store_tmp_attrs;
   P->rte_better = rip_rte_better;
-  P->rte_igp_metric = rip_rte_igp_metric;
+  P->rte_same = rip_rte_same;
 
   return P;
 }
@@ -1165,20 +1114,6 @@ rip_start(struct proto *P)
   return PS_UP;
 }
 
-static int
-rip_shutdown(struct proto *P)
-{
-  struct rip_proto *p = (void *) P;
-
-  TRACE(D_EVENTS, "Shutdown requested");
-
-  struct rip_iface *ifa;
-  WALK_LIST(ifa, p->iface_list)
-    rip_iface_stop(ifa);
-
-  return PS_DOWN;
-}
-
 static int
 rip_reconfigure(struct proto *P, struct proto_config *CF)
 {
@@ -1210,18 +1145,14 @@ rip_reconfigure(struct proto *P, struct proto_config *CF)
 static void
 rip_get_route_info(rte *rte, byte *buf)
 {
-  struct rip_proto *p = (struct rip_proto *) rte->src->proto;
-  u32 rt_metric = ea_get_int(rte->attrs->eattrs, EA_RIP_METRIC, p->infinity);
-  u32 rt_tag = ea_get_int(rte->attrs->eattrs, EA_RIP_TAG, 0);
+  buf += bsprintf(buf, " (%d/%d)", rte->pref, rte->u.rip.metric);
 
-  buf += bsprintf(buf, " (%d/%d)", rte->attrs->pref, rt_metric);
-
-  if (rt_tag)
-    bsprintf(buf, " [%04x]", rt_tag);
+  if (rte->u.rip.tag)
+    bsprintf(buf, " [%04x]", rte->u.rip.tag);
 }
 
 static int
-rip_get_attr(const eattr *a, byte *buf, int buflen UNUSED)
+rip_get_attr(eattr *a, byte *buf, int buflen UNUSED)
 {
   switch (a->id)
   {
@@ -1239,7 +1170,7 @@ rip_get_attr(const eattr *a, byte *buf, int buflen UNUSED)
 }
 
 void
-rip_show_interfaces(struct proto *P, const char *iff)
+rip_show_interfaces(struct proto *P, char *iff)
 {
   struct rip_proto *p = (void *) P;
   struct rip_iface *ifa = NULL;
@@ -1248,6 +1179,7 @@ rip_show_interfaces(struct proto *P, const char *iff)
   if (p->p.proto_state != PS_UP)
   {
     cli_msg(-1021, "%s: is not up", p->p.name);
+    cli_msg(0, "");
     return;
   }
 
@@ -1266,15 +1198,16 @@ rip_show_interfaces(struct proto *P, const char *iff)
 	nbrs++;
 
     btime now_ = current_time();
-    btime timer = ((ifa->next_regular < TIME_INFINITY) && (ifa->next_regular > now_)) ?
-      (ifa->next_regular - now_) : 0;
+    btime timer = (ifa->next_regular > now_) ? (ifa->next_regular - now_) : 0;
     cli_msg(-1021, "%-10s %-6s %6u %6u %7t",
 	    ifa->iface->name, (ifa->up ? "Up" : "Down"), ifa->cf->metric, nbrs, timer);
   }
+
+  cli_msg(0, "");
 }
 
 void
-rip_show_neighbors(struct proto *P, const char *iff)
+rip_show_neighbors(struct proto *P, char *iff)
 {
   struct rip_proto *p = (void *) P;
   struct rip_iface *ifa = NULL;
@@ -1283,6 +1216,7 @@ rip_show_neighbors(struct proto *P, const char *iff)
   if (p->p.proto_state != PS_UP)
   {
     cli_msg(-1022, "%s: is not up", p->p.name);
+    cli_msg(0, "");
     return;
   }
 
@@ -1305,6 +1239,8 @@ rip_show_neighbors(struct proto *P, const char *iff)
 	      n->nbr->addr, ifa->iface->name, ifa->cf->metric, n->uc, timer);
     }
   }
+
+  cli_msg(0, "");
 }
 
 static void
@@ -1318,12 +1254,8 @@ rip_dump(struct proto *P)
   FIB_WALK(&p->rtable, struct rip_entry, en)
   {
     debug("RIP: entry #%d: %N via %I dev %s valid %d metric %d age %t\n",
-	  i++, en->n.addr, en->next_hop, en->iface ? en->iface->name : "(null)",
+	  i++, en->n.addr, en->next_hop, en->iface->name,
 	  en->valid, en->metric, current_time() - en->changed);
-
-    for (struct rip_rte *e = en->routes; e; e = e->next)
-      debug("RIP:   via %I metric %d expires %t\n",
-	    e->next_hop, e->metric, e->expires - current_time());
   }
   FIB_WALK_END;
 
@@ -1349,14 +1281,7 @@ struct protocol proto_rip = {
   .init =		rip_init,
   .dump =		rip_dump,
   .start =		rip_start,
-  .shutdown =		rip_shutdown,
   .reconfigure =	rip_reconfigure,
   .get_route_info =	rip_get_route_info,
   .get_attr =		rip_get_attr
 };
-
-void
-rip_build(void)
-{
-  proto_build(&proto_rip);
-}
