@@ -121,18 +121,18 @@ rpki_table_add_roa(struct rpki_cache *cache, struct channel *channel, const net_
   struct rpki_proto *p = cache->p;
 
   rta a0 = {
-    .pref = channel->preference,
+    .src = p->p.main_source,
     .source = RTS_RPKI,
     .scope = SCOPE_UNIVERSE,
     .dest = RTD_NONE,
   };
 
   rta *a = rta_lookup(&a0);
-  rte *e = rte_get_temp(a, p->p.main_source);
+  rte *e = rte_get_temp(a);
 
   e->pflags = 0;
 
-  rte_update2(channel, &pfxr->n, e, e->src);
+  rte_update2(channel, &pfxr->n, e, a0.src);
 }
 
 void
@@ -278,13 +278,12 @@ rpki_cache_change_state(struct rpki_cache *cache, const enum rpki_cache_state ne
 
   case RPKI_CS_NO_INCR_UPDATE_AVAIL:
     /* Server was unable to answer the last Serial Query and sent Cache Reset. */
+    rpki_cache_change_state(cache, RPKI_CS_RESET);
+    break;
+
   case RPKI_CS_ERROR_NO_DATA_AVAIL:
     /* No validation records are available on the cache server. */
-
-    if (old_state == RPKI_CS_ESTABLISHED)
-      rpki_cache_change_state(cache, RPKI_CS_RESET);
-    else
-      rpki_schedule_next_retry(cache);
+    rpki_cache_change_state(cache, RPKI_CS_RESET);
     break;
 
   case RPKI_CS_ERROR_FATAL:
@@ -454,11 +453,6 @@ rpki_retry_hook(timer *tm)
     }
     break;
 
-  case RPKI_CS_NO_INCR_UPDATE_AVAIL:
-  case RPKI_CS_ERROR_NO_DATA_AVAIL:
-    rpki_cache_change_state(cache, RPKI_CS_RESET);
-    break;
-
   default:
     rpki_cache_change_state(cache, RPKI_CS_CONNECTING);
     break;
@@ -585,9 +579,7 @@ rpki_init_cache(struct rpki_proto *p, struct rpki_config *cf)
   switch (cf->tr_config.type)
   {
   case RPKI_TR_TCP: rpki_tr_tcp_init(cache->tr_sock); break;
-#if HAVE_LIBSSH
   case RPKI_TR_SSH: rpki_tr_ssh_init(cache->tr_sock); break;
-#endif
   };
 
   CACHE_DBG(cache, "Connection object created");
@@ -645,6 +637,18 @@ rpki_shutdown(struct proto *P)
  * 	RPKI Reconfiguration
  */
 
+static int
+rpki_try_fast_reconnect(struct rpki_cache *cache)
+{
+  if (cache->state == RPKI_CS_ESTABLISHED)
+  {
+    rpki_cache_change_state(cache, RPKI_CS_FAST_RECONNECT);
+    return SUCCESSFUL_RECONF;
+  }
+
+  return NEED_RESTART;
+}
+
 /**
  * rpki_reconfigure_cache - a cache reconfiguration
  * @p: RPKI protocol instance
@@ -660,7 +664,6 @@ rpki_shutdown(struct proto *P)
 static int
 rpki_reconfigure_cache(struct rpki_proto *p UNUSED, struct rpki_cache *cache, struct rpki_config *new, struct rpki_config *old)
 {
-  u8 try_reset = 0;
   u8 try_fast_reconnect = 0;
 
   if (strcmp(old->hostname, new->hostname) != 0)
@@ -680,30 +683,21 @@ rpki_reconfigure_cache(struct rpki_proto *p UNUSED, struct rpki_cache *cache, st
     CACHE_TRACE(D_EVENTS, cache, "Transport type changed");
     return NEED_RESTART;
   }
-
-  if (old->ignore_max_length != new->ignore_max_length)
-  {
-    CACHE_TRACE(D_EVENTS, cache, "Ignore max length changed");
-    try_reset = 1;
-  }
-
-#if HAVE_LIBSSH
   else if (new->tr_config.type == RPKI_TR_SSH)
   {
     struct rpki_tr_ssh_config *ssh_old = (void *) old->tr_config.spec;
     struct rpki_tr_ssh_config *ssh_new = (void *) new->tr_config.spec;
-    if (bstrcmp(ssh_old->bird_private_key, ssh_new->bird_private_key) ||
-	bstrcmp(ssh_old->cache_public_key, ssh_new->cache_public_key) ||
-	bstrcmp(ssh_old->user, ssh_new->user))
+    if ((strcmp(ssh_old->bird_private_key, ssh_new->bird_private_key) != 0) ||
+	(strcmp(ssh_old->cache_public_key, ssh_new->cache_public_key) != 0) ||
+	(strcmp(ssh_old->user, ssh_new->user) != 0))
     {
       CACHE_TRACE(D_EVENTS, cache, "Settings of SSH transport configuration changed");
       try_fast_reconnect = 1;
     }
   }
-#endif
 
 #define TEST_INTERVAL(name, Name) 						\
-    if (old->name##_interval != new->name##_interval ||				\
+    if (cache->name##_interval != new->name##_interval ||			\
 	old->keep_##name##_interval != new->keep_##name##_interval) 		\
     { 										\
       cache->name##_interval = new->name##_interval;				\
@@ -715,26 +709,8 @@ rpki_reconfigure_cache(struct rpki_proto *p UNUSED, struct rpki_cache *cache, st
   TEST_INTERVAL(expire, Expire);
 #undef TEST_INTERVAL
 
-  if (try_reset || try_fast_reconnect)
-  {
-    if (cache->state != RPKI_CS_ESTABLISHED)
-      return NEED_RESTART;
-
-    if (try_reset && !try_fast_reconnect)
-      rpki_cache_change_state(cache, RPKI_CS_RESET);
-
-    if (try_fast_reconnect)
-    {
-      if (try_reset)
-      {
-	/* Force reset during reconnect */
-	cache->request_session_id = 1;
-	cache->serial_num = 0;
-      }
-
-      rpki_cache_change_state(cache, RPKI_CS_FAST_RECONNECT);
-    }
-  }
+  if (try_fast_reconnect)
+    return rpki_try_fast_reconnect(cache);
 
   return SUCCESSFUL_RECONF;
 }
@@ -834,27 +810,14 @@ rpki_show_proto_info(struct proto *P)
   if (cache)
   {
     const char *transport_name = "---";
-    uint default_port = 0;
 
     switch (cf->tr_config.type)
     {
-#if HAVE_LIBSSH
-    case RPKI_TR_SSH:
-      transport_name = "SSHv2";
-      default_port = RPKI_SSH_PORT;
-      break;
-#endif
-    case RPKI_TR_TCP:
-      transport_name = "Unprotected over TCP";
-      default_port = RPKI_TCP_PORT;
-      break;
+    case RPKI_TR_SSH: transport_name = "SSHv2"; break;
+    case RPKI_TR_TCP: transport_name = "Unprotected over TCP"; break;
     };
 
-    cli_msg(-1006, "  Cache server:     %s", cf->hostname);
-
-    if (cf->port != default_port)
-      cli_msg(-1006, "  Cache port:       %u", cf->port);
-
+    cli_msg(-1006, "  Cache server:     %s", rpki_get_cache_ident(cache));
     cli_msg(-1006, "  Status:           %s", rpki_cache_state_to_str(cache->state));
     cli_msg(-1006, "  Transport:        %s", transport_name);
     cli_msg(-1006, "  Protocol version: %u", cache->version);
@@ -924,11 +887,9 @@ rpki_check_config(struct rpki_config *cf)
     /* Set default port numbers */
     switch (cf->tr_config.type)
     {
-#if HAVE_LIBSSH
     case RPKI_TR_SSH:
       cf->port = RPKI_SSH_PORT;
       break;
-#endif
     default:
       cf->port = RPKI_TCP_PORT;
     }
@@ -940,7 +901,7 @@ rpki_postconfig(struct proto_config *CF)
 {
   /* Define default channel */
   if (EMPTY_LIST(CF->channels))
-    cf_error("Channel not specified");
+    channel_config_new(NULL, net_label[CF->net_type], CF->net_type, CF);
 }
 
 static void
@@ -966,9 +927,3 @@ struct protocol proto_rpki = {
   .reconfigure = 	rpki_reconfigure,
   .get_status = 	rpki_get_status,
 };
-
-void
-rpki_build(void)
-{
-  proto_build(&proto_rpki);
-}
