@@ -49,18 +49,14 @@
 
 static linpool *static_lp;
 
-static inline struct rte_src * static_get_source(struct static_proto *p, uint i)
-{ return i ? rt_get_source(&p->p, i) : p->p.main_source; }
-
 static void
 static_announce_rte(struct static_proto *p, struct static_route *r)
 {
   rta *a = allocz(RTA_MAX_SIZE);
-  struct rte_src *src = static_get_source(p, r->index);
+  a->src = p->p.main_source;
   a->source = RTS_STATIC;
   a->scope = SCOPE_UNIVERSE;
   a->dest = r->dest;
-  a->pref = p->p.main_channel->preference;
 
   if (r->dest == RTD_UNICAST)
   {
@@ -103,24 +99,13 @@ static_announce_rte(struct static_proto *p, struct static_route *r)
     return;
 
   /* We skip rta_lookup() here */
-  rte *e = rte_get_temp(a, src);
+  rte *e = rte_get_temp(a);
   e->pflags = 0;
 
   if (r->cmds)
-  {
-    /* Create a temporary table node */
-    e->net = alloca(sizeof(net) + r->net->length);
-    memset(e->net, 0, sizeof(net) + r->net->length);
-    net_copy(e->net->n.addr, r->net);
-
-    /* Evaluate the filter */
     f_eval_rte(r->cmds, &e, static_lp);
 
-    /* Remove the temporary node */
-    e->net = NULL;
-  }
-
-  rte_update2(p->p.main_channel, r->net, e, src);
+  rte_update(&p->p, r->net, e);
   r->state = SRS_CLEAN;
 
   if (r->cmds)
@@ -132,7 +117,7 @@ withdraw:
   if (r->state == SRS_DOWN)
     return;
 
-  rte_update2(p->p.main_channel, r->net, NULL, src);
+  rte_update(&p->p, r->net, NULL);
   r->state = SRS_DOWN;
 }
 
@@ -149,48 +134,15 @@ static_mark_rte(struct static_proto *p, struct static_route *r)
     ev_schedule(p->event);
 }
 
-static void
-static_mark_all(struct static_proto *p)
-{
-  struct static_config *cf = (void *) p->p.cf;
-  struct static_route *r;
-
-  /* We want to reload all routes, mark them as dirty */
-
-  WALK_LIST(r, cf->routes)
-    if (r->state == SRS_CLEAN)
-      r->state = SRS_DIRTY;
-
-  p->marked_all = 1;
-  BUFFER_FLUSH(p->marked);
-
-  if (!ev_active(p->event))
-    ev_schedule(p->event);
-}
-
-
 static void
 static_announce_marked(void *P)
 {
   struct static_proto *p = P;
-  struct static_config *cf = (void *) p->p.cf;
-  struct static_route *r;
 
-  if (p->marked_all)
-  {
-    WALK_LIST(r, cf->routes)
-      if (r->state == SRS_DIRTY)
-	static_announce_rte(p, r);
+  BUFFER_WALK(p->marked, r)
+    static_announce_rte(P, r);
 
-    p->marked_all = 0;
-  }
-  else
-  {
-    BUFFER_WALK(p->marked, r)
-      static_announce_rte(p, r);
-
-    BUFFER_FLUSH(p->marked);
-  }
+  BUFFER_FLUSH(p->marked);
 }
 
 static void
@@ -207,9 +159,8 @@ static_update_bfd(struct static_proto *p, struct static_route *r)
   if (bfd_up && !r->bfd_req)
   {
     // ip_addr local = ipa_nonzero(r->local) ? r->local : nb->ifa->ip;
-    r->bfd_req = bfd_request_session(p->p.pool, r->via, nb->ifa->ip,
-				     nb->iface, p->p.vrf,
-				     static_bfd_notify, r, NULL);
+    r->bfd_req = bfd_request_session(p->p.pool, r->via, nb->ifa->ip, nb->iface,
+				     static_bfd_notify, r);
   }
 
   if (!bfd_up && r->bfd_req)
@@ -298,7 +249,7 @@ static void
 static_remove_rte(struct static_proto *p, struct static_route *r)
 {
   if (r->state)
-    rte_update2(p->p.main_channel, r->net, NULL, static_get_source(p, r->index));
+    rte_update(&p->p, r->net, NULL);
 
   static_reset_rte(p, r);
 }
@@ -357,7 +308,7 @@ static inline int
 static_same_rte(struct static_route *or, struct static_route *nr)
 {
   /* Note that i_same() requires arguments in (new, old) order */
-  return static_same_dest(or, nr) && f_same(nr->cmds, or->cmds);
+  return static_same_dest(or, nr) && i_same(nr->cmds, or->cmds);
 }
 
 static void
@@ -401,33 +352,12 @@ static_bfd_notify(struct bfd_request *req)
     static_mark_rte(p, r->mp_head);
 }
 
-static void
-static_reload_routes(struct channel *C)
-{
-  struct static_proto *p = (void *) C->proto;
-
-  TRACE(D_EVENTS, "Scheduling route reload");
-
-  static_mark_all(p);
-}
-
 static int
-static_rte_better(rte *new, rte *old)
+static_rte_mergable(rte *pri UNUSED, rte *sec UNUSED)
 {
-  u32 n = ea_get_int(new->attrs->eattrs, EA_GEN_IGP_METRIC, IGP_METRIC_UNKNOWN);
-  u32 o = ea_get_int(old->attrs->eattrs, EA_GEN_IGP_METRIC, IGP_METRIC_UNKNOWN);
-  return n < o;
+  return 1;
 }
 
-static int
-static_rte_mergable(rte *pri, rte *sec)
-{
-  u32 a = ea_get_int(pri->attrs->eattrs, EA_GEN_IGP_METRIC, IGP_METRIC_UNKNOWN);
-  u32 b = ea_get_int(sec->attrs->eattrs, EA_GEN_IGP_METRIC, IGP_METRIC_UNKNOWN);
-  return a == b;
-}
-
-static void static_index_routes(struct static_config *cf);
 
 static void
 static_postconfig(struct proto_config *CF)
@@ -435,7 +365,7 @@ static_postconfig(struct proto_config *CF)
   struct static_config *cf = (void *) CF;
   struct static_route *r;
 
-  if (! proto_cf_main_channel(CF))
+  if (EMPTY_LIST(CF->channels))
     cf_error("Channel not specified");
 
   struct channel_config *cc = proto_cf_main_channel(CF);
@@ -451,8 +381,6 @@ static_postconfig(struct proto_config *CF)
   WALK_LIST(r, cf->routes)
     if (r->net && (r->net->type != CF->net_type))
       cf_error("Route %N incompatible with channel type", r->net);
-
-  static_index_routes(cf);
 }
 
 static struct proto *
@@ -465,8 +393,6 @@ static_init(struct proto_config *CF)
   P->main_channel = proto_add_channel(P, proto_cf_main_channel(CF));
 
   P->neigh_notify = static_neigh_notify;
-  P->reload_routes = static_reload_routes;
-  P->rte_better = static_rte_better;
   P->rte_mergable = static_rte_mergable;
 
   if (cf->igp_table_ip4)
@@ -486,7 +412,7 @@ static_start(struct proto *P)
   struct static_route *r;
 
   if (!static_lp)
-    static_lp = lp_new(&root_pool);
+    static_lp = lp_new(&root_pool, LP_GOOD_SIZE(1024));
 
   if (p->igp_table_ip4)
     rt_lock_table(p->igp_table_ip4);
@@ -536,7 +462,7 @@ static_cleanup(struct proto *P)
 static void
 static_dump_rte(struct static_route *r)
 {
-  debug("%-1N (%u): ", r->net, r->index);
+  debug("%-1N: ", r->net);
   if (r->dest == RTD_UNICAST)
     if (r->iface && ipa_zero(r->via))
       debug("dev %s\n", r->iface->name);
@@ -559,48 +485,11 @@ static_dump(struct proto *P)
 
 #define IGP_TABLE(cf, sym) ((cf)->igp_table_##sym ? (cf)->igp_table_##sym ->table : NULL )
 
-static inline int srt_equal(const struct static_route *a, const struct static_route *b)
-{ return net_equal(a->net, b->net) && (a->index == b->index); }
-
-static inline int srt_compare(const struct static_route *a, const struct static_route *b)
-{ return net_compare(a->net, b->net) ?: uint_cmp(a->index, b->index); }
-
-static inline int srt_compare_qsort(const void *A, const void *B)
+static inline int
+static_cmp_rte(const void *X, const void *Y)
 {
-  return srt_compare(*(const struct static_route * const *)A,
-		     *(const struct static_route * const *)B);
-}
-
-static void
-static_index_routes(struct static_config *cf)
-{
-  struct static_route *rt, **buf;
-  uint num, i, v;
-
-  num = list_length(&cf->routes);
-  buf = xmalloc(num * sizeof(void *));
-
-  /* Initialize with sequential indexes to ensure stable sorting */
-  i = 0;
-  WALK_LIST(rt, cf->routes)
-  {
-    buf[i] = rt;
-    rt->index = i++;
-  }
-
-  qsort(buf, num, sizeof(struct static_route *), srt_compare_qsort);
-
-  /* Compute proper indexes - sequential for routes with same network */
-  for (i = 0, v = 0, rt = NULL; i < num; i++, v++)
-  {
-    if (rt && !net_equal(buf[i]->net, rt->net))
-      v = 0;
-
-    rt = buf[i];
-    rt->index = v;
-  }
-
-  xfree(buf);
+  struct static_route *x = *(void **)X, *y = *(void **)Y;
+  return net_compare(x->net, y->net);
 }
 
 static int
@@ -629,7 +518,7 @@ static_reconfigure(struct proto *P, struct proto_config *CF)
 
   /* Reconfigure initial matching sequence */
   for (or = HEAD(o->routes), nr = HEAD(n->routes);
-       NODE_VALID(or) && NODE_VALID(nr) && srt_equal(or, nr);
+       NODE_VALID(or) && NODE_VALID(nr) && net_equal(or->net, nr->net);
        or = NODE_NEXT(or), nr = NODE_NEXT(nr))
     static_reconfigure_rte(p, or, nr);
 
@@ -655,12 +544,12 @@ static_reconfigure(struct proto *P, struct proto_config *CF)
   for (i = 0, nr2 = nr; i < nrnum; i++, nr2 = NODE_NEXT(nr2))
     nrbuf[i] = nr2;
 
-  qsort(orbuf, ornum, sizeof(struct static_route *), srt_compare_qsort);
-  qsort(nrbuf, nrnum, sizeof(struct static_route *), srt_compare_qsort);
+  qsort(orbuf, ornum, sizeof(struct static_route *), static_cmp_rte);
+  qsort(nrbuf, nrnum, sizeof(struct static_route *), static_cmp_rte);
 
   while ((orpos < ornum) && (nrpos < nrnum))
   {
-    int x = srt_compare(orbuf[orpos], nrbuf[nrpos]);
+    int x = net_compare(orbuf[orpos]->net, nrbuf[nrpos]->net);
     if (x < 0)
       static_remove_rte(p, orbuf[orpos++]);
     else if (x > 0)
@@ -678,10 +567,6 @@ static_reconfigure(struct proto *P, struct proto_config *CF)
   xfree(orbuf);
   xfree(nrbuf);
 
-  /* All dirty routes were announced anyways */
-  BUFFER_FLUSH(p->marked);
-  p->marked_all = 0;
-
   return 1;
 }
 
@@ -703,7 +588,6 @@ static_copy_config(struct proto_config *dest, struct proto_config *src)
     {
       dnh = cfg_alloc(sizeof(struct static_route));
       memcpy(dnh, snh, sizeof(struct static_route));
-      memset(&dnh->n, 0, sizeof(node));
 
       if (!drt)
 	add_tail(&d->routes, &(dnh->n));
@@ -717,16 +601,6 @@ static_copy_config(struct proto_config *dest, struct proto_config *src)
   }
 }
 
-static void
-static_get_route_info(rte *rte, byte *buf)
-{
-  eattr *a = ea_find(rte->attrs->eattrs, EA_GEN_IGP_METRIC);
-  if (a)
-    buf += bsprintf(buf, " (%d/%u)", rte->attrs->pref, a->u.data);
-  else
-    buf += bsprintf(buf, " (%d)", rte->attrs->pref);
-}
-
 static void
 static_show_rt(struct static_route *r)
 {
@@ -772,6 +646,7 @@ static_show(struct proto *P)
 
   WALK_LIST(r, c->routes)
     static_show_rt(r);
+  cli_msg(0, "");
 }
 
 
@@ -790,12 +665,5 @@ struct protocol proto_static = {
   .shutdown =		static_shutdown,
   .cleanup =		static_cleanup,
   .reconfigure =	static_reconfigure,
-  .copy_config =	static_copy_config,
-  .get_route_info =	static_get_route_info,
+  .copy_config =	static_copy_config
 };
-
-void
-static_build(void)
-{
-  proto_build(&proto_static);
-}
