@@ -6,6 +6,7 @@
  *	Can be freely distributed and used under the terms of the GNU GPL.
  */
 
+#include <alloca.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <fcntl.h>
@@ -27,16 +28,10 @@
 #include "lib/hash.h"
 #include "conf/conf.h"
 
-#ifdef CONFIG_LINUX_NETLINK
 #include <asm/types.h>
 #include <linux/if.h>
 #include <linux/netlink.h>
 #include <linux/rtnetlink.h>
-#endif
-#ifdef CONFIG_FREEBSD_NETLINK
-#include <netlink/netlink.h>
-#include <netlink/netlink_route.h>
-#endif
 
 #ifdef HAVE_MPLS_KERNEL
 #include <linux/lwtunnel.h>
@@ -74,21 +69,49 @@
 #define RTA_ENCAP  22
 #endif
 
-#ifndef NETLINK_GET_STRICT_CHK
-#define NETLINK_GET_STRICT_CHK 12
-#endif
-
-#define krt_ipv4(p) ((p)->af == AF_INET)
+#define krt_ecmp6(p) ((p)->af == AF_INET6)
 
 const int rt_default_ecmp = 16;
 
+/*
+ * Structure nl_parse_state keeps state of received route processing. Ideally,
+ * we could just independently parse received Netlink messages and immediately
+ * propagate received routes to the rest of BIRD, but older Linux kernel (before
+ * version 4.11) represents and announces IPv6 ECMP routes not as one route with
+ * multiple next hops (like RTA_MULTIPATH in IPv4 ECMP), but as a sequence of
+ * routes with the same prefix. More recent kernels work as with IPv4.
+ *
+ * Therefore, BIRD keeps currently processed route in nl_parse_state structure
+ * and postpones its propagation until we expect it to be final; i.e., when
+ * non-matching route is received or when the scan ends. When another matching
+ * route is received, it is merged with the already processed route to form an
+ * ECMP route. Note that merging is done only for IPv6 (merge == 1), but the
+ * postponing is done in both cases (for simplicity). All IPv4 routes or IPv6
+ * routes with RTA_MULTIPATH set are just considered non-matching.
+ *
+ * This is ignored for asynchronous notifications (every notification is handled
+ * as a separate route). It is not an issue for our routes, as we ignore such
+ * notifications anyways. But importing alien IPv6 ECMP routes does not work
+ * properly with older kernels.
+ *
+ * Whatever the kernel version is, IPv6 ECMP routes are sent as multiple routes
+ * for the same prefix.
+ */
+
 struct nl_parse_state
 {
-  struct krt_proto *proto;
   struct linpool *pool;
   int scan;
+  int merge;
 
-  u32 rta_flow;
+  net *net;
+  rta *attrs;
+  struct krt_proto *proto;
+  s8 new;
+  s8 krt_src;
+  u8 krt_type;
+  u8 krt_proto;
+  u32 krt_metric;
 };
 
 /*
@@ -104,7 +127,7 @@ struct nl_sock
   uint last_size;
 };
 
-#define NL_RX_SIZE 32768
+#define NL_RX_SIZE 8192
 
 #define NL_OP_DELETE	0
 #define NL_OP_ADD	(NLM_F_CREATE|NLM_F_EXCL)
@@ -131,51 +154,11 @@ nl_open_sock(struct nl_sock *nl)
     }
 }
 
-static int
-nl_set_strict_dump(struct nl_sock *nl UNUSED, int strict UNUSED)
-{
-#ifdef SOL_NETLINK
-  return setsockopt(nl->fd, SOL_NETLINK, NETLINK_GET_STRICT_CHK, &strict, sizeof(strict));
-#else
-  return -1;
-#endif
-}
-
-static void
-nl_set_rcvbuf(int fd, uint val)
-{
-  if (setsockopt(fd, SOL_SOCKET, SO_RCVBUFFORCE, &val, sizeof(val)) < 0)
-    log(L_WARN "KRT: Cannot set netlink rx buffer size to %u: %m", val);
-}
-
-static uint
-nl_cfg_rx_buffer_size(struct config *cfg)
-{
-  uint bufsize = 0;
-
-  struct proto_config *pc;
-  WALK_LIST(pc, cfg->protos)
-    if ((pc->protocol == &proto_unix_kernel) && !pc->disabled)
-      bufsize = MAX(bufsize, ((struct krt_config *) pc)->sys.netlink_rx_buffer);
-
-  return bufsize;
-}
-
-
 static void
 nl_open(void)
 {
-  if ((nl_scan.fd >= 0) && (nl_req.fd >= 0))
-    return;
-
   nl_open_sock(&nl_scan);
   nl_open_sock(&nl_req);
-
-  if (nl_set_strict_dump(&nl_scan, 1) < 0)
-  {
-    log(L_WARN "KRT: Netlink strict checking failed, will scan all tables at once");
-    krt_use_shared_scan();
-  }
 }
 
 static void
@@ -187,79 +170,26 @@ nl_send(struct nl_sock *nl, struct nlmsghdr *nh)
   sa.nl_family = AF_NETLINK;
   nh->nlmsg_pid = 0;
   nh->nlmsg_seq = ++(nl->seq);
-  nh->nlmsg_len = NLMSG_ALIGN(nh->nlmsg_len);
   if (sendto(nl->fd, nh, nh->nlmsg_len, 0, (struct sockaddr *)&sa, sizeof(sa)) < 0)
     die("rtnetlink sendto: %m");
   nl->last_hdr = NULL;
 }
 
 static void
-nl_request_dump_link(void)
+nl_request_dump(int af, int cmd)
 {
   struct {
     struct nlmsghdr nh;
-    struct ifinfomsg ifi;
+    struct rtgenmsg g;
   } req = {
-    .nh.nlmsg_type = RTM_GETLINK,
-    .nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifinfomsg)),
+    .nh.nlmsg_type = cmd,
+    .nh.nlmsg_len = sizeof(req),
     .nh.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP,
-    .nh.nlmsg_seq = ++(nl_scan.seq),
-    .ifi.ifi_family = AF_UNSPEC,
+    .g.rtgen_family = af
   };
-
-  send(nl_scan.fd, &req, sizeof(req), 0);
-  nl_scan.last_hdr = NULL;
+  nl_send(&nl_scan, &req.nh);
 }
 
-static void
-nl_request_dump_addr(int af)
-{
-  struct {
-    struct nlmsghdr nh;
-    struct ifaddrmsg ifa;
-  } req = {
-    .nh.nlmsg_type = RTM_GETADDR,
-    .nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifaddrmsg)),
-    .nh.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP,
-    .nh.nlmsg_seq = ++(nl_scan.seq),
-    .ifa.ifa_family = af,
-  };
-
-  send(nl_scan.fd, &req, sizeof(req), 0);
-  nl_scan.last_hdr = NULL;
-}
-
-static void
-nl_request_dump_route(int af, int table_id)
-{
-  struct {
-    struct nlmsghdr nh;
-    struct rtmsg rtm;
-    struct rtattr rta;
-    u32 table_id;
-  } req = {
-    .nh.nlmsg_type = RTM_GETROUTE,
-    .nh.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg)),
-    .nh.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP,
-    .nh.nlmsg_seq = ++(nl_scan.seq),
-    .rtm.rtm_family = af,
-  };
-
-  if (table_id < 256)
-    req.rtm.rtm_table = table_id;
-  else
-  {
-    req.rta.rta_type = RTA_TABLE;
-    req.rta.rta_len = RTA_LENGTH(4);
-    req.table_id = table_id;
-    req.nh.nlmsg_len = NLMSG_ALIGN(req.nh.nlmsg_len) + req.rta.rta_len;
-  }
-
-  send(nl_scan.fd, &req, req.nh.nlmsg_len, 0);
-  nl_scan.last_hdr = NULL;
-}
-
-
 static struct nlmsghdr *
 nl_get_reply(struct nl_sock *nl)
 {
@@ -412,26 +342,17 @@ static struct nl_want_attrs ifa_attr_want6[BIRD_IFA_MAX] = {
 
 static struct nl_want_attrs nexthop_attr_want4[BIRD_RTA_MAX] = {
   [RTA_GATEWAY]	  = { 1, 1, sizeof(ip4_addr) },
-  [RTA_VIA]	  = { 1, 0, 0 },
-  [RTA_FLOW]	  = { 1, 1, sizeof(u32) },
   [RTA_ENCAP_TYPE]= { 1, 1, sizeof(u16) },
   [RTA_ENCAP]	  = { 1, 0, 0 },
 };
 
 static struct nl_want_attrs nexthop_attr_want6[BIRD_RTA_MAX] = {
   [RTA_GATEWAY]	  = { 1, 1, sizeof(ip6_addr) },
-  [RTA_VIA]	  = { 1, 0, 0 },
-  [RTA_FLOW]	  = { 1, 1, sizeof(u32) },
   [RTA_ENCAP_TYPE]= { 1, 1, sizeof(u16) },
   [RTA_ENCAP]	  = { 1, 0, 0 },
 };
 
 #ifdef HAVE_MPLS_KERNEL
-static struct nl_want_attrs nexthop_attr_want_mpls[BIRD_RTA_MAX] = {
-  [RTA_VIA]	  = { 1, 0, 0 },
-  [RTA_NEWDST]	  = { 1, 0, 0 },
-};
-
 static struct nl_want_attrs encap_mpls_want[BIRD_RTA_MAX] = {
   [RTA_DST]       = { 1, 0, 0 },
 };
@@ -447,7 +368,6 @@ static struct nl_want_attrs rtm_attr_want4[BIRD_RTA_MAX] = {
   [RTA_MULTIPATH] = { 1, 0, 0 },
   [RTA_FLOW]	  = { 1, 1, sizeof(u32) },
   [RTA_TABLE]	  = { 1, 1, sizeof(u32) },
-  [RTA_VIA]	  = { 1, 0, 0 },
   [RTA_ENCAP_TYPE]= { 1, 1, sizeof(u16) },
   [RTA_ENCAP]	  = { 1, 0, 0 },
 };
@@ -464,7 +384,6 @@ static struct nl_want_attrs rtm_attr_want6[BIRD_RTA_MAX] = {
   [RTA_MULTIPATH] = { 1, 0, 0 },
   [RTA_FLOW]	  = { 1, 1, sizeof(u32) },
   [RTA_TABLE]	  = { 1, 1, sizeof(u32) },
-  [RTA_VIA]	  = { 1, 0, 0 },
   [RTA_ENCAP_TYPE]= { 1, 1, sizeof(u16) },
   [RTA_ENCAP]	  = { 1, 0, 0 },
 };
@@ -476,7 +395,6 @@ static struct nl_want_attrs rtm_attr_want_mpls[BIRD_RTA_MAX] = {
   [RTA_OIF]	  = { 1, 1, sizeof(u32) },
   [RTA_PRIORITY]  = { 1, 1, sizeof(u32) },
   [RTA_METRICS]	  = { 1, 0, 0 },
-  [RTA_MULTIPATH] = { 1, 0, 0 },
   [RTA_FLOW]	  = { 1, 1, sizeof(u32) },
   [RTA_TABLE]	  = { 1, 1, sizeof(u32) },
   [RTA_VIA]	  = { 1, 0, 0 },
@@ -548,21 +466,10 @@ static inline ip_addr rta_get_via(struct rtattr *a)
 static u32 rta_mpls_stack[MPLS_MAX_LABEL_STACK];
 static inline int rta_get_mpls(struct rtattr *a, u32 *stack)
 {
-  if (!a)
-    return 0;
-
   if (RTA_PAYLOAD(a) % 4)
     log(L_WARN "KRT: Strange length of received MPLS stack: %u", RTA_PAYLOAD(a));
 
-  int labels = mpls_get(RTA_DATA(a), RTA_PAYLOAD(a) & ~0x3, stack);
-
-  if (labels < 0)
-  {
-    log(L_WARN "KRT: Too long MPLS stack received, ignoring");
-    labels = 0;
-  }
-
-  return labels;
+  return mpls_get(RTA_DATA(a), RTA_PAYLOAD(a) & ~0x3, stack);
 }
 #endif
 
@@ -703,12 +610,10 @@ nl_add_nexthop(struct nlmsghdr *h, uint bufsize, struct nexthop *nh, int af UNUS
       nl_add_attr_mpls_encap(h, bufsize, nh->labels, nh->label);
 
   if (ipa_nonzero(nh->gw))
-  {
-    if (af == (ipa_is_ip4(nh->gw) ? AF_INET : AF_INET6))
-      nl_add_attr_ipa(h, bufsize, RTA_GATEWAY, nh->gw);
-    else
+    if (af == AF_MPLS)
       nl_add_attr_via(h, bufsize, nh->gw);
-  }
+    else
+      nl_add_attr_ipa(h, bufsize, RTA_GATEWAY, nh->gw);
 #else
 
   if (ipa_nonzero(nh->gw))
@@ -717,10 +622,9 @@ nl_add_nexthop(struct nlmsghdr *h, uint bufsize, struct nexthop *nh, int af UNUS
 }
 
 static void
-nl_add_multipath(struct nlmsghdr *h, uint bufsize, struct nexthop *nh, int af, ea_list *eattrs)
+nl_add_multipath(struct nlmsghdr *h, uint bufsize, struct nexthop *nh, int af)
 {
   struct rtattr *a = nl_open_attr(h, bufsize, RTA_MULTIPATH);
-  eattr *flow = ea_find(eattrs, EA_KRT_REALM);
 
   for (; nh; nh = nh->next)
   {
@@ -735,11 +639,6 @@ nl_add_multipath(struct nlmsghdr *h, uint bufsize, struct nexthop *nh, int af, e
     if (nh->flags & RNF_ONLINK)
       rtnh->rtnh_flags |= RTNH_F_ONLINK;
 
-    /* Our KRT_REALM is per-route, but kernel RTA_FLOW is per-nexthop.
-       Therefore, we need to attach the same attribute to each nexthop. */
-    if (flow)
-      nl_add_attr_u32(h, bufsize, RTA_FLOW, flow->u.data);
-
     nl_close_nexthop(h, rtnh);
   }
 
@@ -747,7 +646,7 @@ nl_add_multipath(struct nlmsghdr *h, uint bufsize, struct nexthop *nh, int af, e
 }
 
 static struct nexthop *
-nl_parse_multipath(struct nl_parse_state *s, struct krt_proto *p, const net_addr *n, struct rtattr *ra, int af, int krt_src)
+nl_parse_multipath(struct nl_parse_state *s, struct krt_proto *p, struct rtattr *ra, int af)
 {
   struct rtattr *a[BIRD_RTA_MAX];
   struct rtnexthop *nh = RTA_DATA(ra);
@@ -761,10 +660,7 @@ nl_parse_multipath(struct nl_parse_state *s, struct krt_proto *p, const net_addr
     {
       /* Use RTNH_OK(nh,len) ?? */
       if ((len < sizeof(*nh)) || (len < nh->rtnh_len))
-	goto err;
-
-      if ((nh->rtnh_flags & RTNH_F_DEAD) && (krt_src != KRT_SRC_BIRD))
-	goto next;
+	return NULL;
 
       *last = rv = lp_allocz(s->pool, NEXTHOP_MAX_SIZE);
       last = &(rv->next);
@@ -772,10 +668,7 @@ nl_parse_multipath(struct nl_parse_state *s, struct krt_proto *p, const net_addr
       rv->weight = nh->rtnh_hops;
       rv->iface = if_find_by_index(nh->rtnh_ifindex);
       if (!rv->iface)
-	{
-	  log(L_ERR "KRT: Received route %N with unknown ifindex %u", n, nh->rtnh_ifindex);
-	  return NULL;
-	}
+	return NULL;
 
       /* Nonexistent RTNH_PAYLOAD ?? */
       nl_attr_len = nh->rtnh_len - RTNH_LENGTH(0);
@@ -783,62 +676,39 @@ nl_parse_multipath(struct nl_parse_state *s, struct krt_proto *p, const net_addr
         {
 	case AF_INET:
 	  if (!nl_parse_attrs(RTNH_DATA(nh), nexthop_attr_want4, a, sizeof(a)))
-	    goto err;
+	    return NULL;
 	  break;
 
 	case AF_INET6:
 	  if (!nl_parse_attrs(RTNH_DATA(nh), nexthop_attr_want6, a, sizeof(a)))
-	    goto err;
+	    return NULL;
 	  break;
 
-#ifdef HAVE_MPLS_KERNEL
-	case AF_MPLS:
-	  if (!nl_parse_attrs(RTNH_DATA(nh), nexthop_attr_want_mpls, a, sizeof(a)))
-	    goto err;
-
-	  if (a[RTA_NEWDST])
-	    rv->labels = rta_get_mpls(a[RTA_NEWDST], rv->label);
-
-	  break;
-#endif
-
 	default:
-	  goto err;
+	  return NULL;
 	}
 
       if (a[RTA_GATEWAY])
-	rv->gw = rta_get_ipa(a[RTA_GATEWAY]);
-
-      if (a[RTA_FLOW])
-	s->rta_flow = rta_get_u32(a[RTA_FLOW]);
-
-#ifdef HAVE_MPLS_KERNEL
-      if (a[RTA_VIA])
-	rv->gw = rta_get_via(a[RTA_VIA]);
-#endif
-
-      if (nh->rtnh_flags & RTNH_F_ONLINK)
-	rv->flags |= RNF_ONLINK;
-
-      if (ipa_nonzero(rv->gw))
 	{
+	  rv->gw = rta_get_ipa(a[RTA_GATEWAY]);
+
+	  if (nh->rtnh_flags & RTNH_F_ONLINK)
+	    rv->flags |= RNF_ONLINK;
+
 	  neighbor *nbr;
 	  nbr = neigh_find(&p->p, rv->gw, rv->iface,
 			   (rv->flags & RNF_ONLINK) ? NEF_ONLINK : 0);
 	  if (!nbr || (nbr->scope == SCOPE_HOST))
-	    {
-	        log(L_ERR "KRT: Received route %N with strange next-hop %I", n, rv->gw);
-	        return NULL;
-	    }
+	    return NULL;
 	}
+      else
+	rv->gw = IPA_NONE;
 
 #ifdef HAVE_MPLS_KERNEL
-      if (a[RTA_ENCAP] && a[RTA_ENCAP_TYPE])
+      if (a[RTA_ENCAP_TYPE])
       {
-	if (rta_get_u16(a[RTA_ENCAP_TYPE]) != LWTUNNEL_ENCAP_MPLS)
-	{
-	  log(L_WARN "KRT: Received route %N with unknown encapsulation method %d",
-	      n, rta_get_u16(a[RTA_ENCAP_TYPE]));
+	if (rta_get_u16(a[RTA_ENCAP_TYPE]) != LWTUNNEL_ENCAP_MPLS) {
+	  log(L_WARN "KRT: Unknown encapsulation method %d in multipath", rta_get_u16(a[RTA_ENCAP_TYPE]));
 	  return NULL;
 	}
 
@@ -846,23 +716,16 @@ nl_parse_multipath(struct nl_parse_state *s, struct krt_proto *p, const net_addr
 	nl_attr_len = RTA_PAYLOAD(a[RTA_ENCAP]);
 	nl_parse_attrs(RTA_DATA(a[RTA_ENCAP]), encap_mpls_want, enca, sizeof(enca));
 	rv->labels = rta_get_mpls(enca[RTA_DST], rv->label);
+	break;
       }
 #endif
 
-    next:
+
       len -= NLMSG_ALIGN(nh->rtnh_len);
       nh = RTNH_NEXT(nh);
     }
 
-  /* Ensure nexthops are sorted to satisfy nest invariant */
-  if (!nexthop_is_sorted(first))
-    first = nexthop_sort(first);
-
   return first;
-
-err:
-  log(L_ERR "KRT: Received strange multipath route %N", n);
-  return NULL;
 }
 
 static void
@@ -1064,15 +927,18 @@ nl_parse_addr4(struct ifaddrmsg *i, int scan, int new)
       if (i->ifa_prefixlen == IP4_MAX_PREFIX_LENGTH - 2)
 	ifa.opposite = ipa_opposite_m2(ifa.ip);
 
-      if (ifi->flags & IF_BROADCAST)
-        {
-	  /* If kernel offers us a broadcast address, we trust it */
-	  if (a[IFA_BROADCAST])
-	    ifa.brd = ipa_from_ip4(rta_get_ip4(a[IFA_BROADCAST]));
-	  /* Otherwise we create one (except for /31) */
-	  else if (i->ifa_prefixlen < (IP4_MAX_PREFIX_LENGTH - 1))
-	    ifa.brd = ipa_from_ip4(ip4_or(ipa_to_ip4(ifa.ip),
-					  ip4_not(ip4_mkmask(i->ifa_prefixlen))));
+      if ((ifi->flags & IF_BROADCAST) && a[IFA_BROADCAST])
+	{
+	  ip4_addr xbrd = rta_get_ip4(a[IFA_BROADCAST]);
+	  ip4_addr ybrd = ip4_or(ipa_to_ip4(ifa.ip), ip4_not(ip4_mkmask(i->ifa_prefixlen)));
+
+	  if (ip4_equal(xbrd, net4_prefix(&ifa.prefix)) || ip4_equal(xbrd, ybrd))
+	    ifa.brd = ipa_from_ip4(xbrd);
+	  else if (ifi->flags & IF_TMP_DOWN) /* Complain only during the first scan */
+	    {
+	      log(L_ERR "KIF: Invalid broadcast address %I4 for %s", xbrd, ifi->name);
+	      ifa.brd = ipa_from_ip4(ybrd);
+	    }
 	}
     }
 
@@ -1217,7 +1083,7 @@ kif_do_scan(struct kif_proto *p UNUSED)
 
   if_start_update();
 
-  nl_request_dump_link();
+  nl_request_dump(AF_UNSPEC, RTM_GETLINK);
   while (h = nl_get_scan())
     if (h->nlmsg_type == RTM_NEWLINK || h->nlmsg_type == RTM_DELLINK)
       nl_parse_link(h, 1);
@@ -1244,14 +1110,14 @@ kif_do_scan(struct kif_proto *p UNUSED)
       }
     }
 
-  nl_request_dump_addr(AF_INET);
+  nl_request_dump(AF_INET, RTM_GETADDR);
   while (h = nl_get_scan())
     if (h->nlmsg_type == RTM_NEWADDR || h->nlmsg_type == RTM_DELADDR)
       nl_parse_addr(h, 1);
     else
       log(L_DEBUG "nl_scan_ifaces: Unknown packet received (type=%d)", h->nlmsg_type);
 
-  nl_request_dump_addr(AF_INET6);
+  nl_request_dump(AF_INET6, RTM_GETADDR);
   while (h = nl_get_scan())
     if (h->nlmsg_type == RTM_NEWADDR || h->nlmsg_type == RTM_DELADDR)
       nl_parse_addr(h, 1);
@@ -1311,7 +1177,7 @@ nh_bufsize(struct nexthop *nh)
 }
 
 static int
-nl_send_route(struct krt_proto *p, rte *e, int op)
+nl_send_route(struct krt_proto *p, rte *e, int op, int dest, struct nexthop *nh)
 {
   eattr *ea;
   net *net = e->net;
@@ -1383,6 +1249,8 @@ nl_send_route(struct krt_proto *p, rte *e, int op)
 
   if (p->af == AF_MPLS)
     priority = 0;
+  else if (a->source == RTS_DUMMY)
+    priority = e->u.krt.metric;
   else if (KRT_CF->sys.metric)
     priority = KRT_CF->sys.metric;
   else if ((op != NL_OP_DELETE) && (ea = ea_find(eattrs, EA_KRT_METRIC)))
@@ -1393,17 +1261,15 @@ nl_send_route(struct krt_proto *p, rte *e, int op)
 
   /* For route delete, we do not specify remaining route attributes */
   if (op == NL_OP_DELETE)
-    goto done;
+    goto dest;
 
   /* Default scope is LINK for device routes, UNIVERSE otherwise */
   if (p->af == AF_MPLS)
     r->r.rtm_scope = RT_SCOPE_UNIVERSE;
   else if (ea = ea_find(eattrs, EA_KRT_SCOPE))
     r->r.rtm_scope = ea->u.data;
-  else if (a->dest == RTD_UNICAST && ipa_zero(a->nh.gw))
-    r->r.rtm_scope = RT_SCOPE_LINK;
   else
-    r->r.rtm_scope = RT_SCOPE_UNIVERSE;
+    r->r.rtm_scope = (dest == RTD_UNICAST && ipa_zero(nh->gw)) ? RT_SCOPE_LINK : RT_SCOPE_UNIVERSE;
 
   if (ea = ea_find(eattrs, EA_KRT_PREFSRC))
     nl_add_attr_ipa(&r->h, rsize, RTA_PREFSRC, *(ip_addr *)ea->u.ptr->data);
@@ -1426,13 +1292,14 @@ nl_send_route(struct krt_proto *p, rte *e, int op)
   if (metrics[0])
     nl_add_metrics(&r->h, rsize, metrics, KRT_METRICS_MAX);
 
-  switch (a->dest)
+
+dest:
+  switch (dest)
     {
     case RTD_UNICAST:
       r->r.rtm_type = RTN_UNICAST;
-      struct nexthop *nh = &(a->nh);
-      if (nh->next)
-	nl_add_multipath(&r->h, rsize, nh, p->af, eattrs);
+      if (nh->next && !krt_ecmp6(p))
+	nl_add_multipath(&r->h, rsize, nh, p->af);
       else
       {
 	nl_add_attr_u32(&r->h, rsize, RTA_OIF, nh->iface->index);
@@ -1457,67 +1324,132 @@ nl_send_route(struct krt_proto *p, rte *e, int op)
       bug("krt_capable inconsistent with nl_send_route");
     }
 
-done:
   /* Ignore missing for DELETE */
   return nl_exchange(&r->h, (op == NL_OP_DELETE));
 }
 
 static inline int
-nl_allow_replace(struct krt_proto *p, rte *new)
+nl_add_rte(struct krt_proto *p, rte *e)
 {
-  /*
-   * We use NL_OP_REPLACE for IPv4, it has an issue with not checking for
-   * matching rtm_protocol, but that is OK when dedicated priority is used.
-   *
-   * For IPv6, the NL_OP_REPLACE is still broken even in Linux 4.19 LTS
-   * (although it seems to be fixed in Linux 5.10 LTS) for sequence:
-   *
-   * ip route add 2001:db8::/32 via fe80::1 dev eth0
-   * ip route replace 2001:db8::/32 dev eth0
-   *
-   * (it ends with two routes instead of replacing the first by the second one)
-   *
-   * Replacing with direct and special type (e.g. unreachable) routes does not
-   * work, but replacing with regular routes work reliably
-   */
+  rta *a = e->attrs;
+  int err = 0;
 
-  if (krt_ipv4(p))
-    return 1;
+  if (krt_ecmp6(p) && a->nh.next)
+  {
+    struct nexthop *nh = &(a->nh);
 
-  rta *a = new->attrs;
-  return (a->dest == RTD_UNICAST) && ipa_nonzero(a->nh.gw);
+    err = nl_send_route(p, e, NL_OP_ADD, RTD_UNICAST, nh);
+    if (err < 0)
+      return err;
+
+    for (nh = nh->next; nh; nh = nh->next)
+      err += nl_send_route(p, e, NL_OP_APPEND, RTD_UNICAST, nh);
+
+    return err;
+  }
+
+  return nl_send_route(p, e, NL_OP_ADD, a->dest, &(a->nh));
 }
 
-void
-krt_replace_rte(struct krt_proto *p, net *n UNUSED, rte *new, rte *old)
+static inline int
+nl_delete_rte(struct krt_proto *p, rte *e)
 {
   int err = 0;
 
-  if (old && new && nl_allow_replace(p, new))
-  {
-    err = nl_send_route(p, new, NL_OP_REPLACE);
-  }
-  else
-  {
-    if (old)
-      nl_send_route(p, old, NL_OP_DELETE);
+  /* For IPv6, we just repeatedly request DELETE until we get error */
+  do
+    err = nl_send_route(p, e, NL_OP_DELETE, RTD_NONE, NULL);
+  while (krt_ecmp6(p) && !err);
 
-    if (new)
-      err = nl_send_route(p, new, NL_OP_ADD);
-  }
+  return err;
+}
+
+void
+krt_replace_rte(struct krt_proto *p, net *n, rte *new, rte *old)
+{
+  int err = 0;
+
+  /*
+   * We could use NL_OP_REPLACE, but route replace on Linux has some problems:
+   *
+   * 1) Does not check for matching rtm_protocol
+   * 2) Has broken semantics for IPv6 ECMP
+   * 3) Crashes some kernel version when used for IPv6 ECMP
+   *
+   * So we use NL_OP_DELETE and then NL_OP_ADD. We also do not trust the old
+   * route value, so we do not try to optimize IPv6 ECMP reconfigurations.
+   */
+
+  if (old)
+    nl_delete_rte(p, old);
 
   if (new)
-  {
-    if (err < 0)
-      bmap_clear(&p->sync_map, new->id);
-    else
-      bmap_set(&p->sync_map, new->id);
-  }
+    err = nl_add_rte(p, new);
+
+  if (err < 0)
+    n->n.flags |= KRF_SYNC_ERROR;
+  else
+    n->n.flags &= ~KRF_SYNC_ERROR;
+}
+
+static int
+nl_mergable_route(struct nl_parse_state *s, net *net, struct krt_proto *p, uint priority, uint krt_type)
+{
+  /* Route merging must be active */
+  if (!s->merge)
+    return 0;
+
+  /* Saved and new route must have same network, proto/table, and priority */
+  if ((s->net != net) || (s->proto != p) || (s->krt_metric != priority))
+    return 0;
+
+  /* Both must be regular unicast routes */
+  if ((s->krt_type != RTN_UNICAST) || (krt_type != RTN_UNICAST))
+    return 0;
+
+  return 1;
+}
+
+static void
+nl_announce_route(struct nl_parse_state *s)
+{
+  rte *e = rte_get_temp(s->attrs);
+  e->net = s->net;
+  e->u.krt.src = s->krt_src;
+  e->u.krt.proto = s->krt_proto;
+  e->u.krt.seen = 0;
+  e->u.krt.best = 0;
+  e->u.krt.metric = s->krt_metric;
+
+  if (s->scan)
+    krt_got_route(s->proto, e);
+  else
+    krt_got_route_async(s->proto, e, s->new);
+
+  s->net = NULL;
+  s->attrs = NULL;
+  s->proto = NULL;
+  lp_flush(s->pool);
+}
+
+static inline void
+nl_parse_begin(struct nl_parse_state *s, int scan, int merge)
+{
+  memset(s, 0, sizeof (struct nl_parse_state));
+  s->pool = nl_linpool;
+  s->scan = scan;
+  s->merge = merge;
+}
+
+static inline void
+nl_parse_end(struct nl_parse_state *s)
+{
+  if (s->net)
+    nl_announce_route(s);
 }
 
 
-#define SKIP0(ARG, ...) do { DBG("KRT: Ignoring route - " ARG, ##__VA_ARGS__); return; } while(0)
-#define SKIP(ARG, ...)  do { DBG("KRT: Ignoring route %N - " ARG, &dst, ##__VA_ARGS__); return; } while(0)
+#define SKIP(ARG...) do { DBG("KRT: Ignoring route - " ARG); return; } while(0)
 
 static void
 nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
@@ -1570,10 +1502,10 @@ nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
 	return;
 
       if (!a[RTA_DST])
-	SKIP0("MPLS route without RTA_DST\n");
+	SKIP("MPLS route without RTA_DST");
 
       if (rta_get_mpls(a[RTA_DST], rta_mpls_stack) != 1)
-	SKIP0("MPLS route with multi-label RTA_DST\n");
+	SKIP("MPLS route with multi-label RTA_DST");
 
       net_fill_mpls(&dst, rta_mpls_stack[0]);
       break;
@@ -1591,9 +1523,6 @@ nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
   else
     table_id = i->rtm_table;
 
-  if (i->rtm_flags & RTM_F_CLONED)
-    SKIP("cloned\n");
-
   /* Do we know this table? */
   p = HASH_FIND(nl_table_map, RTH, i->rtm_family, table_id);
   if (!p)
@@ -1652,34 +1581,14 @@ nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
 
   net *net = net_get(p->p.main_channel->table, n);
 
+  if (s->net && !nl_mergable_route(s, net, p, priority, i->rtm_type))
+    nl_announce_route(s);
+
   rta *ra = lp_allocz(s->pool, RTA_MAX_SIZE);
+  ra->src = p->p.main_source;
   ra->source = RTS_INHERIT;
   ra->scope = SCOPE_UNIVERSE;
 
-  {
-    ea_list *ea = lp_alloc(s->pool, sizeof(ea_list) + 2 * sizeof(eattr));
-    *ea = (ea_list) { .flags = EALF_SORTED, .count = 2 };
-    ea->next = ra->eattrs;
-    ra->eattrs = ea;
-
-    ea->attrs[0] = (eattr) {
-      .id = EA_KRT_SOURCE,
-      .type = EAF_TYPE_INT,
-      .u.data = i->rtm_protocol
-    };
-
-    ea->attrs[1] = (eattr) {
-      .id = EA_KRT_METRIC,
-      .type = EAF_TYPE_INT,
-      .u.data = priority,
-    };
-  }
-
-  if (a[RTA_FLOW])
-    s->rta_flow = rta_get_u32(a[RTA_FLOW]);
-  else
-    s->rta_flow = 0;
-
   switch (i->rtm_type)
     {
     case RTN_UNICAST:
@@ -1687,17 +1596,17 @@ nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
 
       if (a[RTA_MULTIPATH])
         {
-	  struct nexthop *nh = nl_parse_multipath(s, p, n, a[RTA_MULTIPATH], i->rtm_family, krt_src);
+	  struct nexthop *nh = nl_parse_multipath(s, p, a[RTA_MULTIPATH], i->rtm_family);
 	  if (!nh)
-	    SKIP("strange RTA_MULTIPATH\n");
+	    {
+	      log(L_ERR "KRT: Received strange multipath route %N", net->n.addr);
+	      return;
+	    }
 
-	  nexthop_link(ra, nh);
+	  ra->nh = *nh;
 	  break;
 	}
 
-      if ((i->rtm_flags & RTNH_F_DEAD) && (krt_src != KRT_SRC_BIRD))
-	SKIP("ignore RTNH_F_DEAD\n");
-
       ra->nh.iface = if_find_by_index(oif);
       if (!ra->nh.iface)
 	{
@@ -1705,24 +1614,27 @@ nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
 	  return;
 	}
 
-      if (a[RTA_GATEWAY])
-	ra->nh.gw = rta_get_ipa(a[RTA_GATEWAY]);
-
+      if ((i->rtm_family != AF_MPLS) && a[RTA_GATEWAY]
 #ifdef HAVE_MPLS_KERNEL
-      if (a[RTA_VIA])
-	ra->nh.gw = rta_get_via(a[RTA_VIA]);
+	  || (i->rtm_family == AF_MPLS) && a[RTA_VIA]
 #endif
-
-      if (i->rtm_flags & RTNH_F_ONLINK)
-	ra->nh.flags |= RNF_ONLINK;
-
-      if (ipa_nonzero(ra->nh.gw))
+	  )
 	{
+#ifdef HAVE_MPLS_KERNEL
+	  if (i->rtm_family == AF_MPLS)
+	    ra->nh.gw = rta_get_via(a[RTA_VIA]);
+	  else
+#endif
+	    ra->nh.gw = rta_get_ipa(a[RTA_GATEWAY]);
+
 	  /* Silently skip strange 6to4 routes */
 	  const net_addr_ip6 sit = NET_ADDR_IP6(IP6_NONE, 96);
 	  if ((i->rtm_family == AF_INET6) && ipa_in_netX(ra->nh.gw, (net_addr *) &sit))
 	    return;
 
+	  if (i->rtm_flags & RTNH_F_ONLINK)
+	    ra->nh.flags |= RNF_ONLINK;
+
 	  neighbor *nbr;
 	  nbr = neigh_find(&p->p, ra->nh.gw, ra->nh.iface,
 			   (ra->nh.flags & RNF_ONLINK) ? NEF_ONLINK : 0);
@@ -1751,8 +1663,9 @@ nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
     }
 
 #ifdef HAVE_MPLS_KERNEL
+  int labels = 0;
   if ((i->rtm_family == AF_MPLS) && a[RTA_NEWDST] && !ra->nh.next)
-    ra->nh.labels = rta_get_mpls(a[RTA_NEWDST], ra->nh.label);
+    labels = rta_get_mpls(a[RTA_NEWDST], ra->nh.label);
 
   if (a[RTA_ENCAP] && a[RTA_ENCAP_TYPE] && !ra->nh.next)
     {
@@ -1763,7 +1676,7 @@ nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
 	      struct rtattr *enca[BIRD_RTA_MAX];
 	      nl_attr_len = RTA_PAYLOAD(a[RTA_ENCAP]);
 	      nl_parse_attrs(RTA_DATA(a[RTA_ENCAP]), encap_mpls_want, enca, sizeof(enca));
-	      ra->nh.labels = rta_get_mpls(enca[RTA_DST], ra->nh.label);
+	      labels = rta_get_mpls(enca[RTA_DST], ra->nh.label);
 	      break;
 	    }
 	  default:
@@ -1771,6 +1684,14 @@ nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
 	    break;
 	}
     }
+
+  if (labels < 0)
+  {
+    log(L_WARN "KRT: Too long MPLS stack received, ignoring.");
+    ra->nh.labels = 0;
+  }
+  else
+    ra->nh.labels = labels;
 #endif
 
   if (i->rtm_scope != def_scope)
@@ -1780,49 +1701,40 @@ nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
       ra->eattrs = ea;
       ea->flags = EALF_SORTED;
       ea->count = 1;
-      ea->attrs[0] = (eattr) {
-        .id = EA_KRT_SCOPE,
-        .flags = 0,
-        .type = EAF_TYPE_INT,
-        .u.data = i->rtm_scope,
-      };
+      ea->attrs[0].id = EA_KRT_SCOPE;
+      ea->attrs[0].flags = 0;
+      ea->attrs[0].type = EAF_TYPE_INT;
+      ea->attrs[0].u.data = i->rtm_scope;
     }
 
   if (a[RTA_PREFSRC])
     {
       ip_addr ps = rta_get_ipa(a[RTA_PREFSRC]);
 
-      struct adata *ad = lp_alloc(s->pool, sizeof(struct adata) + sizeof(ps));
-      ad->length = sizeof(ps);
-      memcpy(ad->data, &ps, sizeof(ps));
-
       ea_list *ea = lp_alloc(s->pool, sizeof(ea_list) + sizeof(eattr));
       ea->next = ra->eattrs;
       ra->eattrs = ea;
       ea->flags = EALF_SORTED;
       ea->count = 1;
-      ea->attrs[0] = (eattr) {
-	.id = EA_KRT_PREFSRC,
-        .flags = 0,
-        .type = EAF_TYPE_IP_ADDRESS,
-        .u.ptr = ad,
-      };
+      ea->attrs[0].id = EA_KRT_PREFSRC;
+      ea->attrs[0].flags = 0;
+      ea->attrs[0].type = EAF_TYPE_IP_ADDRESS;
+      ea->attrs[0].u.ptr = lp_alloc(s->pool, sizeof(struct adata) + sizeof(ps));
+      ea->attrs[0].u.ptr->length = sizeof(ps);
+      memcpy(ea->attrs[0].u.ptr->data, &ps, sizeof(ps));
     }
 
-  /* Can be set per-route or per-nexthop */
-  if (s->rta_flow)
+  if (a[RTA_FLOW])
     {
       ea_list *ea = lp_alloc(s->pool, sizeof(ea_list) + sizeof(eattr));
       ea->next = ra->eattrs;
       ra->eattrs = ea;
       ea->flags = EALF_SORTED;
       ea->count = 1;
-      ea->attrs[0] = (eattr) {
-	.id = EA_KRT_REALM,
-	.flags = 0,
-	.type = EAF_TYPE_INT,
-	.u.data = s->rta_flow,
-      };
+      ea->attrs[0].id = EA_KRT_REALM;
+      ea->attrs[0].flags = 0;
+      ea->attrs[0].type = EAF_TYPE_INT;
+      ea->attrs[0].u.data = rta_get_u32(a[RTA_FLOW]);
     }
 
   if (a[RTA_METRICS])
@@ -1839,12 +1751,13 @@ nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
 
       for (t = 1; t < KRT_METRICS_MAX; t++)
 	if (metrics[0] & (1 << t))
-	  ea->attrs[n++] = (eattr) {
-	    .id = EA_CODE(PROTOCOL_KERNEL, KRT_METRICS_OFFSET + t),
-	    .flags = 0,
-	    .type = EAF_TYPE_INT, /* FIXME: Some are EAF_TYPE_BITFIELD */
-	    .u.data = metrics[t],
-	  };
+	  {
+	    ea->attrs[n].id = EA_CODE(PROTOCOL_KERNEL, KRT_METRICS_OFFSET + t);
+	    ea->attrs[n].flags = 0;
+	    ea->attrs[n].type = EAF_TYPE_INT; /* FIXME: Some are EAF_TYPE_BITFIELD */
+	    ea->attrs[n].u.data = metrics[t];
+	    n++;
+	  }
 
       if (n > 0)
         {
@@ -1855,40 +1768,80 @@ nl_parse_route(struct nl_parse_state *s, struct nlmsghdr *h)
 	}
     }
 
-  rte *e = rte_get_temp(ra, p->p.main_source);
-  e->net = net;
+  /*
+   * Ideally, now we would send the received route to the rest of kernel code.
+   * But IPv6 ECMP routes before 4.11 are sent as a sequence of routes, so we
+   * postpone it and merge next hops until the end of the sequence. Note that
+   * when doing merging of next hops, we expect the new route to be unipath.
+   * Otherwise, we ignore additional next hops in nexthop_insert().
+   */
 
-  if (s->scan)
-    krt_got_route(p, e, krt_src);
+  if (!s->net)
+  {
+    /* Store the new route */
+    s->net = net;
+    s->attrs = ra;
+    s->proto = p;
+    s->new = new;
+    s->krt_src = krt_src;
+    s->krt_type = i->rtm_type;
+    s->krt_proto = i->rtm_protocol;
+    s->krt_metric = priority;
+  }
   else
-    krt_got_route_async(p, e, new, krt_src);
+  {
+    /* Merge next hops with the stored route */
+    rta *oa = s->attrs;
 
-  lp_flush(s->pool);
+    struct nexthop *nhs = &oa->nh;
+    nexthop_insert(&nhs, &ra->nh);
+
+    /* Perhaps new nexthop is inserted at the first position */
+    if (nhs == &ra->nh)
+    {
+      /* Swap rtas */
+      s->attrs = ra;
+
+      /* Keep old eattrs */
+      ra->eattrs = oa->eattrs;
+    }
+  }
 }
 
 void
-krt_do_scan(struct krt_proto *p)
+krt_do_scan(struct krt_proto *p UNUSED)	/* CONFIG_ALL_TABLES_AT_ONCE => p is NULL */
 {
-  struct nl_parse_state s = {
-    .proto = p,
-    .pool = nl_linpool,
-    .scan = 1,
-  };
-
-  /* Table-specific scan or shared scan */
-  if (p)
-    nl_request_dump_route(p->af, krt_table_id(p));
-  else
-    nl_request_dump_route(AF_UNSPEC, 0);
-
   struct nlmsghdr *h;
+  struct nl_parse_state s;
+
+  nl_parse_begin(&s, 1, 0);
+  nl_request_dump(AF_INET, RTM_GETROUTE);
   while (h = nl_get_scan())
-  {
     if (h->nlmsg_type == RTM_NEWROUTE || h->nlmsg_type == RTM_DELROUTE)
       nl_parse_route(&s, h);
     else
       log(L_DEBUG "nl_scan_fire: Unknown packet received (type=%d)", h->nlmsg_type);
-  }
+  nl_parse_end(&s);
+
+  nl_parse_begin(&s, 1, 1);
+  nl_request_dump(AF_INET6, RTM_GETROUTE);
+  while (h = nl_get_scan())
+    if (h->nlmsg_type == RTM_NEWROUTE || h->nlmsg_type == RTM_DELROUTE)
+      nl_parse_route(&s, h);
+    else
+      log(L_DEBUG "nl_scan_fire: Unknown packet received (type=%d)", h->nlmsg_type);
+  nl_parse_end(&s);
+
+#ifdef HAVE_MPLS_KERNEL
+  nl_parse_begin(&s, 1, 1);
+  nl_request_dump(AF_MPLS, RTM_GETROUTE);
+  while (h = nl_get_scan())
+    if (h->nlmsg_type == RTM_NEWROUTE || h->nlmsg_type == RTM_DELROUTE)
+      nl_parse_route(&s, h);
+    else
+      log(L_DEBUG "nl_scan_fire: Unknown packet received (type=%d)", h->nlmsg_type);
+  nl_parse_end(&s);
+#endif
 }
 
 /*
@@ -1897,24 +1850,20 @@ krt_do_scan(struct krt_proto *p)
 
 static sock *nl_async_sk;		/* BIRD socket for asynchronous notifications */
 static byte *nl_async_rx_buffer;	/* Receive buffer */
-static uint nl_async_bufsize;		/* Kernel rx buffer size for the netlink socket */
-static struct config *nl_last_config;	/* For tracking changes to nl_async_bufsize */
 
 static void
 nl_async_msg(struct nlmsghdr *h)
 {
-  struct nl_parse_state s = {
-    .proto = NULL,
-    .pool = nl_linpool,
-    .scan = 0,
-  };
+  struct nl_parse_state s;
 
   switch (h->nlmsg_type)
     {
     case RTM_NEWROUTE:
     case RTM_DELROUTE:
       DBG("KRT: Received async route notification (%d)\n", h->nlmsg_type);
+      nl_parse_begin(&s, 0, 0);
       nl_parse_route(&s, h);
+      nl_parse_end(&s);
       break;
     case RTM_NEWLINK:
     case RTM_DELLINK:
@@ -2036,32 +1985,6 @@ nl_open_async(void)
     bug("Netlink: sk_open failed");
 }
 
-static void
-nl_update_async_bufsize(void)
-{
-  /* No async socket */
-  if (!nl_async_sk)
-    return;
-
-  /* Already reconfigured */
-  if (nl_last_config == config)
-    return;
-
-  /* Update netlink buffer size */
-  uint bufsize = nl_cfg_rx_buffer_size(config);
-  if (bufsize && (bufsize != nl_async_bufsize))
-  {
-    /* Log message for reconfigurations only */
-    if (nl_last_config)
-      log(L_INFO "KRT: Changing netlink rx buffer size to %u", bufsize);
-
-    nl_set_rcvbuf(nl_async_sk->fd, bufsize);
-    nl_async_bufsize = bufsize;
-  }
-
-  nl_last_config = config;
-}
-
 
 /*
  *	Interface to the UNIX krt module
@@ -2090,7 +2013,6 @@ krt_sys_start(struct krt_proto *p)
 
   nl_open();
   nl_open_async();
-  nl_update_async_bufsize();
 
   return 1;
 }
@@ -2098,16 +2020,12 @@ krt_sys_start(struct krt_proto *p)
 void
 krt_sys_shutdown(struct krt_proto *p)
 {
-  nl_update_async_bufsize();
-
   HASH_REMOVE2(nl_table_map, RTH, krt_pool, p);
 }
 
 int
 krt_sys_reconfigure(struct krt_proto *p UNUSED, struct krt_config *n, struct krt_config *o)
 {
-  nl_update_async_bufsize();
-
   return (n->sys.table_id == o->sys.table_id) && (n->sys.metric == o->sys.metric);
 }
 
@@ -2135,7 +2053,7 @@ static const char *krt_features_names[KRT_FEATURES_MAX] = {
 };
 
 int
-krt_sys_get_attr(const eattr *a, byte *buf, int buflen UNUSED)
+krt_sys_get_attr(eattr *a, byte *buf, int buflen UNUSED)
 {
   switch (a->id)
   {
