@@ -91,7 +91,7 @@ static struct resclass rf_class = {
 };
 
 struct rfile *
-rf_open(pool *p, const char *name, const char *mode)
+rf_open(pool *p, char *name, char *mode)
 {
   FILE *f = fopen(name, mode);
 
@@ -132,7 +132,7 @@ times_init(struct timeloop *loop)
   if (rv < 0)
     die("Monotonic clock is missing");
 
-  if ((ts.tv_sec < 0) || (((u64) ts.tv_sec) > ((u64) 1 << 40)))
+  if ((ts.tv_sec < 0) || (((s64) ts.tv_sec) > ((s64) 1 << 40)))
     log(L_WARN "Monotonic clock is crazy");
 
   loop->last_time = ts.tv_sec S + ts.tv_nsec NS;
@@ -1082,7 +1082,6 @@ sk_passive_connected(sock *s, int type)
   t->fd = fd;
   t->ttl = s->ttl;
   t->tos = s->tos;
-  t->vrf = s->vrf;
   t->rbsize = s->rbsize;
   t->tbsize = s->tbsize;
 
@@ -1436,17 +1435,13 @@ sk_open(sock *s)
 	if (sk_set_high_port(s) < 0)
 	  log(L_WARN "Socket error: %s%#m", s->err);
 
-    if (s->flags & SKF_FREEBIND)
-      if (sk_set_freebind(s) < 0)
-        log(L_WARN "Socket error: %s%#m", s->err);
-
     sockaddr_fill(&sa, s->af, bind_addr, s->iface, bind_port);
     if (bind(fd, &sa.sa, SA_LEN(sa)) < 0)
       ERR2("bind");
   }
 
   if (s->password)
-    if (sk_set_md5_auth(s, s->saddr, s->daddr, -1, s->iface, s->password, 0) < 0)
+    if (sk_set_md5_auth(s, s->saddr, s->daddr, s->iface, s->password, 0) < 0)
       goto err;
 
   switch (s->type)
@@ -1499,9 +1494,7 @@ sk_open_unix(sock *s, char *name)
   if (fcntl(fd, F_SETFL, O_NONBLOCK) < 0)
     return -1;
 
-  /* Path length checked in test_old_bird() but we may need unix sockets for other reasons in future */
-  ASSERT_DIE(strlen(name) < sizeof(sa.sun_path));
-
+  /* Path length checked in test_old_bird() */
   sa.sun_family = AF_UNIX;
   strcpy(sa.sun_path, name);
 
@@ -1854,8 +1847,8 @@ sk_read_ssh(sock *s)
 
  /* sk_read() and sk_write() are called from BFD's event loop */
 
-static inline int
-sk_read_noflush(sock *s, int revents)
+int
+sk_read(sock *s, int revents)
 {
   switch (s->type)
   {
@@ -1918,15 +1911,7 @@ sk_read_noflush(sock *s, int revents)
 }
 
 int
-sk_read(sock *s, int revents)
-{
-  int e = sk_read_noflush(s, revents);
-  tmp_flush();
-  return e;
-}
-
-static inline int
-sk_write_noflush(sock *s)
+sk_write(sock *s)
 {
   switch (s->type)
   {
@@ -1974,14 +1959,6 @@ sk_write_noflush(sock *s)
   }
 }
 
-int
-sk_write(sock *s)
-{
-  int e = sk_write_noflush(s);
-  tmp_flush();
-  return e;
-}
-
 int sk_is_ipv4(sock *s)
 { return s->af == AF_INET; }
 
@@ -2000,7 +1977,6 @@ sk_err(sock *s, int revents)
     }
 
   s->err_hook(s, se);
-  tmp_flush();
 }
 
 void
@@ -2063,8 +2039,8 @@ io_update_time(void)
     event_open->duration = last_time - event_open->timestamp;
 
     if (event_open->duration > config->latency_limit)
-      log(L_WARN "Event 0x%p 0x%p took %u.%03u ms",
-	  event_open->hook, event_open->data, (uint) (event_open->duration TO_MS), (uint) (event_open->duration % 1000));
+      log(L_WARN "Event 0x%p 0x%p took %d ms",
+	  event_open->hook, event_open->data, (int) (event_open->duration TO_MS));
 
     event_open = NULL;
   }
@@ -2168,8 +2144,8 @@ watchdog_stop(void)
 
   btime duration = last_time - loop_time;
   if (duration > config->watchdog_warning)
-    log(L_WARN "I/O loop cycle took %u.%03u ms for %d events",
-	(uint) (duration TO_MS), (uint) (duration % 1000), event_log_num);
+    log(L_WARN "I/O loop cycle took %d ms for %d events",
+	(int) (duration TO_MS), event_log_num);
 }
 
 
@@ -2177,12 +2153,15 @@ watchdog_stop(void)
  *	Main I/O Loop
  */
 
+volatile int async_config_flag;		/* Asynchronous reconfiguration/dump scheduled */
+volatile int async_dump_flag;
+volatile int async_shutdown_flag;
+
 void
 io_init(void)
 {
   init_list(&sock_list);
   init_list(&global_event_list);
-  init_list(&global_work_list);
   krt_io_init();
   // XXX init_times();
   // XXX update_times();
@@ -2194,7 +2173,6 @@ io_init(void)
 
 static int short_loops = 0;
 #define SHORT_LOOP_MAX 10
-#define WORK_EVENTS_MAX 10
 
 void
 io_loop(void)
@@ -2212,7 +2190,6 @@ io_loop(void)
     {
       times_update(&main_timeloop);
       events = ev_run_list(&global_event_list);
-      events = ev_run_list_limited(&global_work_list, WORK_EVENTS_MAX) || events;
       timers_fire(&main_timeloop);
       io_close_event();
 
