@@ -23,7 +23,7 @@ void async_config(void);
 void async_dump(void);
 void async_shutdown(void);
 void cmd_check_config(char *name);
-void cmd_reconfig(char *name, int type, int timeout);
+void cmd_reconfig(char *name, int type, uint timeout);
 void cmd_reconfig_confirm(void);
 void cmd_reconfig_undo(void);
 void cmd_shutdown(void);
@@ -48,14 +48,6 @@ typedef struct sockaddr_bird {
 } sockaddr;
 
 
-#ifdef IPV6
-#define BIRD_AF AF_INET6
-#define ipa_from_sa(x) ipa_from_sa6(x)
-#else
-#define BIRD_AF AF_INET
-#define ipa_from_sa(x) ipa_from_sa4(x)
-#endif
-
 
 /* This is sloppy hack, it should be detected by configure script */
 /* Linux systems have it defined so this is definition for BSD systems */
@@ -64,29 +56,36 @@ typedef struct sockaddr_bird {
 #endif
 
 
-static inline ip_addr ipa_from_in4(struct in_addr a UNUSED6)
+static inline ip_addr ipa_from_in4(struct in_addr a)
 { return ipa_from_u32(ntohl(a.s_addr)); }
 
-static inline ip_addr ipa_from_in6(struct in6_addr a UNUSED4)
+static inline ip_addr ipa_from_in6(struct in6_addr a)
 { return ipa_build6(ntohl(a.s6_addr32[0]), ntohl(a.s6_addr32[1]), ntohl(a.s6_addr32[2]), ntohl(a.s6_addr32[3])); }
 
-static inline ip_addr ipa_from_sa4(sockaddr *sa UNUSED6)
+static inline ip_addr ipa_from_sa4(sockaddr *sa)
 { return ipa_from_in4(((struct sockaddr_in *) sa)->sin_addr); }
 
-static inline ip_addr ipa_from_sa6(sockaddr *sa UNUSED4)
+static inline ip_addr ipa_from_sa6(sockaddr *sa)
 { return ipa_from_in6(((struct sockaddr_in6 *) sa)->sin6_addr); }
 
+static inline ip_addr ipa_from_sa(sockaddr *sa)
+{
+  switch (sa->sa.sa_family)
+  {
+  case AF_INET:		return ipa_from_sa4(sa);
+  case AF_INET6:	return ipa_from_sa6(sa);
+  default:		return IPA_NONE;
+  }
+}
+
 static inline struct in_addr ipa_to_in4(ip_addr a)
 { return (struct in_addr) { htonl(ipa_to_u32(a)) }; }
 
-#ifdef IPV6
+static inline struct in_addr ip4_to_in4(ip4_addr a)
+{ return (struct in_addr) { htonl(ip4_to_u32(a)) }; }
+
 static inline struct in6_addr ipa_to_in6(ip_addr a)
 { return (struct in6_addr) { .s6_addr32 = { htonl(_I0(a)), htonl(_I1(a)), htonl(_I2(a)), htonl(_I3(a)) } }; }
-#else
-/* Temporary dummy */
-static inline struct in6_addr ipa_to_in6(ip_addr a UNUSED)
-{ return (struct in6_addr) { .s6_addr32 = { 0, 0, 0, 0 } }; }
-#endif
 
 void sockaddr_fill(sockaddr *sa, int af, ip_addr a, struct iface *ifa, uint port);
 int sockaddr_read(sockaddr *sa, int af, ip_addr *a, struct iface **ifa, uint *port);
@@ -109,7 +108,6 @@ void *rf_file(struct rfile *f);
 int rf_fileno(struct rfile *f);
 void test_old_bird(char *path);
 
-
 /* krt.c bits */
 
 void krt_io_init(void);
@@ -118,12 +116,17 @@ void krt_io_init(void);
 
 void main_thread_init(void);
 void log_init_debug(char *);		/* Initialize debug dump to given file (NULL=stderr, ""=off) */
-void log_switch(int debug, list *l, char *); /* Use l=NULL for initial switch */
+void log_switch(int initial, list *l, char *);
 
 struct log_config {
   node n;
   uint mask;				/* Classes to log */
   void *fh;				/* FILE to log to, NULL=syslog */
+  struct rfile *rf;			/* Resource for log file */
+  char *filename;			/* Log filename */
+  char *backup;				/* Secondary filename (for log rotation) */
+  off_t pos;				/* Position/size of current log */
+  off_t limit;				/* Log size limit */
   int terminal_flag;
 };
 
